/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
#include "gaplm.h"
#include "utils.h"

/**
 * This constructor is used for stand-alone training mode.
 * in_filename - file containing training data, one instance per line
 * out_filename - root of filenames to use for saved models
 */
GapLM::GapLM(const string& in_filename, const string& count_out_filename, const string& score_filename, const string& solo_words_filename,
		const string& srclang_in_filename, const string& srctgt_align_filename, const string& srclang_solo_words_filename,
		bool discrim, bool bigram, bool bilingual, bool onlyNull,
		double alpha, double alpha2, double lambdaW, double lambdaS, double initNewColor, double initPropagate,
		int burnin, bool useGapWidths, int maxGapWidth) {

	// TODO seed random number generator using a seed from a command-line parameter. line below used for debugging.
	srand(123);
	m_burnin = burnin;
	// set the filename variables in the model
	m_inputFilename = in_filename;
	m_scoreFilename = score_filename;
	m_srclangInputFilename = srclang_in_filename;
	m_srctgtAlignFilename = srctgt_align_filename;
	m_countOutputFilename = count_out_filename;
	m_countWidthOutputFilename = count_out_filename + ".widths";
	m_soloWordsFilename = solo_words_filename;
	m_srclangSoloWordsFilename = srclang_solo_words_filename;

	// Set the hyperparameter values and cache log values where appropriate.
	m_alpha = alpha;
	m_logalpha = log(alpha);
	m_alpha2 = alpha2;
	m_logalpha2 = log(alpha2);
	m_lambdaW = lambdaW;
	m_loglambdaW = log(lambdaW);
	m_lambdaS = lambdaS;
	m_loglambdaS = log(lambdaS);

	m_pInitNewColor = initNewColor;
	m_pInitPropagate = initPropagate;
	m_logPInitPropagate = log(m_pInitPropagate);
	m_logOneMinusPInitPropagate = log(1.0 - m_pInitPropagate);

	m_useGapWidths = useGapWidths;
	m_maxGapWidth = maxGapWidth;

	initCommon();

	cerr << "Reading data..." << endl;
	// Load the training data into memory, initialize boundary positions randomly, and
	// compute initial counts.
	readData(discrim, bigram, bilingual, onlyNull);
	cerr << "Done reading data." << endl;
}

void GapLM::initCommon() {
	m_logging = false;
	m_numWords = 0;
	m_numColors = 0;
	m_numColors2 = 0;
	m_numTables = 0;
	m_timeProfile = 0.0;
	m_cumSumCounts = 0.0;
	m_cumSumCounts2 = 0.0;

	// Create and initialize the vocabulary.
	// Since index 0 indicates word was not found, add the word ~~DUMMYDUMMYDUMMY~~
	// (unlikely to ever appear) as the word to fill in slot 0.
	m_vocabulary.storeIfNew("~~DUMMYDUMMYDUMMY~~");

	// cache the Poisson pdf for values 0 up to 1000
	for (int i = 0; i <= 1000; i++) {
		m_logPoissonWPDF.push_back(logPoisson(m_loglambdaW, m_lambdaW, i));
		m_logPoissonSPDF.push_back(logPoisson(m_loglambdaS, m_lambdaS, i));
	}

	// Cache the "n choose k" quantities.
	for (int i = 0; i < 30; ++i) {
		for (int j = 0; j < 30; ++j) {
			m_precomputedLogOneOverNChooseK[i][j] = log(1.0 / ((double) nchoosek(i, j)));
		}
	}

	// Set empty and deleted keys in counts hash map.
	GENERIC_STRING nullString = HASHMAP_NULL_STRING;
	GENERIC_STRING deletedString = HASHMAP_DELETED_STRING;

	m_counts.set_empty_key(nullString);
	m_counts.set_deleted_key(deletedString);
	m_counts2.set_empty_key(nullString);
	m_counts2.set_deleted_key(deletedString);

	m_unigramProbabilities.set_empty_key(nullString);
	m_srclangUnigramProbabilities.set_empty_key(nullString);
	m_soloWords.set_empty_key(nullString);
	m_srclangSoloWords.set_empty_key(nullString);

	// set the string to use as delimiter for bigram counts to "~~"
	// todo: ensure that "~~" never appears in the training corpus
	//m_delimiterString = DELIMITER_STRING;
	//m_beginningOfSentence = BEGINNING_OF_SENTENCE_STRING;
	//m_endOfSentence = END_OF_SENTENCE_STRING;
}

GapLM::~GapLM() {}

void GapLM::setInitialParameterValue(double newInitialParameterValue) {
	m_initialParameterValue = newInitialParameterValue;
}

bool GapLM::openFilesSuccess(bool discrim, bool bilingual) {
	bool success = true;
    input_file.open(m_inputFilename.c_str());
    if (!input_file) {
   	    cerr << "Cannot open input file " << m_inputFilename << ". Exiting..." << endl;
		success = false;
    } else {
		if (USING_WSTRINGS) {
	    	input_file.imbue(locale("en_US.utf8"));
		}
	}
	if (discrim) {
	    score_file.open(m_scoreFilename.c_str());
	    if (!score_file) {
   		    cerr << "Cannot open score file " << m_scoreFilename << ". Exiting..." << endl;
			success = false;
		}
	}
	if (bilingual) {
	    srclang_input_file.open(m_srclangInputFilename.c_str());
	    if (!srclang_input_file) {
   		    cerr << "Cannot open source language input file " << m_srclangInputFilename << ". Exiting..." << endl;
			success = false;
	    } else {
			if (USING_WSTRINGS) {
	    		srclang_input_file.imbue(locale("en_US.utf8"));
			}
		}

	    srctgt_align_file.open(m_srctgtAlignFilename.c_str());
	    if (!srctgt_align_file) {
   		    cerr << "Cannot open source-target alignment file " << m_srctgtAlignFilename << ". Exiting..." << endl;
			success = false;
	    } else {
			if (USING_WSTRINGS) {
	    		srctgt_align_file.imbue(locale("en_US.utf8"));
			}
		}
	}
	return success;
}

void GapLM::readData(bool discrim, bool bigram, bool bilingual, bool bilingnull) {
	// Open the files, exit immediately if any failed.
	if (!openFilesSuccess(discrim, bilingual)) {
		exit(1);
	}

	int ctr = 1;

	// Before reading in any data, read the solo words files.
	if (m_soloWordsFilename != "||NOFILESPECIFIED||") {
		GENERIC_IFSTREAM solo_words_file;
		solo_words_file.open(m_soloWordsFilename.c_str());
		if (!solo_words_file) {
			cerr << "Cannot open input file " << m_soloWordsFilename << ". Exiting..." << endl;
		} else {
			if (USING_WSTRINGS) {
				solo_words_file.imbue(locale("en_US.utf8"));
			}

			// Read the file: there should be one solo word per line.
			GENERIC_STRING line;
			// Go through all lines in the file.
		    while (!solo_words_file.eof()) {
			    getline(solo_words_file, line);
			    if (line != EMPTY_STRING) {
					// Insert non-empty lines into the hash set of solo words.
					m_soloWords.insert(line);
				}
			}
		}
	}

	// Read the solo words file for the source language if we're using the bilingual model.
	if (bilingual && m_srclangSoloWordsFilename != "||NOFILESPECIFIED||") {
		GENERIC_IFSTREAM solo_words_file;
		solo_words_file.open(m_srclangSoloWordsFilename.c_str());
		if (!solo_words_file) {
			cerr << "Cannot open input file " << m_srclangSoloWordsFilename << ". Exiting..." << endl;
		} else {
			if (USING_WSTRINGS) {
				solo_words_file.imbue(locale("en_US.utf8"));
			}

			// Read the file: there should be one solo word per line.
			GENERIC_STRING line;
			// Go through all lines in the file.
		    while (!solo_words_file.eof()) {
			    getline(solo_words_file, line);
			    if (line != EMPTY_STRING) {
					// Insert non-empty lines into the hash set of solo words.
					m_srclangSoloWords.insert(line);
				}
			}
		}
	}

	GENERIC_STRING str1, str2, alignmentString;

    while (!input_file.eof()) {
		// Get a line of target language data.
	    getline(input_file, str1);
		if (bilingual) {
			// Get a line of source language data.
			getline(srclang_input_file, str2);
			// Get a line of source-target word alignments.
			getline(srctgt_align_file, alignmentString);
			if (str1 != EMPTY_STRING && str2 != EMPTY_STRING && alignmentString != EMPTY_STRING) {
				if (bilingnull) {
					AlignedSentence as = readSentenceBilingNull(str2, str1, alignmentString, ctr);
					if (!(as.onlySoloWords)) m_alignedData.push_back(as);
					++ctr;
				} else {
					SentencePair sp = readSentenceBiling(str2, str1, alignmentString, ctr);
					m_bilingualData.push_back(sp);
					++ctr;
				}
			}
		} else {
		    if (str1 != EMPTY_STRING) {

				Sentence s;
				if (discrim) {
					// Get the next line from the score file.
					getline(score_file, str2);
					double sentenceScore = 0.0;
					if (str2 == EMPTY_STRING) {
						cerr << "Warning: for sentence " << ctr << ", no score provided. Using 0.0." << endl;
					} else {
						sentenceScore = atof(str2.c_str());
					}
					s = readSentenceMonolingDiscrim(str1, ctr, sentenceScore);
				} else {
					s = readSentenceMonoling(str1, ctr);
				}

				m_data.push_back(s);
				++ctr;
			}
		}
	}
	cerr << "Done reading data" << endl;
	input_file.close();
	if (bilingual || bilingnull) {
		srclang_input_file.close();
		srctgt_align_file.close();
	}

	// Compute unigram probabilities.
	if (!bilingual) computeUnigramProbabilities();

	cerr << "Counting words" << endl;
	bool includeWidthsFalse = false;
	if (bilingual) {
		if (bilingnull) {
			countWordsBilingNull(includeWidthsFalse);
		} else {
			countWordsBiling(includeWidthsFalse);
		}
	} else {
		if (discrim) {
			countWordsMonolingDiscrim(includeWidthsFalse);
		} else {
			countWordsMonoling(includeWidthsFalse);
		}
	}
}

void GapLM::computeUnigramProbabilities() {
	double sumCounts = 0.0;

	dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::const_iterator iter;
	for (iter = m_unigramProbabilities.begin(); iter != m_unigramProbabilities.end(); ++iter) {
		sumCounts += iter->second;
	}
	if (sumCounts < 0.0) {
		cerr << "Error: sumCounts (" << sumCounts << ") < 0.0!" << endl;
		sumCounts = -sumCounts;
	}
	sumCounts = log(sumCounts);
	cerr << "sumCounts = " << sumCounts << endl;
	dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator iter2;
	for (iter2 = m_unigramProbabilities.begin(); iter2 != m_unigramProbabilities.end(); ++iter2) {
		iter2->second = log(iter2->second) - sumCounts;
	}
}

int GapLM::normalizeAndSampleMultinomial(vector<MultinomialEvent>& m) {
	// Go through all events in the multinomial and add their scores.
	double totalScore = neg_inf;
	vector<MultinomialEvent>::iterator iter;
	for (iter = m.begin(); iter != m.end(); ++iter) {
		double currScore = iter->score;
		if (currScore != currScore) {
			cerr << "Warning: multinomial event has score " << currScore << ", phraseWithCurr = " << iter->phraseWithCurr
				 << ", phraseWithoutCurr = " << iter->phraseWithoutCurr << ", color = " << iter->color << ", srcColor = " << iter->srcColor
				 << ", oldDistIndex = "
				 << iter->oldDistIndex << ", distIndex = " << iter->distIndex << endl;
		} else {
			totalScore = logAdd(totalScore, currScore);
		}
	}
	// Go back through and divide by the sum to get the probability for each event.
	// At the same time, fill in the cumulative probability and sample.
	int sampledEvent = -1;
	double randDouble = log(((double) rand()) / ((double) RAND_MAX));
	double cumProb = neg_inf;
	int ctr = 0;
	//cerr << "randDouble = " << randDouble << endl;
	for (iter = m.begin(); iter != m.end(); ++iter) {
		// Ignore NaNs.
		if (iter->score == iter->score) {
			// Log-divide by the sum to get the log-probability of this event.
			iter->score = iter->score - totalScore;
			// Update the cumulative probability for this multinomial by log-adding the log-prob.
			cumProb = logAdd(cumProb, iter->score);
			// Check if the random draw chose this event.
			if (randDouble <= cumProb) {
				//	cerr << "setting sampledEvent to ctr = " << ctr << endl;
				sampledEvent = ctr;
				break;
			}
		}
		ctr++;
	}
	if (sampledEvent == -1) {
		ctr = 0;
		int lastValidEvent = -1;
		for (iter = m.begin(); iter != m.end(); ++iter) {
			//cerr << "ctr = " << ctr << ", color = " << iter->color << ", score = " << iter->score << endl;
			if (iter->score == iter->score && ctr > lastValidEvent) {
				lastValidEvent = ctr;
			}
			/*if ((*iter).boundary) {
				cerr << ", src1 = " << (*iter).src1
					<< ", trg1 = " << (*iter).trg1
					<< ", src2 = " << (*iter).src2
					<< ", trg2 = " << (*iter).trg2 << endl;
			} else {
				cerr << ", src1 = " << (*iter).src1 << ", trg1 = " << (*iter).trg1 << endl;
			}*/
			ctr++;
		}
		if (lastValidEvent == -1) {
			cerr << "Warning: sampling unsuccessful; no valid events were found. log(random #) = " << randDouble
				 << ", total log-probability mass = " << cumProb << ". Using event 0." << endl;
			lastValidEvent = 0;
		}
		sampledEvent = lastValidEvent;
	}
	return sampledEvent;
}

void GapLM::printSegmentedCorpus(string& filename) {
	GENERIC_OFSTREAM outfile(filename.c_str());
	if (!outfile.good()) {
		cerr << "error opening output file " << filename << endl;
		exit(1);
	}
	if (USING_WSTRINGS) {
    	outfile.imbue(locale("en_US.utf8"));
	}

	// Go through all the sentences.
	for (unsigned int i = 0; i < m_data.size(); i++) {
		// Go through all words in the current sentence.
		for (int j = 0; j < m_data[i].wordsToColors.size(); ++j) {
			outfile << m_data[i].words[j] << "~" << m_data[i].wordsToColors[j] << " ";
		}
		outfile << endl;
	}
	outfile.close();
}

// Returns the log of the pdf at x of a Poisson with parameter lambda,
// and where logLambda = log(lambda).
// Computed in the log domain to avoid overflow.
double GapLM::logPoisson(double logLambda, double lambda, int x) {
	return ( (logLambda * ((double) x) ) - lambda) - logFactorial(x);
}

// Returns the log of the pdf at x of a negative binomial with parameters
// m_negBinR and m_negBinP. Computed in the log domain to avoid overflow.
//double GapLM::logNegBin(int x) {
//	return (m_negBinR * m_logNegBinP + ((double) x) * m_logOneMinusNegBinP +
//		logFactorial(((int) m_negBinR) + x - 1)) - (logFactorial(x) + logFactorial(((int) m_negBinR) - 1));
//}

void GapLM::accumulateCounts(bool discrim, bool bilingual, bool bilingnull) {
	// Add current counts to the cumulative counts hashmap.
	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator iter;
	for (iter = m_counts.begin(); iter != m_counts.end(); iter++) {
		double currCount = (double) (iter->second);
		m_cumCounts[iter->first] += currCount;
		m_cumSumCounts += currCount;
	}
	if (discrim || bilingual) {
		// Do the same thing for the second distribution.
		for (iter = m_counts2.begin(); iter != m_counts2.end(); iter++) {
			double currCount = (double) (iter->second);
			m_cumCounts2[iter->first] += currCount;
			m_cumSumCounts2 += currCount;
		}
	}
	// Get gap widths.
	bool includeWidthsTrue = true;
	if (!discrim && bilingual && !bilingnull) {
		countWordsBiling(includeWidthsTrue);
	} else if (!discrim && !bilingual && bilingnull) {
		countWordsBilingNull(includeWidthsTrue);
	} else if (discrim && !bilingual && !bilingnull) {
		countWordsMonolingDiscrim(includeWidthsTrue);
	} else {
		countWordsMonoling(includeWidthsTrue);
	}
}

void GapLM::divideCounts(bool discrim, int numIters) {
	double dNumIters = (double) numIters;
	// Divide counts in the cumulative counts hashmap by the number of iterations.
	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator iter;
	for (iter = m_cumCounts.begin(); iter != m_cumCounts.end(); iter++) {
		//m_cumCounts[iter->first] /= dNumIters;
		iter->second /= dNumIters;
	}

	m_cumSumCounts /= dNumIters;

	// Divide counts in the cumulative counts with widths hashmap by the number of iterations.
	for (iter = m_cumCountsWidths.begin(); iter != m_cumCountsWidths.end(); iter++) {
		iter->second /= dNumIters;
	}

	if (discrim) {
		// Divide counts in the cumulative counts hashmap by the number of iterations.
		for (iter = m_cumCounts2.begin(); iter != m_cumCounts2.end(); iter++) {
			//m_cumCounts[iter->first] /= dNumIters;
			iter->second /= dNumIters;
		}
		m_cumSumCounts2 /= dNumIters;
		// Divide counts in the cumulative counts with widths hashmap by the number of iterations.
		for (iter = m_cumCountsWidths2.begin(); iter != m_cumCountsWidths2.end(); iter++) {
			iter->second /= dNumIters;
		}
	}
}

/************************** Testing Functions *****************************/
// Simple function to test the multinomial sampling code.
void GapLM::testMult() {
	vector<MultinomialEvent> samplingChoices;
	for (int ii = 0; ii < 5; ii++) {
		MultinomialEvent me;
		double randDouble = log(((double) rand()) / ((double) RAND_MAX));
		me.score = randDouble - ((double) ii);
		cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
		samplingChoices.push_back(me);
	}
	int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
	cerr << "sampledEvent = " << sampledEvent << endl << endl;
}

