/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
/**
 * Implements training for nonparametric Bayesian gappy pattern models.
 *
 * @author Kevin Gimpel
 * @date 7/19/2010
 */
#ifndef GAPLM_H_
#define GAPLM_H_

#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <stack>
#include <cstdlib>
#include <map>
#include <math.h>
#include <time.h>
#include <queue>
#include "dense_vocabulary.h"
#include "../external/include/google/sparse_hash_map"
#include "../external/include/google/dense_hash_map"
#include "../external/include/google/dense_hash_set"

using google::sparse_hash_map;
using google::dense_hash_map;
using google::dense_hash_set;

using namespace std;

typedef unsigned int WORD_ID;

// declarations for handling ordinary STL strings
#define USING_WSTRINGS 0
typedef string GENERIC_STRING;
typedef char GENERIC_CHAR;
typedef ifstream GENERIC_IFSTREAM;
typedef ofstream GENERIC_OFSTREAM;
#define DELIMITER_STRING "~~"
#define BOUNDARY_STRING "|"
#define BOUNDARY_SPACE_STRING "| "
#define SPACE_BOUNDARY_STRING " |"
#define LANG_BOUNDARY_STRING " # "
#define LANG_BOUNDARY_SPACE_STRING "# "
#define SPACE_LANG_BOUNDARY_STRING " #"
#define SPACE_STRING " "
#define EMPTY_STRING ""
#define EPSILON_STRING "EPSILON"
#define BILINGUAL_PADDING_STRING "~"
#define UNALIGNED_PREFIX_STRING "~"
#define HASHMAP_NULL_STRING "~~~"
#define HASHMAP_DELETED_STRING "~~~~"
#define OPEN_PARENS_STRING "("
#define CLOSE_PARENS_STRING ")"
#define DUMMY_STRING "DUMMYDUMMY"
#define DUMMY_PREFIX_STRING "DUMMY "
#define BEGINNING_OF_SENTENCE_STRING "<<BOS>>"
#define END_OF_SENTENCE_STRING "<<EOS>>"
#define CERR cerr

// declarations for handling wide-character strings
/*#define USING_WSTRINGS 1
typedef wstring GENERIC_STRING;
typedef wchar_t GENERIC_CHAR;
typedef wifstream GENERIC_IFSTREAM;
typedef wofstream GENERIC_OFSTREAM;
#define DELIMITER_STRING L"~~"
#define BOUNDARY_STRING L"|"
#define SPACE_STRING L" "
#define EMPTY_STRING L""
#define HASHMAP_NULL_STRING L"~~"
#define OPEN_PARENS_STRING L"("
#define CLOSE_PARENS_STRING L")"
#define DUMMY_STRING L"DUMMYDUMMY"
#define DUMMY_PREFIX_STRING L"DUMMY "
#define BEGINNING_OF_SENTENCE_STRING L"<<BOS>>"
#define END_OF_SENTENCE_STRING L"<<EOS>>"
#define CERR wcerr
*/

struct Sentence {
	GENERIC_STRING seq;				// The actual sentence.
	vector<GENERIC_STRING> words;	// Tokens in the sentence.
	vector<WORD_ID> wordIDs;	// Word IDs of the tokens in the sentence.
	int numWords;				// Number of words in the sentence.
	int numColors;				// Number of colors (i.e., patterns) in the sentence.
	stack<int> freeColors;		// Stack of free color indices.
	// Map colors to vectors of word indices that are currently that color.
	// Note: could use a vector instead.
	dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint> colorsToWords;

	// Map word indices to colors.
	vector<int> wordsToColors;

	// The following variables are only used for the monoling-discrim model:

	// Map colors to their mixture component indices.
	map<int, int> colorsToDists;

	// This is the score of a baseline MT system on matching the translation
	// of a source sentence with this reference sentence, e.g., the smoothed
	// sentence-level BLEU score.
	double score;
};

// An aligned sentence is a target-language sentence along with information about
// which of its words are aligned to words in a source-language translation.
// We don't need the actual word alignments, so we throw them out when reading
// in the sentences and only retain lists of words that aligned and unaligned.
struct AlignedSentence {
	GENERIC_STRING seq;
	vector<GENERIC_STRING> words;
	vector<WORD_ID> wordIDs;
	int numWords;

	// Holds all indices of aligned words in "words" and "wordIDs".
	vector<int> alignedWords;
	// Holds all indices of words in "words" and "wordIDs" that this aligned word owns, including itself.
	// Indices of "alignedWords" correspond to indices of "alignedWordChildren".
	vector< set<int> > alignedWordChildren;
	// Holds all indices of unaligned words in "words" and "wordIDs".
	vector<int> unalignedWords;
	// For each unaligned word index in "unalignedWords", holds the index of the aligned
	// word in "alignedWords" (NOT in "words"/"wordIDs") that owns the unaligned word. Each unaligned word
	// is owned by exactly one aligned word. Multiple unaligned words can have the same owner.
	vector<int> unalignedWordOwners;

	// If the aligned words in the sentence are all solo words, then this sentence cannot be
	// generated by the model, so we simply throw it out.
	bool onlySoloWords;
};

struct WordIndexSets {
	set<int> sets[2];
};

struct SentencePair {
	GENERIC_STRING seq[2];
	vector<GENERIC_STRING> words[2];
	vector<WORD_ID> wordIDs[2];
	int numWords[2];
	int numColors[2];
	stack<int> freeColors[2];
	// Map colors to word indices that are currently that color.
	// The first set in the pair is for sentence 1 and the second set is for sentence 2.
	// Note: could use a vector instead.
	//dense_hash_map<unsigned int, pair< set<int>, set<int> >, MurmurHash2UnsignedInt, equnsignedint> colorsToWords;
	//dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint> srcColorsToWords;

	// colorsToWords[0] maps source colors to source word indices that are currently that color.
	// colorsToWords[1] maps target colors to target word indices that are currently that color.
	// Note: when building the string corresponding to a pattern, we have to check srcWordSrcColor to see
	// whether the source word aligned to each target word is colored with a source or target color.
	// If so, these "propagated" source words should be included in the string following the target
	// words they are linked to.
	dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint> colorsToWords[2];

	// Map word indices to colors, for sentence 1 and 2, respectively.
	vector<int> wordsToColors[2];

	// For each word in the source sentence, tracks whether the color recorded is a source color
	// (true) or a target color (false).
	vector<bool> srcWordSrcColor;
	// Maps target colors to sets of source word indexes corresponding to source words colored
	// with that color.
	map< int, set<int> > tgtColorSrcWordIndexSets;

	// For each source word, holds set of indexes of target words that the source word is linked to.
	vector< set<int> > srcWordLinks;
	// For each target word, holds set of indexes of source words that the target word is linked to.
	vector< set<int> > tgtWordLinks;
};

struct MultinomialEvent {
	// Cached strings holding the active string item before and after sampling this event.
	// These are used for incrementing/decrementing the appropriate counts during sampling.
	GENERIC_STRING phraseWithCurr, phraseWithoutCurr;
	GENERIC_STRING srcPhraseWithCurr, srcPhraseWithoutCurr;
	// Placeholder for any index of interest attached to this event.
	// For monoling and biling models, represents the color of the pattern.
	// For biling-null model, represents the owner of the word being sampled.
	int color;
	bool srcColor;
	set<int> dependentSrcWords;
	int dependentSrcWord;
	int dependentSrcWordColor;
	int oldDistIndex, distIndex;
	// Score of the event (in the log domain).
	double score;
	// Sum of all scores: gets filled in and used by normalizeAndSampleMultinomial().
	double cumScore;
};

struct Color {
	GENERIC_STRING phrase;
	double count;
	double logProbability;
};

struct CPTEvent {
	GENERIC_STRING phrase;
	double condProb;
	double rawCount;
};

// The following struct can be used with sort() to sort Colors in
// decreasing order by count.
struct CompareColors {
	bool operator() (Color a, Color b) {
		return a.count > b.count;
	}
};

// The following struct can be used with sort() to sort Colors in
// alphabetical order.
struct CompareColorsAlphabetical {
	bool operator() (Color a, Color b) {
		return a.phrase < b.phrase;
	}
};

class GapLM {
	public:
		GapLM(const string& in_filename, const string& count_out_filename, const string& score_filename, const string& solo_words_filename,
			const string& srclang_in_filename, const string& srctgt_align_filename, const string& srclang_solo_words_filename,
			bool discrim, bool bigram, bool bilingual, bool onlyNull, double alpha, double alpha2, double lambdaW, double lambdaS, 
			double initNewColor, double initPropagate, int burnin, bool useGapWidths, int maxGapWidth);
		~GapLM();
		void setInitialParameterValue(double newInitialParameterValue);
		void trainMonoling(int numIters);
		void trainMonolingDiscrim(int numIters, const string& count_out_filename2);
		void trainBiling(int numIters, const string& count_out_filename2);
		void trainBilingNull(int numIters);
		void decodeFile(string& decode_filename, string& output_decode_filename, bool bigram);

		void printSegmentedCorpus(string& filename);
	private:
		/********************* In gaplm.cpp: ************************/
		void initCommon();
		void testMult();

		void accumulateCounts(bool discrim, bool bilingual, bool bilingnull);
		void divideCounts(bool discrim, int numIters);

		void computeUnigramProbabilities();
		bool openFilesSuccess(bool discrim, bool bilingual);
		void readData(bool discrim, bool bigram, bool bilingual, bool bilingnull);
		void initializeExpectedTableCounts();

		int normalizeAndSampleMultinomial(vector<MultinomialEvent>& m);

		// Returns the log of the pdf at x of a Poisson with parameter lambda,
		// and where logLambda = log(lambda).
		// Computed in the log domain to avoid overflow.
		double logPoisson(double logLambda, double lambda, int x);
		// Returns the log of the pdf at x of a negative binomial with parameters
		// m_negBinR and m_negBinP. Computed in the log domain to avoid overflow.
		double logNegBin(int x);

		double logBaseP1(GENERIC_STRING& token);
		double logBaseP1UsingCumCounts(GENERIC_STRING& token, bool bilingual);

		double logBaseDistribution(vector<GENERIC_STRING>& fragmentTokens, int numGaps);
		double logBaseDistributionUniform(GENERIC_STRING& token);
		double baseDistributionUniform(GENERIC_STRING& token);

		// The following function is inline and therefore declared below.
		double computeGapPortionOfBaseProbability(int numTokens, int numGaps);

		/********************* In gaplm_monoling.cpp: ************************/
		bool compareColors(Color a, Color b);
		bool compareColorsAlphabetical(Color a, Color b);
		Sentence readSentenceMonoling(GENERIC_STRING& str1, int lineCtr);
		Sentence readSentenceOld(GENERIC_STRING& str1, int lineCtr);
		// Goes through saved Sentence objects and counts instances of patterns (i.e., updates m_counts map).
		void countWordsMonoling(bool includeWidths);
		double appendWordsToString(Sentence& s, const set<int>& wordIndices, bool computeBaseProb,
			double uniformWordDist, GENERIC_STRING* result);

		/******************* In gaplm_monolingdisc.cpp: **********************/
		Sentence readSentenceMonolingDiscrim(GENERIC_STRING& str1, int lineCtr, double sentenceScore);
		void countWordsMonolingDiscrim(bool includeWidths);

		/********************* In gaplm_biling.cpp: ************************/
		void insertSourceWordIndexIntoMapWithTargetColor(map< int, set<int> >& tgtColorSrcWordIndexSets, int srcWordIndex, int tgtColor);
		SentencePair readSentenceBiling(GENERIC_STRING& str1, GENERIC_STRING& str2, GENERIC_STRING& alignmentString, int lineCtr);
		void countWordsBiling(bool includeWidths);
		double appendWordsToString(SentencePair& sp, const set<int>& wordIndices, bool tgtSentence,
			int tgtColor, bool computeBaseProb, double srcUniformWordDist, double tgtUniformWordDist, GENERIC_STRING* result);

		void buildMultinomialEventTargetNewColor(SentencePair& sp, MultinomialEvent& me, int currWord,
			int newColor, int dependentSrcWordIndex, double srcUniformBaseProb, double tgtUniformBaseProb);

		void buildMultinomialEventSourceColor(SentencePair& sp, MultinomialEvent& me, int currWord,
			int color, set<int>& wordIndices, double uniformBaseProb);

		void buildMultinomialEventTargetColorWithSourceDepWord(SentencePair& sp, MultinomialEvent& me,
			int currWord, int color, set<int>& wordIndices, int dependentSrcWordIndex,
			double srcUniformBaseProb, double tgtUniformBaseProb);

		void buildAllSourceColorMultinomialEvents(SentencePair& sp, vector<MultinomialEvent>& samplingChoices,
			int currWord, int newColor, double uniformBaseProb);

		/********************* In gaplm_bilingnull.cpp: ************************/
		AlignedSentence readSentenceBilingNull(GENERIC_STRING& str1, GENERIC_STRING& str2,
			GENERIC_STRING& alignmentString, int lineCtr);
		void countWordsBilingNull(bool includeWidths);
		double appendWordsToString(AlignedSentence& as, int alignedWord, const set<int>& wordIndices,
			bool computeBaseProb, double uniformWordDist, GENERIC_STRING* result);


		/********************* Member variables: ************************/
		CompareColors m_compareColorsObject;
		CompareColorsAlphabetical m_compareColorsAlphabeticalObject;

		bool m_logging;

		// Number of iterations of sampling to discard as burn-in.
		int m_burnin;

		// The training data, in various forms for the different models.
		vector<Sentence> m_data;				// model = monoling
		vector<SentencePair> m_bilingualData;	// model = biling
		vector<AlignedSentence> m_alignedData;	// model = biling-null

		// The vocabulary mapping all words to integers.
		DenseVocabulary m_vocabulary;

		/**
		 * Maps strings to integer counts.
		 * TODO_efficiency: convert this to a set<WORD_ID> -> int map and
		 * use a separate DenseVocabulary to represent tokens as integers.
		 */
		// using Google's dense hash map and the MurmurHash2 function:
		dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring> m_counts;
		sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_cumCounts;
		sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_cumCountsWidths;

		dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring> m_counts2;
		sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_cumCounts2;
		sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_cumCountsWidths2;

		double m_cumSumCounts;
		double m_cumSumCounts2;

		// Unigram word probabilities, possibly used for base distribution.
		dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_unigramProbabilities;
		dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> m_srclangUnigramProbabilities;

		// Sets of words to keep by themselves in their own (non-gappy) patterns, for source and target languages.
		dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring> m_srclangSoloWords;
		dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring> m_soloWords;

		/**
		 * Any count that involves a target-language word is stored in the
		 * targetCounts hashmap. These include target word counts, source-target
		 * word link counts, and counts of source bigrams with the target word linked
		 * to the second token in the source bigram. They are used only by bilingual models.
		 */
		//dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring> targetCounts;
		//dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> cumTargetCounts;

		//dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> tableCounts;
		//dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> cumTableCounts;
		//dense_hash_map<GENERIC_STRING, set<GENERIC_STRING>, MurmurHash2, eqstring> bigramPartners;

		//dense_hash_map<GENERIC_STRING, set<GENERIC_STRING>, MurmurHash2, eqstring> bilingualBigramPartners;
		//dense_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> bilingualCumTableCounts;
		//dense_hash_map<string, dense_hash_set<string, MurmurHash2, eqstring>, MurmurHash2, eqstring> bigramPartners;

		int m_numWords;		// The total number of tokens.
		int m_numColors;
		int m_numColors2;
		int m_numTables;	// The total number of tables in all restaurants.

		double m_pInitNewColor; // Probability of creating a new color during initialization.
		double m_pInitPropagate; // Probability of propagating a target sentence color to a source
								 // sentence word across a word link.
		double m_logPInitPropagate;
		double m_logOneMinusPInitPropagate;

		double m_alpha;		// Concentration parameter for Dirichlet process prior for multinomial over tokens.
		double m_logalpha;	// log(alpha), precomputed for efficiency.
		double m_alpha2;	// Concentration parameter for Dirichlet process prior for multinomial over tokens.
		double m_logalpha2;	// log(alpha), precomputed for efficiency.

		// The concentration parameter for each Dirichlet process prior for each conditional multinomial
		// over tokens given a previous token.
		double m_alpha1;
		double m_logalpha1;		// log(alpha1), precomputed for efficiency.
		double m_logalpha1PlusOne;

		// The mean of the Poisson over the number of words in each pattern.
		double m_lambdaW;
		// log(lambdaW), precomputed for efficiency.
		double m_loglambdaW;
		// The mean of the Poisson over the number of source words for a target word in a bilingual pattern.
		double m_lambdaS;
		// log(lambdaS), precomputed for efficiency.
		double m_loglambdaS;

		// Precomputed Poisson PDFs.
		vector<double> m_logPoissonWPDF;
		vector<double> m_logPoissonSPDF;

		bool m_useGapWidths;
		int m_maxGapWidth;

		// Precomputed table with values of log( 1 / (n choose k)), for various n and k values.
		double m_precomputedLogOneOverNChooseK[30][30];

		double m_timeProfile;

		// initial value for all parameters
		double m_initialParameterValue;

		string m_inputFilename;
		string m_scoreFilename;
		string m_srclangInputFilename;
		string m_srctgtAlignFilename;
		string m_parallelInputFilename;
		string m_alignmentInitFilename;
		string m_countOutputFilename;
		string m_countWidthOutputFilename;
		string m_soloWordsFilename;
		string m_srclangSoloWordsFilename;

		// file handles for input files (read each time a Formula object for a training example is needed)
		GENERIC_IFSTREAM input_file;
		GENERIC_IFSTREAM score_file;
		GENERIC_IFSTREAM srclang_input_file;
		GENERIC_IFSTREAM srctgt_align_file;
		GENERIC_IFSTREAM parallel_input_file;
		ifstream alignment_init_file;
		GENERIC_OFSTREAM output_file;
};

// For the portion of the base probability of a gappy pattern that involves gaps,
// we have a uniform distribution over the number of gaps given the number of words
// and a uniform distribution over the possible arrangements of the gaps.
// Precondition: numTokens should never be less than 1.
inline double GapLM::computeGapPortionOfBaseProbability(int numTokens, int numGaps) {
	double ret = 0.0;

	// Log-multiply in the probability of the number of gaps given the number of words,
	// for which we use a uniform distribution over the possible numbers of gaps.
	// The largest possible number of gaps is the number of tokens minus 1.
	// There can also be zero gaps, so there are a total of numTokens possible numbers
	// of gaps.
	// Note: If there is only one token, the number of gaps must be zero and the probability of
	// zero gaps given one token is 1, so we simply avoid computing this probability in
	// this case as a code optimization.
	if (numTokens > 1) ret -= log((double)numTokens);
	// Note: we originally used a Poisson for the number of gaps, but this allows patterns
	// with more gaps than tokens.


	// Note: the fact that we only enter the following block if numGaps > 0 is merely
	// a code optimization, since "n choose k" for k == 0 is always 1. I.e., the code
	// below should work fine in this case, but it is not necessary.
	if (numGaps > 0) {
		// We use a uniform distribution over all arrangements of the gaps.
		// Where there are n tokens and k gaps, there are "(n-1) choose k" arrangements
		// of gaps. We cache log(1.0 / nchoosek) for values of n and k.

		// Log-multiply the log-probability of this particular arrangement of gaps.
		// Since this is uniform, we only care about the number of gaps (numGaps)
		// and do not look at the actual positions of the gaps.
		//if (ret != ret) cerr << "Warning: gap portion of base prob NaN!" << endl;
		ret += m_precomputedLogOneOverNChooseK[numTokens - 1][numGaps];
		if (ret != ret) {
			cerr << "Warning: after including arrangement of gaps, gap portion of base prob NaN! numTokens = "
				 << numTokens << ", numGaps = " << numGaps << endl;
		}
	}
	return ret;
}

#endif /*GAPLM_H_*/
