/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// This file contains code that implements the bilingual gappy pattern model. 
// The comments at the top of gaplm_monoling.cpp also apply here. --Kevin Gimpel, 7/18/2011

#include "gaplm.h"
#include "utils.h"

// Inserts given source word index into the set paired with the given target color in the provided map.
// The map is modified by this function.
void GapLM::insertSourceWordIndexIntoMapWithTargetColor(map< int, set<int> >& tgtColorSrcWordIndexSets, int srcWordIndex, int tgtColor) {
	map< int, set<int> >::iterator iter = tgtColorSrcWordIndexSets.find(tgtColor);
	if (iter == tgtColorSrcWordIndexSets.end()) {
		// Create a set for the target color with the given source word index.
		set<int> theSet;
		theSet.insert(srcWordIndex);
		tgtColorSrcWordIndexSets.insert(make_pair(tgtColor, theSet));
	} else {
		// Insert the source word index into the existing set for the target color.
		iter->second.insert(srcWordIndex);
	}
}

// Converts a line of input into a Sentence instance, which is returned.
// Assumes tokens are delimited by spaces.
SentencePair GapLM::readSentenceBiling(GENERIC_STRING& str1, GENERIC_STRING& str2, GENERIC_STRING& alignmentString, int lineCtr) {
	SentencePair s;
	s.seq[0] = str1;
	s.seq[1] = str2;

	// Tokenize the sentence.
	s.words[0].clear();
	s.words[1].clear();
	tokenize(str1, ' ', s.words[0]);
	tokenize(str2, ' ', s.words[1]);
	s.numWords[0] = s.words[0].size();
	s.numWords[1] = s.words[1].size();
	s.numColors[0] = 0;
	s.numColors[1] = 0;

	s.colorsToWords[0].set_empty_key(9999999);
	s.colorsToWords[0].set_deleted_key(8888888);
	s.colorsToWords[1].set_empty_key(9999999);
	s.colorsToWords[1].set_deleted_key(8888888);

	// Initialize the vectors of word alignments.
	s.srcWordLinks.resize(s.numWords[0]);
	s.tgtWordLinks.resize(s.numWords[1]);

	// Tokenize the string of word alignments.
	vector<string> alignments;
	tokenize(alignmentString, ' ', alignments);
	set<int> alignedWordSet;
	// Go through the word links.
	for (int i = 0; i < alignments.size(); ++i) {
		// Get the target word in the word link.
		vector<string> wordIndices;
		tokenize(alignments[i], '-', wordIndices);

		int srcWordIndex = atoi(wordIndices[0].c_str());
		int tgtWordIndex = atoi(wordIndices[1].c_str());
		if (srcWordIndex >= s.numWords[0] || tgtWordIndex >= s.numWords[1]) {
			cerr << "Error: during bilingual model initialization, on line " << lineCtr << ", invalid alignment (" << alignments[i]
				 << ") for src-tgt sentence pair with " << s.numWords[0] << " and " << s.numWords[1] << " words." << endl;
		}
		s.srcWordLinks[srcWordIndex].insert(tgtWordIndex);
		s.tgtWordLinks[tgtWordIndex].insert(srcWordIndex);
	}

	dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;

	set<int> tgtSoloWordColors;
	set<int>::iterator soloColorIter;

	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	////////////////////////////////////////////////////////////////
	// Initialize target sentence color assignments first. Source
	// sentence color assignments will be done below.

	// Go through all target words.
	for (int i = 0; i < s.numWords[1]; ++i) {
		bool soloWord = false;
		soloWordsIter = m_soloWords.find(s.words[1][i]);
		if (soloWordsIter != m_soloWords.end()) soloWord = true;

		if (soloWord) {
			// The new color will be s.numColors for the target sentence.
			int newColor = s.numColors[1];
			// Increment the number of colors for this sentence.
			++(s.numColors[1]);
			// Add the color assignment for this word.
			s.wordsToColors[1].push_back(newColor);
			// Add this word to the list of word indices for the new color.
			set<int> wordIndices;
			wordIndices.insert(i);
			s.colorsToWords[1].insert(make_pair(newColor, wordIndices));
			// Add this new color to the hashset of solo word colors.
			tgtSoloWordColors.insert(newColor);
		} else {
			// The probability of creating a new color for this word is m_pInitNewColor.
			// The remaining probability mass is divided among the colors used already in this sentence.

			bool createNewColor = true;
			// We MUST create a new color if we are on the first word or if all previous colors
			// were assigned to solo words.
			// So, if we are past the first word and there are non-solo-word colors, see whether
			// or not we should create a new color.
			if (i > 0 && s.numColors[1] > tgtSoloWordColors.size()) {
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				if (randDouble > m_pInitNewColor) {
					createNewColor = false;
				}
			}

			if (createNewColor) {
				// The new color will be s.numColors.
				int newColor = s.numColors[1];
				// Increment the number of colors for this sentence.
				++(s.numColors[1]);
				// Add the color assignment for this word.
				s.wordsToColors[1].push_back(newColor);
				// Add this word to the list of word indices for the new color.
				set<int> wordIndices;
				wordIndices.insert(i);
				s.colorsToWords[1].insert(make_pair(newColor, wordIndices));
			} else {
				// If we're not creating a new color, choose uniformly from among the colors
				// that already exist for this sentence, except for any solo word colors.
				int numPossibleColors = s.numColors[1] - tgtSoloWordColors.size();
				vector<MultinomialEvent> samplingChoices;
				double eventLogProb = -log(numPossibleColors);

				for (int ii = 0; ii < s.numColors[1]; ++ii) {
					// Look for the current color in the solo colors set.
					soloColorIter = tgtSoloWordColors.find(ii);
					// If the current color is not in the solo colors set, add it to be
					// considered as a multinomial event.
					if (soloColorIter == tgtSoloWordColors.end()) {
						MultinomialEvent me;
						me.score = eventLogProb;
						me.color = ii;
						//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
						samplingChoices.push_back(me);
					}
				}
				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				//cerr << "sampledEvent = " << sampledEvent << ", sampledColor = " << sampledColor << endl;

				// Add the color assignment for this word.
				s.wordsToColors[1].push_back(sampledColor);

				// Search the color map for the color we just sampled.
				iter = s.colorsToWords[1].find(sampledColor);
				// If it already exists (should always be the case! -- make this more efficient),
				if (iter != s.colorsToWords[1].end()) {
					// Add the index of the current word to the list of words for the current color.
					iter->second.insert(i);
				} else {
					cerr << "Error: sampled color (" << sampledColor << ") not found in hash map!" << endl;
					// Create a new vector and add it to the hash map.
					set<int> wordIndices;
					wordIndices.insert(i);
					s.colorsToWords[1].insert(make_pair(sampledColor, wordIndices));
				}
			}
			//cerr << s.words[i] << "~" << prevColor << " ";
		}
		// Increment the count of this word for the unigram probabilities.
		m_unigramProbabilities[s.words[1][i]] += 1.0;
		//cerr << tokens[i] << "\t" << m_unigramProbabilities[tokens[i]] << endl;
	}
	//cerr << endl;

	// We want to add colors to the freeColors stack for each word in the target sentence.
	// First, figure out how many more target colors there can possibly be.
	if (s.numColors[1] != s.colorsToWords[1].size()) {
		cerr << "Warning: s.colorsToWords[1].size() = " << s.colorsToWords[1].size() << ", s.numColors[1] = " << s.numColors[1] << endl;
	}
	// TODO: we only use <= in the following line instead of < because we want to add one extra
	// color in case anything bad happens. But the algorithm should only need s.numWords colors,
	// so the <= should be able to be changed to a <.
	for (int i = s.numColors[1]; i <= s.numWords[1]; ++i) {
		s.freeColors[1].push(i);
	}

	if (s.wordsToColors[1].size() != s.numWords[1]) {
		cerr << "Warning: s.wordsToColors[1].size() (" << s.wordsToColors[1].size()
			 << ") != s.numWords[1] (" << s.numWords[1] << ")" << endl;
	}

	///////////////////////////////////////////////////////////////////////////////////
	// Now initialize colors for source sentence.
	// To do this, we go through all source words and first decide whether to propagate
	// the target color to the source word. We do this with probability m_pInitPropagate.
	// If the source word is unaligned, there is no target color to propagate.
	// If the source word is a solo word, then don't propagate it; force it to be by itself as a source color.

	s.srcWordSrcColor.reserve(s.numWords[0]);
	set<int> srcSoloWordColors;

	// Go through all source words.
	for (int i = 0; i < s.numWords[0]; ++i) {
		bool soloWord = false;
		soloWordsIter = m_srclangSoloWords.find(s.words[0][i]);
		if (soloWordsIter != m_srclangSoloWords.end()) soloWord = true;

		if (soloWord) {
			// The new color will be s.numColors for the source sentence.
			int newColor = s.numColors[0];
			// Increment the number of colors for this sentence.
			++(s.numColors[0]);
			// Add the color assignment for this word.
			s.wordsToColors[0].push_back(newColor);
			// Add this word to the list of word indices for the new color.
			set<int> wordIndices;
			wordIndices.insert(i);
			s.colorsToWords[0].insert(make_pair(newColor, wordIndices));
			// Add this new color to the hashset of solo word colors.
			srcSoloWordColors.insert(newColor);
			// The current word has a source color, so push "true" onto the back of the s.srcWordSrcColor vector.
			s.srcWordSrcColor.push_back(true);
		} else {
			// If the word is not a solo word, we first have to decide whether it should
			// propagate the target color of a target word it is linked to.
			bool propagateTargetColor = false;

			// If the current source word has any word links to target words,
			if (s.srcWordLinks[i].size() > 0) {
				// Decide whether to propagate.
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				if (randDouble < m_pInitPropagate) {
					propagateTargetColor = true;
				}
			}

			// If we should propagate a target color, sample uniformly at random from the target
			// colors among target words that are linked to this source word.
			if (propagateTargetColor) {
				//
				vector<MultinomialEvent> targetWords;
				set<int>::const_iterator tgtWordIter;
				for (tgtWordIter = s.srcWordLinks[i].begin(); tgtWordIter != s.srcWordLinks[i].end(); ++tgtWordIter) {
					// Get the color of the current target word.
					int currTgtWordColor = s.wordsToColors[1][*tgtWordIter];
					// Check if the color was identified as a target solo color.
					soloColorIter = tgtSoloWordColors.find(currTgtWordColor);
					// If the current color is not in the target solo colors set, add it to be
					// considered as a multinomial event.
					if (soloColorIter == tgtSoloWordColors.end()) {
						MultinomialEvent me;
						me.score = 0.0;	// This is in log-domain. Scores will be normalized when sampling is performed.
						me.color = currTgtWordColor;	// The event is the target color.
						//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
						targetWords.push_back(me);
					}
				}
				if (targetWords.size() == 0) {
					// If all target words were solo words, then this source word can not propagate any
					// target colors, so set propagateTargetColor to false to enter the block below this one.
					propagateTargetColor = false;
				} else {
					int sampledEvent = normalizeAndSampleMultinomial(targetWords);
					int sampledColor = targetWords[sampledEvent].color;
					// Since we chose a target color, insert this source word index in the map.
					insertSourceWordIndexIntoMapWithTargetColor(s.tgtColorSrcWordIndexSets, i, sampledColor);
					// Add the color assignment for this source word.
					s.wordsToColors[0].push_back(sampledColor);
					// The current word has a target color, so push "false" onto the back of the s.srcWordSrcColor vector.
					s.srcWordSrcColor.push_back(false);
					// Note: we don't have to insert anything into s.colorsToWords[0], because that
					// is only for source colors, while here we have used a target color for a source word.
				}
			}

			// We use a separate conditional here because we may have set propagateTargetColor to false
			// in the block above if we wanted to propagate but all linked target words were solo words.
			if (!propagateTargetColor) {
				// Here, we are choosing from among existing source colors or may generate a new one.

				// The probability of creating a new color for this word is m_pInitNewColor.
				// The remaining probability mass is divided among the source colors used already in the source
				// sentence.

				bool createNewColor = true;
				// We MUST create a new color if we are on the first word or if all previous colors
				// were assigned to solo words.
				// So, if we are past the first word and there are non-solo-word colors, see whether
				// or not we should create a new color.
				if (i > 0 && s.numColors[0] > srcSoloWordColors.size()) {
					double randDouble = ((double) rand()) / ((double) RAND_MAX);
					if (randDouble > m_pInitNewColor) {
						createNewColor = false;
					}
				}

				if (createNewColor) {
					// The new color will be s.numColors[0].
					int newColor = s.numColors[0];
					// Increment the number of colors for the source sentence.
					++(s.numColors[0]);
					// Add the color assignment for this word.
					s.wordsToColors[0].push_back(newColor);
					// Add this word to the list of word indices for the new color.
					set<int> wordIndices;
					wordIndices.insert(i);
					s.colorsToWords[0].insert(make_pair(newColor, wordIndices));
					// The current word has a source color, so push "true" onto the back of the s.srcWordSrcColor vector.
					s.srcWordSrcColor.push_back(true);

				} else {
					// If we're not creating a new color, choose uniformly from among the colors
					// that already exist for this sentence, except for any solo word colors.
					//int numPossibleColors = s.numColors[0] - srcSoloWordColors.size();
					vector<MultinomialEvent> samplingChoices;
					//double eventLogProb = -log(numPossibleColors);

					for (int ii = 0; ii < s.numColors[0]; ++ii) {
						// Look for the current color in the solo colors set.
						soloColorIter = srcSoloWordColors.find(ii);
						// If the current color is not in the solo colors set, add it to be
						// considered as a multinomial event.
						if (soloColorIter == srcSoloWordColors.end()) {
							MultinomialEvent me;
							me.score = 0.0;//eventLogProb;
							me.color = ii;
							//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
							samplingChoices.push_back(me);
						}
					}
					int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
					int sampledColor = samplingChoices[sampledEvent].color;
					//cerr << "sampledEvent = " << sampledEvent << ", sampledColor = " << sampledColor << endl;

					// Add the color assignment for this word.
					s.wordsToColors[0].push_back(sampledColor);
					// The current word has a source color, so push "true" onto the back of the s.srcWordSrcColor vector.
					s.srcWordSrcColor.push_back(true);

					// Search the color map for the color we just sampled.
					iter = s.colorsToWords[0].find(sampledColor);
					// If it already exists (should always be the case! -- make this more efficient),
					if (iter != s.colorsToWords[0].end()) {
						// Add the index of the current word to the list of words for the current color.
						iter->second.insert(i);
					} else {
						cerr << "Error: sampled color (" << sampledColor << ") not found in hash map!" << endl;
						// Create a new vector and add it to the hash map.
						set<int> wordIndices;
						wordIndices.insert(i);
						s.colorsToWords[0].insert(make_pair(sampledColor, wordIndices));
					}
				}
			}
			//cerr << s.words[i] << "~" << prevColor << " ";
		}
		// Increment the count of this word for the unigram probabilities.
		m_srclangUnigramProbabilities[s.words[0][i]] += 1.0;
		//cerr << tokens[i] << "\t" << m_unigramProbabilities[tokens[i]] << endl;
	}
	//cerr << endl;

	// We want to add colors to the freeColors stack for each word in the source sentence.
	// First, figure out how many more source colors there can possibly be.
	if (s.numColors[0] != s.colorsToWords[0].size()) {
		cerr << "Warning: s.colorsToWords[0].size() = " << s.colorsToWords[0].size() << ", s.numColors[0] = " << s.numColors[0] << endl;
	}
	// TODO: we only use <= in the following line instead of < because we want to add one extra
	// color in case anything bad happens. But the algorithm should only need s.numWords colors,
	// so the <= should be able to be changed to a <.
	for (int i = s.numColors[0]; i <= s.numWords[0]; ++i) {
		s.freeColors[0].push(i);
	}

	if (s.wordsToColors[0].size() != s.numWords[0]) {
		cerr << "Warning: s.wordsToColors[0].size() (" << s.wordsToColors[0].size()
			 << ") != s.numWords[0] (" << s.numWords[0] << ")" << endl;
	}

	// Print out the data structures in the sentence pair to ensure there weren't any bugs.
	/*cerr << endl;
	cerr << "Colors for each source word: " << endl;
	for (int i = 0; i < s.numWords[0]; ++i) {
		cerr << i << "\t" << s.words[0][i] << "\t" << s.wordsToColors[0][i] << "\t" << s.srcWordSrcColor[i] << endl;
	}
	cerr << "Words for each source color: " << endl;
	for (int i = 0; i < s.numColors[0]; ++i) {
		cerr << i << ":" << endl;
		set<int>& wordIndices = s.colorsToWords[0][i];
		set<int>::iterator setIntIter = wordIndices.begin();
		for (; setIntIter != wordIndices.end(); ++setIntIter) {
			cerr << "\t" << *setIntIter << "\t" << s.words[0][*setIntIter] << endl;
		}
	}
	cerr << "Colors for each target word: " << endl;
	for (int i = 0; i < s.numWords[1]; ++i) {
		cerr << i << "\t" << s.words[1][i] << "\t" << s.wordsToColors[1][i] << endl;
	}
	cerr << "Words for each target color: " << endl;
	for (int i = 0; i < s.numColors[1]; ++i) {
		cerr << i << ":" << endl;
		set<int>& wordIndices = s.colorsToWords[1][i];
		set<int>::iterator setIntIter = wordIndices.begin();
		for (; setIntIter != wordIndices.end(); ++setIntIter) {
			cerr << "\t" << *setIntIter << "\t" << s.words[1][*setIntIter]<< endl;
		}
		// Now print source words colored with this target color.
		cerr << "\tsource words:" << endl;
		map< int, set<int> >::iterator mapSetIter = s.tgtColorSrcWordIndexSets.find(i);
		if (mapSetIter != s.tgtColorSrcWordIndexSets.end()) {
			set<int>::iterator iter3 = mapSetIter->second.begin();
			for (; iter3 != mapSetIter->second.end(); ++iter3) {
				cerr << "\t\t" << *iter3 << "\t" << s.words[0][*iter3] << endl;
			}
		}

	}*/
	return s;
}

// Goes through sentences in memory and counts fragments (i.e., updates counts hashmaps)
// given current colors.
void GapLM::countWordsBiling(bool includeWidths) {
	// Clear all counts.
	if (!includeWidths) {
		m_counts.clear();
		m_counts2.clear();
		m_numColors = 0;
		m_numColors2 = 0;
	}

	bool computeBaseProbFalse = false;
	bool tgtSentenceTrue = true;
	bool tgtSentenceFalse = false;

	// Go through all sentence pairs.
	for (unsigned int i = 0; i < m_bilingualData.size(); ++i) {
		SentencePair& s = m_bilingualData[i];
		// Go through all source colors in the sentence pair.

		dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::const_iterator iter;

		for (iter = s.colorsToWords[0].begin(); iter != s.colorsToWords[0].end(); ++iter) {
			if (!includeWidths) ++m_numColors;
			GENERIC_STRING currWordPhrase = EMPTY_STRING;
			const set<int>& wordIndices = iter->second;

			// Build the string containing the fragment corresponding to the current source color
			// using words from the source sentence in the current sentence pair.
			appendWordsToString(s, wordIndices, tgtSentenceFalse, -1, computeBaseProbFalse, 0.0, 0.0, &currWordPhrase);

			if (includeWidths) {
				++m_cumCountsWidths[currWordPhrase];
			} else {
				//cerr << "SRC: Incrementing the count of ||" << currWordPhrase << "||, ";
				++m_counts[currWordPhrase];
				//cerr << "SRC: New count: " << m_counts[currWordPhrase] << endl;
			}
		}

		for (iter = s.colorsToWords[1].begin(); iter != s.colorsToWords[1].end(); ++iter) {
			if (!includeWidths) ++m_numColors2;
			GENERIC_STRING currWordPhrase = EMPTY_STRING;
			const set<int>& wordIndices = iter->second;
			// Build the string containing the fragment corresponding to the current target color
			// using words from the target sentence (and possibly source sentence) in the current
			// sentence pair.
			appendWordsToString(s, wordIndices, tgtSentenceTrue, iter->first, computeBaseProbFalse, 0.0, 0.0, &currWordPhrase);

			if (includeWidths) {
				++m_cumCountsWidths2[currWordPhrase];
			} else {
				//cerr << "TGT: Incrementing the count of ||" << currWordPhrase << "||, ";
				++m_counts2[currWordPhrase];
				//cerr << "TGT: New count: " << m_counts2[currWordPhrase] << endl;
			}
		}
	}

	if (!includeWidths) {
		cerr << "Done counting words in sentence pairs" << endl;
	}
}

double GapLM::appendWordsToString(SentencePair& sp, const set<int>& wordIndices, bool tgtSentence,
		int tgtColor, bool computeBaseProb, double srcUniformWordDist, double tgtUniformWordDist, GENERIC_STRING* result) {

	int lastWordIndex = -1;
	set<int>::const_iterator setIter;
	int numWords = wordIndices.size();
	double baseProb = 0.0;
	if (computeBaseProb) baseProb += m_logPoissonWPDF[numWords];
	int numGaps = 0;
	int ii = 0;

	int sentIndex = 0;
	if (tgtSentence) sentIndex = 1;

	const set<int>* srcWordIndexSet;
	bool foundSrcWordIndexSet = false;
	if (tgtSentence) {
		map< int, set<int> >::const_iterator iter = sp.tgtColorSrcWordIndexSets.find(tgtColor);
		if (iter != sp.tgtColorSrcWordIndexSets.end()) {
			srcWordIndexSet = &(iter->second);
			foundSrcWordIndexSet = true;
		}
	}

	// If this is a target pattern, put the language boundary character at the beginning.
	//if (tgtSentence) result->append(LANG_BOUNDARY_SPACE_STRING);

	for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
		//cerr << "\tB4 " << *setIter << endl;
		// If there is a gap between the current word and the last one, add a "|" as a separator.
		if (ii > 0) {
			int distance = *setIter - lastWordIndex;
			if (distance > 1) {
				if (m_useGapWidths) {
					int gapWidthToPrint = distance - 1;
					if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
					result->append(BOUNDARY_STRING);
					result->append(itos(gapWidthToPrint));
					result->append(SPACE_STRING);
				} else {
					result->append(BOUNDARY_SPACE_STRING);
				}
				++numGaps;
			}
		}
		// Append the current word for this color.
		result->append(sp.words[sentIndex][*setIter]);

		// If we're dealing with the target sentence, we have to include
		// all source words that are linked to this sentence.
		int numSourceWordsAdded = 0;
		if (tgtSentence && foundSrcWordIndexSet) {
			// Get the set of all source words that are linked to the current target word.
			set<int>& srcWordCurrSet = sp.tgtWordLinks[*setIter];
			// Go through each one and check if it's one of the source words colored with
			// the current target color.
			set<int>::const_iterator iter;
			for (iter = srcWordCurrSet.begin(); iter != srcWordCurrSet.end(); ++iter) {
				// Check if the current source word is in the set of source words colored with the target color.
				set<int>::const_iterator iter2 = srcWordIndexSet->find(*iter);
				// If we found it,
				if (iter2 != srcWordIndexSet->end()) {
					// Append it to the string, padded by special characters.
					result->append(BILINGUAL_PADDING_STRING);
					result->append(sp.words[0][*iter]);
					//result->append(BILINGUAL_PADDING_STRING);
					// We added a source word to this pattern, so increment the counter (for computing the base probability).
					if (computeBaseProb) {
						++numSourceWordsAdded;
						baseProb += srcUniformWordDist;
					}
				}
			}
		}
		if (computeBaseProb && tgtSentence) baseProb += m_logPoissonSPDF[numSourceWordsAdded];

		if (computeBaseProb) {
			if (tgtSentence) {
				baseProb += tgtUniformWordDist;
			} else {
				baseProb += srcUniformWordDist;
			}
		}

		// If we're not at the end, append a space.
		if (ii < numWords - 1) {
			result->append(SPACE_STRING);
		}
		// Update the index of the last word added.
		lastWordIndex = *setIter;
		++ii;
	}

	// If this is a source pattern, tack on the language boundary character at the end.
	//if (!tgtSentence) result->append(SPACE_LANG_BOUNDARY_STRING);

	// Log-multiply in the probability of the number of gaps.
	if (computeBaseProb) baseProb += computeGapPortionOfBaseProbability(numWords, numGaps);
	return baseProb;
}

// Creates a multinomial event corresponding to choosing the given new target color (newColor)
// for the given target word index (currWord) and the given dependent source word index.
void GapLM::buildMultinomialEventTargetNewColor(SentencePair& sp, MultinomialEvent& me, int currWord,
		int newColor, int dependentSrcWordIndex, double srcUniformBaseProb, double tgtUniformBaseProb) {

	// Compute the log-probability from the base distribution for tokens.
	double base = m_logPoissonWPDF[1];	// There is just one token.
	// Log-multiply in the unigram target token probability for the single target token.
	base += tgtUniformBaseProb;
	// We don't have to log-multiply in the probability of the number of gaps
	// since there's only one word, there are no gaps, and the probability of
	// 0 gaps given 1 word is 1.0.
	// We also don't need to log-multiply in the log-probability of the
	// arrangement of gaps since it is always one when there aren't any gaps.
	// Therefore, the following line is commented out. If the following function
	// call does not return 0.0, then uncomment the line.
	//base += computeGapPortionOfBaseProbability(1, 0);

	string pattern = sp.words[1][currWord];
	// Go through each one and check if it's one of the source words colored with
	// the current target color.
	int numSourceWordsAdded = 0;
	// Commented code below is for allowing many-to-one links:
	//set<int>::const_iterator iter;
	//for (iter = dependentSrcWordIndices.begin(); iter != dependentSrcWordIndices.end(); ++iter) {
		// Append it to the string, padded by special characters.
	//	pattern += BILINGUAL_PADDING_STRING;
	//	pattern += sp.words[0][*iter];

		// We added a source word to this pattern, so increment the counter (for computing the base probability).
	//	++numSourceWordsAdded;
	//	base += srcUniformBaseProb;
	//}
	if (dependentSrcWordIndex != -1) {
		pattern += BILINGUAL_PADDING_STRING;
		pattern += sp.words[0][dependentSrcWordIndex];

		// We added a source word to this pattern, so increment the counter (for computing the base probability).
		++numSourceWordsAdded;
		base += srcUniformBaseProb;
	}
	base += m_logPoissonSPDF[numSourceWordsAdded];

	double eventLogProb;
	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;
	checkCountsIter = m_counts2.find(sp.words[1][currWord]);
	// If we found an entry for the current word, use the count.
	if (checkCountsIter != m_counts2.end()) {
		eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + base);
	} else {
		// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
		eventLogProb = m_logalpha + base;
	}
	// We have the numerator, so we now need to compute the denominator.
	double denom = logAdd(log((double) m_numColors2), m_logalpha);
	if (denom == neg_inf) {
		cerr << "Error: dividing by neg_inf!" << endl;
	}
	// Subtract the denominator from the numerator.
	eventLogProb -= denom;

	// Build a multinomial event for the choice of a new source color.
	me.phraseWithCurr = pattern;
	me.phraseWithoutCurr = EMPTY_STRING;
	me.score = eventLogProb;
	me.color = newColor;
	me.srcColor = false;
	//me.dependentSrcWords = dependentSrcWordIndices;
	me.dependentSrcWord = dependentSrcWordIndex;
}

void GapLM::buildMultinomialEventSourceColor(SentencePair& sp, MultinomialEvent& me, int currWord,
		int color, set<int>& wordIndices, double uniformBaseProb) {

	// Build the strings containing the fragment corresponding to the current color,
	// both with and without the current word currWord.
	GENERIC_STRING phraseWithoutCurr = EMPTY_STRING;
	GENERIC_STRING phraseWithCurr = EMPTY_STRING;

	// Count the number of gaps both with and without the current word.
	int tokensWithoutCurrNumGaps = 0;
	int tokensWithCurrNumGaps = 0;

	int lastWordIndexWithCurr = -1;
	int ii = 0;
	int lastWordIndexWithoutCurr = -1;
	int jj = 0;

	int numWordsCurrColorWithoutCurr = wordIndices.size();
	// Since we're handling a source color in the source sentence,
	// add the current word index to the set for this color.
	wordIndices.insert(currWord);
	int numWordsCurrColorWithCurr = wordIndices.size();

	if (numWordsCurrColorWithoutCurr == 0) {
		cerr << "Warning: numWordsCurrColorWithoutCurr == 0! currWord = " << currWord << ", color = " << color << endl;
	}
	if (numWordsCurrColorWithCurr == 0) {
		cerr << "Warning: numWordsCurrColorWithCurr == 0! currWord = " << currWord << ", color = " << color << endl;
	}

	// Initialize the base probabilities to the probability of the number of words,
	// for both the color without the current word and with the current word.
	double baseWithoutCurr = m_logPoissonWPDF[numWordsCurrColorWithoutCurr];
	double baseWithCurr = m_logPoissonWPDF[numWordsCurrColorWithCurr];

	bool soloWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	set<int>::const_iterator setIter;
	for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
		// Don't add word currWord to phraseWithoutCurr.
		if (*setIter != currWord) {
			// Add to phraseWithoutCurr.
			// If there is a gap between the current word and the last one, add a
			// space and then "|" as a separator.
			if (jj > 0) {
				int distance = *setIter - lastWordIndexWithoutCurr;
				if (distance > 1) {
					if (m_useGapWidths) {
						int gapWidthToPrint = distance - 1;
						if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
						phraseWithoutCurr += SPACE_STRING;
						phraseWithoutCurr += BOUNDARY_STRING;
						phraseWithoutCurr += itos(gapWidthToPrint);
					} else {
						phraseWithoutCurr += SPACE_BOUNDARY_STRING;
					}
					// Increment the number of gaps in the phrase without the current word.
					++tokensWithoutCurrNumGaps;
				}
			}
			// If we're not at the beginning, insert a space before the current word.
			if (jj > 0) {
				phraseWithoutCurr += SPACE_STRING;
			}
			// Append the current word.
			phraseWithoutCurr += sp.words[0][*setIter];
			// Check if the current word is a solo word.
			soloWord = false;
			if (numWordsCurrColorWithoutCurr > 1) {
				soloWordsIter = m_srclangSoloWords.find(sp.words[0][*setIter]);
				if (soloWordsIter != m_srclangSoloWords.end()) soloWord = true;
			}

			if (soloWord) {
				// Set the base probability to zero because a solo word is with another word in a pattern.
				baseWithoutCurr = neg_inf;
			} else {
				// Log-multiply in the unigram token probability for the single token.
				baseWithoutCurr += uniformBaseProb;
			}
			// Update the index of the last word added to the phrase without word currWord.
			lastWordIndexWithoutCurr = *setIter;
			// Increment the counter of the number of words we've seen so far
			// in the phrase without the current word.
			++jj;
		}
		// If there is a gap between the current word and the last one, add a "|" as a separator.
		if (ii > 0) {
			int distance = *setIter - lastWordIndexWithCurr;
			if (distance > 1) {
				if (m_useGapWidths) {
					int gapWidthToPrint = distance - 1;
					if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
					phraseWithCurr += SPACE_STRING;
					phraseWithCurr += BOUNDARY_STRING;
					phraseWithCurr += itos(gapWidthToPrint);
				} else {
					phraseWithCurr += SPACE_BOUNDARY_STRING;
				}
				++tokensWithCurrNumGaps;
			}
		}
		// If we're not at the beginning, insert a space before the current word.
		if (ii > 0) {
			phraseWithCurr += SPACE_STRING;
		}
		// Append the current word.
		phraseWithCurr += sp.words[0][*setIter];

		// Check if the current word is a solo word.
		soloWord = false;
		if (numWordsCurrColorWithCurr > 1) {
			soloWordsIter = m_srclangSoloWords.find(sp.words[0][*setIter]);
			if (soloWordsIter != m_srclangSoloWords.end()) soloWord = true;
		}

		if (soloWord) {
			// Set the base probability to zero because a solo word is with another word in a single pattern.
			baseWithCurr = neg_inf;
		} else {
			// Log-multiply in the unigram token probability for the single token.
			baseWithCurr += uniformBaseProb;
		}
		// Update the index of the last word added.
		lastWordIndexWithCurr = *setIter;

		// Increment the counter of the number of words we've seen so far
		// in the phrase with the current word.
		++ii;
	}

	// Log-multiply in the terms for the number of gaps and the arrangement of gaps.
	baseWithoutCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithoutCurr, tokensWithoutCurrNumGaps);
	baseWithCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithCurr, tokensWithCurrNumGaps);

	// Now that we have the string, remove the current word from the current color's list to
	// reset things to how they were at the start of this loop iteration.
	wordIndices.erase(currWord);

	double eventLogProb, denom;
	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;

	if (baseWithCurr == neg_inf) {
		eventLogProb = neg_inf;
	} else {
		checkCountsIter = m_counts.find(phraseWithCurr);
		// If we found an entry for the current word, use the count.
		if (checkCountsIter != m_counts.end()) {
			eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithCurr);
		} else {
			// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
			eventLogProb = m_logalpha + baseWithCurr;
		}
	}

	if (baseWithoutCurr == neg_inf) {
		denom = neg_inf;
	} else {
		checkCountsIter = m_counts.find(phraseWithoutCurr);
		if (checkCountsIter != m_counts.end()) {
			denom = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithoutCurr);
		} else {
			// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
			denom = m_logalpha + baseWithoutCurr;
		}
	}

	if (denom == neg_inf) {
		eventLogProb = neg_inf;
	} else {
		eventLogProb -= denom;
	}

	// Fill the fields of the multinomial event.
	me.srcPhraseWithCurr = phraseWithCurr;
	me.srcPhraseWithoutCurr = phraseWithoutCurr;
	me.score = eventLogProb;
	me.dependentSrcWord = currWord;
	me.dependentSrcWordColor = color;
	me.srcColor = false;
}

void GapLM::buildMultinomialEventTargetColorWithSourceDepWord(SentencePair& sp, MultinomialEvent& me,
		int currWord, int color, set<int>& wordIndices, int dependentSrcWordIndex,
		double srcUniformBaseProb, double tgtUniformBaseProb) {

	// Build the strings containing the fragment corresponding to the current color,
	// both with and without the current word currWord.
	GENERIC_STRING phraseWithoutCurr = EMPTY_STRING;
	GENERIC_STRING phraseWithCurr = EMPTY_STRING;

	// Count the number of gaps both with and without the current word.
	int tokensWithoutCurrNumGaps = 0;
	int tokensWithCurrNumGaps = 0;

	int ii = 0;
	int jj = 0;
	int lastWordIndexWithCurr = -1;
	int lastWordIndexWithoutCurr = -1;

	int numWordsCurrColorWithoutCurr = wordIndices.size();
	// Since we're handling a source color in the source sentence,
	// add the current word index to the set for this color.
	wordIndices.insert(currWord);
	int numWordsCurrColorWithCurr = wordIndices.size();

	if (numWordsCurrColorWithoutCurr == 0) {
		cerr << "Warning: TGT: numWordsCurrColorWithoutCurr == 0! currWord = " << currWord << ", color = " << color << endl;
	}
	if (numWordsCurrColorWithCurr == 0) {
		cerr << "Warning: TGT: numWordsCurrColorWithCurr == 0! currWord = " << currWord << ", color = " << color << endl;
	}

	//int numSrcWordsCurrColorWithoutCurr = -1;
	//int numSrcWordsCurrColorWithCurr = -1;

	const set<int>* srcWordIndexSet;
	bool foundSrcWordIndexSet = false;
	map< int, set<int> >::const_iterator iter0 = sp.tgtColorSrcWordIndexSets.find(color);
	if (iter0 != sp.tgtColorSrcWordIndexSets.end()) {
		srcWordIndexSet = &(iter0->second);
		foundSrcWordIndexSet = true;

		//numSrcWordsCurrColorWithoutCurr = srcWordIndexSet->size();
		//srcWordIndexSet->insert(srcWordIndices.begin(), srcWordIndices.end());
		//numSrcWordsCurrColorWithCurr = srcWordIndexSet->size();
	}
	bool dependentSrcWordIndexPresent = false;
	if (dependentSrcWordIndex != -1) dependentSrcWordIndexPresent = true;

	// Initialize the base probabilities to the probability of the number of words,
	// for both the color without the current word and with the current word.
	double baseWithoutCurr = m_logPoissonWPDF[numWordsCurrColorWithoutCurr];
	double baseWithCurr = m_logPoissonWPDF[numWordsCurrColorWithCurr];

	//baseWithoutCurr += m_logPoissonSPDF[numSrcWordsCurrColorWithoutCurr];
	//baseWithCurr += m_logPoissonSPDF[numSrcWordsCurrColorWithCurr];

	bool soloWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	int numSourceWordsAdded;

	set<int>::const_iterator setIter;
	for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
		// Don't add word currWord to phraseWithoutCurr.
		if (*setIter != currWord) {
			// Add to phraseWithoutCurr.
			// If there is a gap between the current word and the last one, add a
			// space and then "|" as a separator.
			if (jj > 0) {
				int distance = *setIter - lastWordIndexWithoutCurr;
				if (distance > 1) {
					if (m_useGapWidths) {
						int gapWidthToPrint = distance - 1;
						if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
						phraseWithoutCurr += SPACE_STRING;
						phraseWithoutCurr += BOUNDARY_STRING;
						phraseWithoutCurr += itos(gapWidthToPrint);
					} else {
						phraseWithoutCurr += SPACE_BOUNDARY_STRING;
					}
					// Increment the number of gaps in the phrase without the current word.
					++tokensWithoutCurrNumGaps;
				}
			}
			// If we're not at the beginning, insert a space before the current word.
			if (jj > 0) {
				phraseWithoutCurr += SPACE_STRING;
			}
			// Append the current word.
			phraseWithoutCurr += sp.words[1][*setIter];

			// We have to include all source words that are linked to this word.
			numSourceWordsAdded = 0;
			if (foundSrcWordIndexSet) {
				// Get the set of all source words that are linked to the current target word.
				set<int>& srcWordCurrSet = sp.tgtWordLinks[*setIter];
				// Go through each one and check if it's one of the source words colored with
				// the current target color.
				set<int>::const_iterator iter;
				for (iter = srcWordCurrSet.begin(); iter != srcWordCurrSet.end(); ++iter) {
					// Check if the current source word is in the set of source words colored with the target color.
					set<int>::const_iterator iter2 = srcWordIndexSet->find(*iter);
					// If we found it,
					if (iter2 != srcWordIndexSet->end()) {
						// Append it to the string, padded by special characters.
						phraseWithoutCurr += BILINGUAL_PADDING_STRING;
						phraseWithoutCurr += sp.words[0][*iter];

						++numSourceWordsAdded;
						baseWithoutCurr += srcUniformBaseProb;
					}
				}
			}
			baseWithoutCurr += m_logPoissonSPDF[numSourceWordsAdded];

			// Check if the current word is a solo word.
			soloWord = false;
			if (numWordsCurrColorWithoutCurr > 1) {
				soloWordsIter = m_soloWords.find(sp.words[1][*setIter]);
				if (soloWordsIter != m_soloWords.end()) soloWord = true;
			}

			if (soloWord) {
				// Set the base probability to zero because a solo word is with another word in a pattern.
				baseWithoutCurr = neg_inf;
			} else {
				// Log-multiply in the unigram token probability for the single token.
				baseWithoutCurr += tgtUniformBaseProb;
			}
			// Update the index of the last word added to the phrase without word currWord.
			lastWordIndexWithoutCurr = *setIter;
			// Increment the counter of the number of words we've seen so far
			// in the phrase without the current word.
			++jj;
		}
		// If there is a gap between the current word and the last one, add a "|" as a separator.
		if (ii > 0) {
			int distance = *setIter - lastWordIndexWithCurr;
			if (distance > 1) {
				if (m_useGapWidths) {
					int gapWidthToPrint = distance - 1;
					if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
					phraseWithCurr += SPACE_STRING;
					phraseWithCurr += BOUNDARY_STRING;
					phraseWithCurr += itos(gapWidthToPrint);
				} else {
					phraseWithCurr += SPACE_BOUNDARY_STRING;
				}
				++tokensWithCurrNumGaps;
			}
		}
		// If we're not at the beginning, insert a space before the current word.
		if (ii > 0) {
			phraseWithCurr += SPACE_STRING;
		}
		// Append the current word.
		phraseWithCurr += sp.words[1][*setIter];

		// We have to include all source words that are linked to this word.
		numSourceWordsAdded = 0;
		if (foundSrcWordIndexSet || dependentSrcWordIndexPresent) {
			// Get the set of all source words that are linked to the current target word.
			set<int>& srcWordCurrSet = sp.tgtWordLinks[*setIter];
			// Go through each one and check if it's one of the source words colored with
			// the current target color.
			set<int>::const_iterator iter;
			for (iter = srcWordCurrSet.begin(); iter != srcWordCurrSet.end(); ++iter) {
				bool foundIt = false;
				// Check if the current source word is in the set of source words colored with the target color.
				if (foundSrcWordIndexSet) {
					set<int>::const_iterator iter2 = srcWordIndexSet->find(*iter);
					if (iter2 != srcWordIndexSet->end()) foundIt = true;
				}
				if (!foundIt && dependentSrcWordIndexPresent) {
					// If we didn't find it, check if dependentSrcWordIndex equals it.
					if (dependentSrcWordIndex == *iter) foundIt = true;
					//set<int>::const_iterator iter3 = dependentSrcWordIndices.find(*iter);
					//if (iter3 != dependentSrcWordIndices.end()) foundIt = true;
				}

				if (foundIt) {
					// Append it to the string, padded by special characters.
					phraseWithCurr += BILINGUAL_PADDING_STRING;
					phraseWithCurr += sp.words[0][*iter];

					++numSourceWordsAdded;
					baseWithCurr += srcUniformBaseProb;
				}
			}
		}
		baseWithCurr += m_logPoissonSPDF[numSourceWordsAdded];

		// Check if the current word is a solo word.
		soloWord = false;
		if (numWordsCurrColorWithCurr > 1) {
			soloWordsIter = m_soloWords.find(sp.words[1][*setIter]);
			if (soloWordsIter != m_soloWords.end()) soloWord = true;
		}

		if (soloWord) {
			// Set the base probability to zero because a solo word is with another word in a single pattern.
			baseWithCurr = neg_inf;
		} else {
			// Log-multiply in the unigram token probability for the single token.
			baseWithCurr += tgtUniformBaseProb;
		}
		// Update the index of the last word added.
		lastWordIndexWithCurr = *setIter;

		// Increment the counter of the number of words we've seen so far
		// in the phrase with the current word.
		++ii;
	}

	// Log-multiply in the terms for the number of gaps and the arrangement of gaps.
	baseWithoutCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithoutCurr, tokensWithoutCurrNumGaps);
	baseWithCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithCurr, tokensWithCurrNumGaps);

	// Now that we have the string, remove the current word from the current color's list to
	// reset things to how they were at the start of this loop iteration.
	wordIndices.erase(currWord);

	double eventLogProb, denom;
	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;

	if (baseWithCurr == neg_inf) {
		eventLogProb = neg_inf;
	} else {
		checkCountsIter = m_counts2.find(phraseWithCurr);
		// If we found an entry for the current word, use the count.
		if (checkCountsIter != m_counts2.end()) {
			eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithCurr);
		} else {
			// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
			eventLogProb = m_logalpha + baseWithCurr;
		}
	}

	if (baseWithoutCurr == neg_inf) {
		denom = neg_inf;
	} else {
		checkCountsIter = m_counts2.find(phraseWithoutCurr);
		if (checkCountsIter != m_counts2.end()) {
			denom = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithoutCurr);
		} else {
			// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
			denom = m_logalpha + baseWithoutCurr;
		}
	}

	if (denom == neg_inf) {
		eventLogProb = neg_inf;
	} else {
		eventLogProb -= denom;
	}

	// Fill the fields of the multinomial event.
	me.phraseWithCurr = phraseWithCurr;
	me.phraseWithoutCurr = phraseWithoutCurr;
	me.score = eventLogProb;
	me.color = color;
	me.srcColor = false;
	me.dependentSrcWord = dependentSrcWordIndex;
}

void GapLM::buildAllSourceColorMultinomialEvents(SentencePair& sp, vector<MultinomialEvent>& samplingChoices,
		int currWord, int newColor, double uniformBaseProb) {

	// The possible moves consist of changing the color of the current source word
	// to any of the source colors in the sentence, including newColor.

	// First, create the multinomial event corresponding to choosing the new source color.

	// Compute the log-probability from the base distribution for tokens.
	double base = m_logPoissonWPDF[1];	// There is just one token.
	// Log-multiply in the unigram source token probability for the single source token.
	base += uniformBaseProb;
	// We don't have to log-multiply in the probability of the number of gaps
	// since there's only one word, there are no gaps, and the probability of
	// 0 gaps given 1 word is 1.0.
	// We also don't need to log-multiply in the log-probability of the
	// arrangement of gaps since it is always one when there aren't any gaps.
	// Therefore, the following line is commented out. If the following function
	// call does not return 0.0, then uncomment the line.
	//base += computeGapPortionOfBaseProbability(1, 0);

	double eventLogProb;
	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;
	checkCountsIter = m_counts.find(sp.words[0][currWord]);

	// If we found an entry for the current word, use the count.
	if (checkCountsIter != m_counts.end()) {
		eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + base);
	} else {
		// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
		eventLogProb = m_logalpha + base;
	}
	// We have the numerator, so we now need to compute the denominator.
	double denom = logAdd(log((double) m_numColors), m_logalpha);
	if (denom == neg_inf) {
		cerr << "Error: dividing by neg_inf!" << endl;
	}
	// Subtract the denominator from the numerator.
	eventLogProb -= denom;

	// Build a multinomial event for the choice of a new source color.
	MultinomialEvent me;
	me.srcPhraseWithCurr = sp.words[0][currWord];
	me.srcPhraseWithoutCurr = EMPTY_STRING;
	me.score = eventLogProb;
	me.dependentSrcWord = currWord;
	me.dependentSrcWordColor = newColor;
	me.srcColor = false;  // We do this because the actual me.color variable represents a target color.
	samplingChoices.push_back(me);

	// Create multinomial events for the existing source colors in the sentence if the word
	// is not a solo word.
	bool soloWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter = m_srclangSoloWords.find(sp.words[0][currWord]);
	if (soloWordsIter != m_srclangSoloWords.end()) soloWord = true;

	// Only create multinomial events if the current word under consideration is not a solo word.
	// Note: we could do this check at the start of this loop iteration for word currWord, but doing it
	// here will cause any solo word that somehow got lumped together with something else to be
	// changed to be a new color by itself. The only downside is a small amount of extra computation.
	if (!soloWord) {
		//cerr << "For word " << currWord << ":" << endl;
		// Go through all source colors in the sentence.
		dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;
		for (iter = sp.colorsToWords[0].begin(); iter != sp.colorsToWords[0].end(); ++iter) {
			MultinomialEvent me;
			//cerr << "Building me for src color " << iter->first << ", which has " << iter->second.size() << " words." << endl;
			buildMultinomialEventSourceColor(sp, me, currWord, iter->first, iter->second, uniformBaseProb);
			samplingChoices.push_back(me);
		}
	}
}

void GapLM::trainBiling(int numIters, const string& count_out_filename2) {
	clock_t clock1, clock2;
	double totalTime;
	int numSentencePairs = m_bilingualData.size();

	cerr << "m_numColors = " << m_numColors << endl;
	cerr << "m_numColors2 = " << m_numColors2 << endl;

	int numTypes[2];
	numTypes[0] = m_srclangUnigramProbabilities.size();
	numTypes[1] = m_unigramProbabilities.size();
	double uniformBaseProb[2];
	uniformBaseProb[0] = log(1.0 / ((double) numTypes[0]));
	uniformBaseProb[1] = log(1.0 / ((double) numTypes[1]));

	bool soloWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	GENERIC_STRING phraseWithCurr;

	bool discrimFalse = false;
	bool bilingualTrue = true;
	bool bilingualNullFalse = false;
	bool computeBaseProbFalse = false;
	bool tgtSentenceTrue = true;
	bool tgtSentenceFalse = false;

	set<int>::const_iterator setIntIter;
	set<int>::const_iterator setIntIter2;

	// go through all the iterations
	for (int m = 1; m <= numIters; ++m) {
		clock1 = clock();
		// Go through every word position in every sentence and record the number of times we change
		// a word's color, keeping track of both source and target words.
		int numSourceColorsChanged = 0;
		int numTargetColorsChanged = 0;
		int numColorsChangedSourceToTarget = 0;

		// Go through all the sentences.
		for (int i = 0; i < numSentencePairs; ++i) {
			SentencePair& sp = m_bilingualData[i];

			///////////////////////////////////////////////////////////////////////////////////////////
			//////////////////////////////// Target Sentence //////////////////////////////////////////
			///////////////////////////////////////////////////////////////////////////////////////////

			// Go through all words in the target sentence.
			for (unsigned int j = 0; j < sp.numWords[1]; ++j) {
				// Get the color of the current target word.
				int currColor = sp.wordsToColors[1][j];
				// Get the set of target words that are colored with this color.
				set<int>& currWordIndices = sp.colorsToWords[1][currColor];
				// Get the number of target words colored with this color.
				int numWordsCurrColor = currWordIndices.size();
				// Revert counts to indicate that we are "removing" the current word from the corpus.
				// First, build the string.
				phraseWithCurr = EMPTY_STRING;
				appendWordsToString(sp, currWordIndices, tgtSentenceTrue, currColor, computeBaseProbFalse, uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);

				// Decrement the count of the pattern.
				//cerr << "T-1TGT: Decrementing count of ||" << phraseWithCurr << "||, before count = " << m_counts2[phraseWithCurr] << endl;
				if (m_counts2[phraseWithCurr] == 1) {
					m_counts2.erase(phraseWithCurr);
				} else {
					--m_counts2[phraseWithCurr];
				}

				// Check if there is a source word linked to target word j that is colored with currColor.
				// When sampling a new color for the target word, we will consider all possibilities for the source word as well.
				int dependentSourceWord = -1;
				int currSrcColor = -1;
				int dependentSourceWordHadSrcColor = -1;

				if (sp.tgtWordLinks[j].size() == 1) {
					dependentSourceWord = *(sp.tgtWordLinks[j].begin());
					// Get the color of the dependent source word.
					currSrcColor = sp.wordsToColors[0][dependentSourceWord];
					// Record whether the color of the dependent source word is a source or target color.
					if (sp.srcWordSrcColor[dependentSourceWord]) {
						dependentSourceWordHadSrcColor = 1;
					} else {
						dependentSourceWordHadSrcColor = 0;
					}

					if (!(sp.srcWordSrcColor[dependentSourceWord]) && currSrcColor != currColor) {
						cerr << "Error: dependentSourceWord (" << dependentSourceWord
							 << ") is colored with a target color but it's different from the color of the linked target word!" << endl;
					}

					if (!(sp.srcWordSrcColor[dependentSourceWord]) && currSrcColor == currColor) {
						// If the linked source word is colored with target color currColor,
						// remove the source word from the source word index set for the target color.
						sp.tgtColorSrcWordIndexSets[currColor].erase(dependentSourceWord);
						if (sp.tgtColorSrcWordIndexSets[currColor].size() == 0) {
							sp.tgtColorSrcWordIndexSets.erase(currColor);
						}
					} else {
						// Here, the linked source word is colored with a source color.

						// Revert the source color counts.
						set<int>& srcCurrWordIndices = sp.colorsToWords[0][currSrcColor];
						// Get the number of words colored with this color.
						int numWordsCurrSrcColor = srcCurrWordIndices.size();
						// Revert counts to indicate that we are "removing" the current word from the corpus.
						// First, build the string.
						phraseWithCurr = EMPTY_STRING;
						appendWordsToString(sp, srcCurrWordIndices, tgtSentenceFalse, -1, computeBaseProbFalse,
							uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);

						//cerr << "T-1SRC: Decrementing count of ||" << phraseWithCurr << "||, before count = " << m_counts[phraseWithCurr] << endl;

						// Decrement the count of the string.
						if (m_counts[phraseWithCurr] == 1) {
							m_counts.erase(phraseWithCurr);
						} else {
							--m_counts[phraseWithCurr];
						}

						if (numWordsCurrSrcColor == 1) {
							// Remove the color of the current word if it's alone.
							// Since the word is alone, remove the entire color entry from the hash map.
							sp.colorsToWords[0].erase(currSrcColor);
							// Remove the color by adding it to the free color stack.
							sp.freeColors[0].push(currSrcColor);
							// Decrement the number of source colors in the sentence pair.
							--(sp.numColors[0]);
							// Decrement the total number of colors in the corpus.
							--m_numColors;
						} else {
							// Remove the current word from the word index set for its color.
							srcCurrWordIndices.erase(dependentSourceWord);

							// Increment the count of the new smaller fragment that remains for this color.
							phraseWithCurr = EMPTY_STRING;
							appendWordsToString(sp, srcCurrWordIndices, tgtSentenceFalse, -1, computeBaseProbFalse,
								uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);
							++m_counts[phraseWithCurr];
							//cerr << "T-2SRC: Incrementing count of ||" << phraseWithCurr << "||, after count = " << m_counts[phraseWithCurr] << endl;

							// TODO: should we also change/update wordsToColors?
						}
					}
				} else if (sp.tgtWordLinks[j].size() > 1) {
					cerr << "Warning: many-to-1 link!" << endl;
				}

				if (numWordsCurrColor == 1) {
					// Remove the color of the current word if it's alone.

					// Since the word is alone, remove the entire color entry from the hash map.
					sp.colorsToWords[1].erase(currColor);
					// Remove the color by adding it to the free color stack.
					sp.freeColors[1].push(currColor);
					// Decrement the number of target colors in the sentence pair.
					--(sp.numColors[1]);
					// Decrement the total number of target colors in the corpus.
					--m_numColors2;
				} else {
					// Remove the current word from the word index set for its color.
					currWordIndices.erase(j);

					// Increment the count of the new smaller pattern that remains for this color.
					phraseWithCurr = EMPTY_STRING;
					appendWordsToString(sp, currWordIndices, tgtSentenceTrue, currColor, computeBaseProbFalse, uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);
					++m_counts2[phraseWithCurr];
					//cerr << "T-2TGT: Incrementing count of ||" << phraseWithCurr << "||, after count = " << m_counts2[phraseWithCurr] << endl;

					// TODO: should we also change/update wordsToColors?
				}

				// Get the next free source color from the source freeColors stack.
				int newSrcColor = -1;
				if (dependentSourceWord != -1) {
					if (sp.freeColors[0].empty()) {
						cerr << "Error: empty source free colors!" << endl;
					}
					newSrcColor = sp.freeColors[0].top();
					// Discard the color we just obtained.
					sp.freeColors[0].pop();
				}

				// Get the next free target color from the target freeColors stack.
				if (sp.freeColors[1].empty()) {
					cerr << "Error: empty target free colors!" << endl;
				}
				int newTgtColor = sp.freeColors[1].top();
				// Discard the color we just obtained.
				sp.freeColors[1].pop();

				// The possible moves consist of changing the color of the current target word
				// to any of the target colors in the sentence, including newColor.

				vector<MultinomialEvent> samplingChoices;

				// Get all possible source color events given the source word index.
				// We use these as the templates that we will reproduce and modify scores of for each possible target event.
				vector<MultinomialEvent> srcColorEvents;
				if (dependentSourceWord != -1) {
					buildAllSourceColorMultinomialEvents(sp, srcColorEvents, dependentSourceWord, newSrcColor, uniformBaseProb[0]);
				}

				MultinomialEvent me;
				buildMultinomialEventTargetNewColor(sp, me, j, newTgtColor, -1, uniformBaseProb[0], uniformBaseProb[1]);
				// If there was a source word, pair the target color event with each source color event.
				if (dependentSourceWord != -1) {
					for (int k = 0; k < srcColorEvents.size(); ++k) {
						MultinomialEvent me2;
						me2.phraseWithCurr = me.phraseWithCurr;
						me2.phraseWithoutCurr = me.phraseWithoutCurr;
						me2.srcPhraseWithCurr = srcColorEvents[k].srcPhraseWithCurr;
						me2.srcPhraseWithoutCurr = srcColorEvents[k].srcPhraseWithoutCurr;
						me2.color = me.color;
						me2.srcColor = false;
						me2.dependentSrcWord = srcColorEvents[k].dependentSrcWord;
						me2.dependentSrcWordColor = srcColorEvents[k].dependentSrcWordColor;
						// For all these events, the target color is not propagated to the source word, so
						// we log-multiply in m_logOneMinusPInitPropagate.
						me2.score = me.score + srcColorEvents[k].score + m_logOneMinusPInitPropagate;
						samplingChoices.push_back(me2);
					}
					// Also create a multinomial event for including the source word in
					// the target color.
					MultinomialEvent me3;
					buildMultinomialEventTargetNewColor(sp, me3, j, newTgtColor, dependentSourceWord,
						uniformBaseProb[0], uniformBaseProb[1]);
					me3.srcPhraseWithCurr = EMPTY_STRING;
					me3.srcPhraseWithoutCurr = EMPTY_STRING;
					me3.srcColor = false;
					me3.dependentSrcWord = dependentSourceWord;
					me3.dependentSrcWordColor = -1;
					// The target color is propagated to the source word, so log-multiply in log(p(propagate)).
					me3.score += m_logPInitPropagate;

					samplingChoices.push_back(me3);

				} else {
					// Set the source fields of the event to dummy values.
					me.srcPhraseWithCurr = EMPTY_STRING;
					me.srcPhraseWithoutCurr = EMPTY_STRING;
					me.srcColor = false;
					me.dependentSrcWord = -1;
					me.dependentSrcWordColor = -1;
					// There is no source word, so we don't need any propagation probability
					// (the probability of not propagating is 1).

					samplingChoices.push_back(me);
				}

				// Create multinomial events for the existing target colors in the sentence if the word
				// is not a solo word.
				soloWord = false;
				soloWordsIter = m_soloWords.find(sp.words[1][j]);
				if (soloWordsIter != m_soloWords.end()) soloWord = true;

				// Only create multinomial events if the current word under consideration is not a solo word.
				// Note: we could do this check at the start of this loop iteration for word j, but doing it
				// here will cause any solo word that somehow got lumped together with something else to be
				// changed to be a new color by itself. The only downside is a small amount of extra computation.
				if (!soloWord) {
					//cerr << "For word " << j << ":" << endl;
					// Go through all target colors in the sentence.
					dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;
					for (iter = sp.colorsToWords[1].begin(); iter != sp.colorsToWords[1].end(); ++iter) {

						MultinomialEvent meWithoutSrc;

						//cerr << "Building me for src color " << iter->first << ", which has " << iter->second.size() << " words." << endl;

						// First build a multinomial event for the current target color without any aligned source word.
						buildMultinomialEventTargetColorWithSourceDepWord(sp, meWithoutSrc, j, iter->first, iter->second,
							-1, uniformBaseProb[0], uniformBaseProb[1]);

						if (dependentSourceWord != -1) {
							// Pair the event without the source word with each of the source color events.
							for (int k = 0; k < srcColorEvents.size(); ++k) {
								MultinomialEvent me2;
								me2.phraseWithCurr = meWithoutSrc.phraseWithCurr;
								me2.phraseWithoutCurr = meWithoutSrc.phraseWithoutCurr;
								me2.srcPhraseWithCurr = srcColorEvents[k].srcPhraseWithCurr;
								me2.srcPhraseWithoutCurr = srcColorEvents[k].srcPhraseWithoutCurr;
								me2.color = meWithoutSrc.color;
								me2.srcColor = false;
								me2.dependentSrcWord = srcColorEvents[k].dependentSrcWord;
								me2.dependentSrcWordColor = srcColorEvents[k].dependentSrcWordColor;
								// For all these events, the target color is not propagated to the source word, so
								// we log-multiply in m_logOneMinusPInitPropagate.
								me2.score = meWithoutSrc.score + srcColorEvents[k].score + m_logOneMinusPInitPropagate;
								samplingChoices.push_back(me2);
							}

							// Also create a multinomial event for including the source word in
							// the target color.
							MultinomialEvent meWithSrc;
							buildMultinomialEventTargetColorWithSourceDepWord(sp, meWithSrc, j, iter->first, iter->second,
								dependentSourceWord, uniformBaseProb[0], uniformBaseProb[1]);

							meWithSrc.srcPhraseWithCurr = EMPTY_STRING;
							meWithSrc.srcPhraseWithoutCurr = EMPTY_STRING;
							meWithSrc.srcColor = false;
							meWithSrc.dependentSrcWord = dependentSourceWord;
							meWithSrc.dependentSrcWordColor = -1;
							// The target color is propagated to the source word, so log-multiply in log(p(propagate)).
							meWithSrc.score += m_logPInitPropagate;

							samplingChoices.push_back(meWithSrc);
						} else {
							// Here, there does not exist any linked source word, so just set the source
							// fields of meWithoutSrc to dummy values.
							meWithoutSrc.srcPhraseWithCurr = EMPTY_STRING;
							meWithoutSrc.srcPhraseWithoutCurr = EMPTY_STRING;
							meWithoutSrc.srcColor = false;
							meWithoutSrc.dependentSrcWord = -1;
							meWithoutSrc.dependentSrcWordColor = -1;
							// There is no source word, so we don't need any propagation probability
							// (the probability of not propagating is 1).

							// Add the event for the current target color to the vector.
							samplingChoices.push_back(meWithoutSrc);
						}
					}
				}

				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				int sampledSrcColor = samplingChoices[sampledEvent].dependentSrcWordColor;
				int sampledSrcWord = samplingChoices[sampledEvent].dependentSrcWord;
				//cerr << "Sampled color: " << sampledColor << "\t" << sampledSrcColor << endl;
				//if (sampledSrcColor) cerr << "Error: while handling target words, chose source color during sampling!" << endl;

				// Set the color of the current target word to sampledColor.
				sp.wordsToColors[1][j] = sampledColor;

				// Increment the count of the new pattern we just added.
				++m_counts2[samplingChoices[sampledEvent].phraseWithCurr];
				//cerr << "T-3TGTNEW: Incrementing the count of ||" << samplingChoices[sampledEvent].phraseWithCurr
				//	 << "||; after count = " << m_counts2[samplingChoices[sampledEvent].phraseWithCurr] << endl;

				// If we changed the color of the current word, increment the counter.
				if (sampledColor != currColor) {
					++numTargetColorsChanged;
				}

				if (sampledSrcWord == -1 && sampledSrcColor == -1) {
					// There is no linked source word.

					if (sampledColor == newTgtColor) {
						// We sampled a new target color, so create a new set for it containing only j.
						set<int> newColorWordIndices;
						newColorWordIndices.insert(j);
						sp.colorsToWords[1].insert(make_pair(newTgtColor, newColorWordIndices));

						// Increment the number of target colors in this sentence pair and the total number
						// since we just created a new one.
						++(sp.numColors[1]);
						++m_numColors2;
					} else {
						//cerr << "Sampled existing target color" << endl;
						// We sampled an existing target color, so just add j to its set.
						sp.colorsToWords[1][sampledColor].insert(j);

						// Decrement the count of the pattern we removed.
						//cerr << "T-3TGTOLD: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr
						//	 << "||; before count = " << m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
						if (m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
							m_counts2.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
						} else {
							--m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr];
						}

						// Since we didn't sample a new target color, push the new color that we
						// were considering back onto the target color stack.
						sp.freeColors[1].push(newTgtColor);
					}
				} else if (sampledSrcColor == -1) {
					// There is a linked source word and it is colored with the target color.

					// Set the color of the source word to the target color.
					sp.wordsToColors[0][sampledSrcWord] = sampledColor;
					sp.srcWordSrcColor[sampledSrcWord] = false;

					// We don't have to increment any counts involving the source word.
					if (newSrcColor == -1) cerr << "Error: newSrcColor == -1" << endl;

					// The new source color was not used, so push it back onto the stack.
					sp.freeColors[0].push(newSrcColor);

					if (sampledColor != currSrcColor && dependentSourceWordHadSrcColor == 0) {
						// If the colors are different and the source word had a target
						// color before, increment the counter.
						++numSourceColorsChanged;
					} else if (dependentSourceWordHadSrcColor == 1) {
						// Or, if the source word had a source color before, increment the counter.
						++numSourceColorsChanged;
						// Since the source word had a source color before and now has a target color,
						// increment the other counter.
						++numColorsChangedSourceToTarget;
					}

					if (sampledColor == newTgtColor) {
						//cerr << "Created new src color: " << newColor << endl;
						// We sampled a new target color, so create a new set for it containing only j.
						set<int> newColorWordIndices;
						newColorWordIndices.insert(j);
						sp.colorsToWords[1].insert(make_pair(newTgtColor, newColorWordIndices));

						// Increment the number of target colors in this sentence pair and the total number
						// since we just created a new one.
						++(sp.numColors[1]);
						++m_numColors2;

						// Adjust the indices of source words colored with target colors appropriately.
						set<int> newColorDependentSrcWords;
						newColorDependentSrcWords.insert(sampledSrcWord);
						sp.tgtColorSrcWordIndexSets.insert(make_pair(newTgtColor, newColorDependentSrcWords));
					} else {
						//cerr << "Sampled existing target color" << endl;
						// We sampled an existing target color, so just add j to its set.
						sp.colorsToWords[1][sampledColor].insert(j);

						// Decrement the count of the pattern we removed.
						//cerr << "T-3SRCTGTOLD: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr
						//	 << "||; before count = " << m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
						if (m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
							m_counts2.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
						} else {
							--m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr];
						}

						// Note: we don't have to do anything with source counts here, because the
						// source word is colored with the target color.

						// Adjust the indices of source words colored with target colors appropriately.
						map< int, set<int> >::iterator mapSetIter = sp.tgtColorSrcWordIndexSets.find(sampledColor);
						if (mapSetIter == sp.tgtColorSrcWordIndexSets.end()) {
							set<int> newColorDependentSrcWords;
							newColorDependentSrcWords.insert(sampledSrcWord);
							sp.tgtColorSrcWordIndexSets.insert(make_pair(sampledColor, newColorDependentSrcWords));
						} else {
							mapSetIter->second.insert(sampledSrcWord);
						}

						// Since we didn't sample a new target color, push the new color that we
						// were considering back onto the target color stack.
						sp.freeColors[1].push(newTgtColor);
					}
				} else {
					// There is a linked source word and it is colored with a source color.

					// Set the color of the source word to its source color.
					sp.wordsToColors[0][sampledSrcWord] = sampledSrcColor;
					sp.srcWordSrcColor[sampledSrcWord] = true;
					// Increment the count of the source word as its sampled color.
					++m_counts[samplingChoices[sampledEvent].srcPhraseWithCurr];
					//cerr << "T-3SRCNEW: Incrementing the count of ||" << samplingChoices[sampledEvent].srcPhraseWithCurr
					//	 << "||; after count = " << m_counts[samplingChoices[sampledEvent].srcPhraseWithCurr] << endl;

					if (sampledSrcColor != currSrcColor && dependentSourceWordHadSrcColor == 1) {
						// If the colors are different and the source word had a source
						// color before, increment the counter.
						++numSourceColorsChanged;
					} else if (dependentSourceWordHadSrcColor == 0) {
						// Or, if the source word had a target color before, increment the counter.
						++numSourceColorsChanged;
						// Since the source word had a target color before and now has a source color,
						// increment the other counter.
						++numColorsChangedSourceToTarget;
					}
					// TODO: figure out when to increment counter of src-to-tgt changes.

					// First, handle the source word.
					if (sampledSrcColor == newSrcColor) {
						//cerr << "Created new src color: " << newColor << endl;
						// We sampled a new source color, so create a new set for it containing only j.
						set<int> newColorWordIndices;
						newColorWordIndices.insert(sampledSrcWord);
						sp.colorsToWords[0].insert(make_pair(newSrcColor, newColorWordIndices));

						// Increment the number of source colors in this sentence pair since we just created a new one.
						++(sp.numColors[0]);
						// Increment the total number of target colors since we just created a new one.
						++m_numColors;
					} else {
						//cerr << "Sampled existing source color" << endl;
						// We sampled an existing source color, so just add sampledSrcWord to its set.
						sp.colorsToWords[0][sampledSrcColor].insert(sampledSrcWord);

						// Decrement the count of the pattern we removed.
						//cerr << "T-3SRCOLD: Decrementing the count of ||" << samplingChoices[sampledEvent].srcPhraseWithoutCurr
						//	 << "||; before count = " << m_counts[samplingChoices[sampledEvent].srcPhraseWithoutCurr] << endl;
						if (m_counts[samplingChoices[sampledEvent].srcPhraseWithoutCurr] == 1) {
							m_counts.erase(samplingChoices[sampledEvent].srcPhraseWithoutCurr);
						} else {
							--m_counts[samplingChoices[sampledEvent].srcPhraseWithoutCurr];
						}

						// Since we didn't sample a new source color, push the new color that we
						// were considering back onto the source color stack.
						sp.freeColors[0].push(newSrcColor);
					}

					if (sampledColor == newTgtColor) {
						// We sampled a new target color, so create a new set for it containing only j.
						set<int> newColorWordIndices;
						newColorWordIndices.insert(j);
						sp.colorsToWords[1].insert(make_pair(newTgtColor, newColorWordIndices));

						// Increment the number of target colors in this sentence pair and the total number
						// since we just created a new one.
						++(sp.numColors[1]);
						++m_numColors2;
					} else {
						//cerr << "Sampled existing target color" << endl;
						// We sampled an existing target color, so just add j to its set.
						sp.colorsToWords[1][sampledColor].insert(j);

						// Decrement the count of the pattern we removed.
						//cerr << "T-3TGTOLDSS: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr
						//	 << "||; before count = " << m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
						if (m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
							m_counts2.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
						} else {
							--m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr];
						}

						// Since we didn't sample a new target color, push the new color that we
						// were considering back onto the target color stack.
						sp.freeColors[1].push(newTgtColor);
					}

				}

				// Handle the source words that came along with target word j.
				//const set<int>& sampledDependentSrcWords = samplingChoices[sampledEvent].dependentSrcWords;
				//for (setIntIter = sampledDependentSrcWords.begin(); setIntIter != sampledDependentSrcWords.end(); ++setIntIter) {
					//sp.wordsToColors[0][*setIntIter] = sampledColor;
					//sp.srcWordSrcColor[*setIntIter] = false;
				//}
			}

			///////////////////////////////////////////////////////////////////////////////////////////
			//////////////////////////////// Source Sentence //////////////////////////////////////////
			///////////////////////////////////////////////////////////////////////////////////////////

			// Go through all words in the source sentence.
			for (unsigned int j = 0; j < sp.numWords[0]; ++j) {
				// Only sample a new color for this word if it is not linked to any target words.
				if (sp.srcWordLinks[j].size() == 0) {
					// Get the color of the current word.
					int currColor = sp.wordsToColors[0][j];
					// Figure out whether it's a source or target color.
					//bool srcColor = sp.srcWordSrcColor[j];

					set<int>& currWordIndices = sp.colorsToWords[0][currColor];
					// Get the number of words colored with this color.
					int numWordsCurrColor = currWordIndices.size();
					// Revert counts to indicate that we are "removing" the current word from the corpus.
					// First, build the string.
					phraseWithCurr = EMPTY_STRING;
					appendWordsToString(sp, currWordIndices, tgtSentenceFalse, -1, computeBaseProbFalse, uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);

					// Decrement the count of the string.
					if (m_counts[phraseWithCurr] == 1) {
						m_counts.erase(phraseWithCurr);
					} else {
						--m_counts[phraseWithCurr];
					}

					if (numWordsCurrColor == 1) {
						// Remove the color of the current word if it's alone.

						// Since the word is alone, remove the entire color entry from the hash map.
						sp.colorsToWords[0].erase(currColor);
						// Remove the color by adding it to the free color stack.
						sp.freeColors[0].push(currColor);
						// Decrement the number of source colors in the sentence pair.
						--(sp.numColors[0]);
						// Decrement the total number of colors in the corpus.
						--m_numColors;
					} else {
						// Remove the current word from the word index set for its color.
						currWordIndices.erase(j);

						// Increment the count of the new smaller fragment that remains for this color.
						phraseWithCurr = EMPTY_STRING;
						appendWordsToString(sp, currWordIndices, tgtSentenceFalse, -1, computeBaseProbFalse, uniformBaseProb[0], uniformBaseProb[1], &phraseWithCurr);
						++m_counts[phraseWithCurr];
						// TODO: should we also change/update wordsToColors?
					}

					// Get the next free source color from the source freeColors stack.
					if (sp.freeColors[0].empty()) {
						cerr << "Error: empty source free colors!" << endl;
					}
					int newColor = sp.freeColors[0].top();
					// Discard the color we just obtained.
					sp.freeColors[0].pop();

					// The possible moves consist of changing the color of the current source word
					// to any of the source colors in the sentence, including newColor.
					vector<MultinomialEvent> samplingChoices;
					buildAllSourceColorMultinomialEvents(sp, samplingChoices, j, newColor, uniformBaseProb[0]);

					// Sample a source color.
					int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
					int sampledColor = samplingChoices[sampledEvent].dependentSrcWordColor;
					//cerr << "Sampled color: " << sampledColor << "\t" << sampledSrcColor << endl;

					// Set the color of the current word to sampledColor.
					sp.wordsToColors[0][j] = sampledColor;
					sp.srcWordSrcColor[j] = true;

					// If we changed the color of the current word, increment the counter.
					if (sampledColor != currColor) {
						++numSourceColorsChanged;
					}

					if (sampledColor == newColor) {
						//cerr << "Created new src color: " << newColor << endl;
						// We sampled a new source color, so create a new set for it containing only j.
						set<int> newColorWordIndices;
						newColorWordIndices.insert(j);
						sp.colorsToWords[0].insert(make_pair(newColor, newColorWordIndices));

						// Increment the count of the new fragment we just added.
						++m_counts[samplingChoices[sampledEvent].srcPhraseWithCurr];
						//cerr << "SRCNEW: Incrementing the count of ||" << samplingChoices[sampledEvent].phraseWithCurr
						//	 << "||; after count = " << m_counts[samplingChoices[sampledEvent].phraseWithCurr] << endl;

						// Increment the total number of colors since we just created a new one.
						++m_numColors;

						// Increment the number of source colors since we just created a new one.
						++(sp.numColors[0]);

					} else {
						//cerr << "Sampled existing source color" << endl;
						// We sampled an existing source color, so just add j to its set.
						sp.colorsToWords[0][sampledColor].insert(j);

						// Increment the count of the new fragment we just added.
						//cerr << "PS: Incrementing the count of ||" << samplingChoices[sampledEvent].phraseWithCurr
						//<< "||; before count = " << m_counts[samplingChoices[sampledEvent].phraseWithCurr] << endl;
						++m_counts[samplingChoices[sampledEvent].srcPhraseWithCurr];
						//cerr << "SRC: Incrementing the count of ||" << samplingChoices[sampledEvent].phraseWithCurr
						//	 << "||; after count = " << m_counts[samplingChoices[sampledEvent].phraseWithCurr] << endl;

						// Decrement the count of the fragment we removed.
						//cerr << "SRC: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr
						//	 << "||; before count = " << m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
						if (m_counts[samplingChoices[sampledEvent].srcPhraseWithoutCurr] == 1) {
							m_counts.erase(samplingChoices[sampledEvent].srcPhraseWithoutCurr);
						} else {
							--m_counts[samplingChoices[sampledEvent].srcPhraseWithoutCurr];
						}

						// Since we didn't sample a new source color, push the new color that we
						// were considering back onto the source color stack.
						sp.freeColors[0].push(newColor);
					}
				}
			}
		}
		clock2 = clock();
		totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
		//cerr << "total time:\t" << totalTime << endl;

		cerr << m << "\t" << totalTime << "\t\t";
		cerr << "src " << numSourceColorsChanged << "\t\ttgt "
			 << numTargetColorsChanged << "\t\tsrctgt " << numColorsChangedSourceToTarget;
		// if we're past the burn-in period, add the current counts to our cumulative counts hashmaps
		double accumTime;
		if (m > m_burnin) {
			//cerr << " Accumulating counts ";
			clock1 = clock();
			accumulateCounts(discrimFalse, bilingualTrue, bilingualNullFalse);
			clock2 = clock();
			accumTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
		}
		cerr << "\tcnts1 " << m_counts.size() << "\tcnts2 " << m_counts2.size();
		if (m > m_burnin) {
			cerr << "\tacc " << accumTime;
		}
		cerr << endl;
	}

	// Now that training is completed, divide the accumulated counts by
	// the number of post-burn-in iterations.
	clock1 = clock();
	divideCounts(bilingualTrue, numIters - m_burnin);
	clock2 = clock();
	double divideTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done computing average counts (" << divideTime << " s).";

	vector<Color> colorsToPrint, colorsToPrint2;
	vector<Color> colorsWidthsToPrint, colorsWidthsToPrint2;

	clock1 = clock();

	double logSumCounts = log(m_cumSumCounts);
	double logSumCounts2 = log(m_cumSumCounts2);

	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator cumCountIter;
	for (cumCountIter = m_cumCounts.begin(); cumCountIter != m_cumCounts.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsToPrint.push_back(c);
		}
	}

	for (cumCountIter = m_cumCounts2.begin(); cumCountIter != m_cumCounts2.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts2;
			colorsToPrint2.push_back(c);
		}
	}

	for (cumCountIter = m_cumCountsWidths.begin(); cumCountIter != m_cumCountsWidths.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsWidthsToPrint.push_back(c);
		}
	}

	for (cumCountIter = m_cumCountsWidths2.begin(); cumCountIter != m_cumCountsWidths2.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts2;
			colorsWidthsToPrint2.push_back(c);
		}
	}

	clock2 = clock();
	double prelimTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done compiling preliminary statistics (" << prelimTime << " s).";

	// First, print raw phrase counts.
	cerr << endl << "Printing phrases sorted by counts (cut-off of 50.0): " << endl;
	cerr << "Total count sum 1: " << m_cumSumCounts << "\t" << logSumCounts << endl;

	// Sort the phrases in decreasing order of raw count.
	sort(colorsToPrint.begin(), colorsToPrint.end(), m_compareColorsObject);
	sort(colorsToPrint2.begin(), colorsToPrint2.end(), m_compareColorsObject);

	// Print phrases to stderr in decreasing order of raw count.
	for (int i = 0; i < colorsToPrint.size(); ++i) {
		if (colorsToPrint[i].count < 50.0) break;
		cerr << colorsToPrint[i].count << "\t\t" << colorsToPrint[i].logProbability << "\t\t" << colorsToPrint[i].phrase << endl;
	}

	cerr << endl << "Total count sum 2: " << m_cumSumCounts2 << "\t" << logSumCounts2 << endl;
	for (int i = 0; i < colorsToPrint2.size(); ++i) {
		if (colorsToPrint2[i].count < 50.0) break;
		cerr << colorsToPrint2[i].count << "\t\t" << colorsToPrint2[i].logProbability << "\t\t" << colorsToPrint2[i].phrase << endl;
	}

	cerr << endl;

	// Print phrases to the count output file in decreasing order of their counts.
	GENERIC_OFSTREAM countOutfile(m_countOutputFilename.c_str());
	bool printToOutputFile = true;
	if (!countOutfile.good()) {
		cerr << "Error opening count output file " << m_countOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsToPrint.size(); ++i) {
			countOutfile << colorsToPrint[i].count << "\t" << colorsToPrint[i].logProbability << "\t" << colorsToPrint[i].phrase << endl;
		}
		countOutfile.close();
	}

	if (printToOutputFile) {
		GENERIC_OFSTREAM countOutfile2(count_out_filename2.c_str());
		if (!countOutfile2.good()) {
			cerr << "Error opening count output file 2 " << count_out_filename2 << ". Only printing abbreviated output to stderr." << endl;
		} else {
			if (USING_WSTRINGS) {
    			countOutfile2.imbue(locale("en_US.utf8"));
			}
			countOutfile2 << m_cumSumCounts2 << "\t" << logSumCounts2 << endl;
			for (int i = 0; i < colorsToPrint2.size(); ++i) {
				countOutfile2 << colorsToPrint2[i].count << "\t" << colorsToPrint2[i].logProbability << "\t" << colorsToPrint2[i].phrase << endl;
			}
			countOutfile2.close();
		}
	}

	// Print phrases with gap widths to the count with gap width output file in decreasing order of their counts.
	/*GENERIC_OFSTREAM countWidthOutfile(m_countWidthOutputFilename.c_str());
	printToOutputFile = true;
	if (!countWidthOutfile.good()) {
		cerr << "Error opening count width output file " << m_countWidthOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countWidthOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countWidthOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
			countWidthOutfile << colorsWidthsToPrint[i].count << "\t" << colorsWidthsToPrint[i].logProbability
				<< "\t" << colorsWidthsToPrint[i].phrase << endl;
		}
		countWidthOutfile.close();
	}*/

	// Now print phrases with gap widths.
	/*cerr << endl << "Printing phrases with widths sorted by counts (cut-off of 20.0): " << endl;
	// Print phrases with widths to stderr in decreasing order of raw count.
	sort(colorsWidthsToPrint.begin(), colorsWidthsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
		if (colorsWidthsToPrint[i].count < 20.0) break;
		cerr << colorsWidthsToPrint[i].count << "\t\t" << colorsWidthsToPrint[i].logProbability
			<< "\t\t" << colorsWidthsToPrint[i].phrase << endl;
	}

	// Print phrases with gap widths to the count with gap width output file in decreasing order of their counts.
	GENERIC_OFSTREAM countWidthOutfile(m_countWidthOutputFilename.c_str());
	printToOutputFile = true;
	if (!countWidthOutfile.good()) {
		cerr << "Error opening count width output file " << m_countWidthOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countWidthOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countWidthOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
			countWidthOutfile << colorsWidthsToPrint[i].count << "\t" << colorsWidthsToPrint[i].logProbability
				<< "\t" << colorsWidthsToPrint[i].phrase << endl;
		}
		countWidthOutfile.close();
	}*/

}

