/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
// This file contains code that implements a gappy pattern model that uses bilingual information 
// to find monolingual (target-side) patterns. It uses word alignments and specifically targets 
// patterns that explain null-aligned target words using other target words. 
// I haven't done extensive testing with this model, but it seems to work. Let me know if you 
// are interested in doing more with this model. --Kevin Gimpel, 7/18/2011
#include "gaplm.h"
#include "utils.h"

// str1 is source language sentence (currently unused)
// str2 is target language sentence
// alignmentString is a string holding the word links
// lineCtr is the line number in the input files of the current tuple (used for printing debug messages)
AlignedSentence GapLM::readSentenceBilingNull(GENERIC_STRING& str1, GENERIC_STRING& str2, GENERIC_STRING& alignmentString, int lineCtr) {
	AlignedSentence s;
	s.seq = str2;
	s.onlySoloWords = false;

	// Tokenize the target sentence.
	s.words.clear();
	tokenize(str2, ' ', s.words);
	int numTokens = s.words.size();
	s.numWords = numTokens;

	// Tokenize the string of word alignments.
	vector<string> alignments;
	tokenize(alignmentString, ' ', alignments);
	set<int> alignedWordSet;
	// Go through the word links.
	for (int i = 0; i < alignments.size(); ++i) {
		// Get the target word in the word link.
		vector<string> wordIndices;
		tokenize(alignments[i], '-', wordIndices);
		alignedWordSet.insert(atoi(wordIndices[1].c_str()));
	}

	// Now, alignedWordSet contains the indices of all the aligned target words.

	// Go through all word indices in the target language sentence and populate the
	// alignedWords and unalignedWords vectors in the AlignedSentence.
	s.alignedWords.reserve(alignedWordSet.size());
	s.alignedWordChildren.reserve(alignedWordSet.size());

	s.unalignedWords.reserve(numTokens - alignedWordSet.size());
	s.unalignedWordOwners.reserve(numTokens - alignedWordSet.size());

	set<int>::const_iterator iter;
	set<int>::const_iterator iter2;

	set<int> currSoloWords;
	set<int>::iterator currSoloWordIter;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	for (int i = 0; i < numTokens; ++i) {
		iter = alignedWordSet.find(i);
		if (iter != alignedWordSet.end()) {
			s.alignedWords.push_back(i);
			set<int> onlyWordi;
			onlyWordi.insert(i);
			s.alignedWordChildren.push_back(onlyWordi);

			// Check if the current aligned word is a solo word.
			bool soloWord = false;
			soloWordsIter = m_soloWords.find(s.words[i]);
			if (soloWordsIter != m_soloWords.end()) soloWord = true;

			if (soloWord) {
				// Add this word index to the set of solo words.
				currSoloWords.insert(i);
			}
		} else {
			// Check if the current unaligned word is a solo word.
			bool soloWord = false;
			soloWordsIter = m_soloWords.find(s.words[i]);
			if (soloWordsIter != m_soloWords.end()) soloWord = true;

			// Only add this word to the unaligned words if it is not a solo word.
			// If it is a solo word, it should always be by itself and therefore
			// will never be changed during sampling and will not change the score
			// of things around it.
			if (!soloWord) s.unalignedWords.push_back(i);
		}
		// Increment the count of this word for the unigram probabilities.
		m_unigramProbabilities[s.words[i]] += 1.0;
	}

	if (s.alignedWords.size() == currSoloWords.size()) {
		cerr << "Sentence " << lineCtr << " contains only solo aligned words: " << str2 << endl;
		s.onlySoloWords = true;
	} else {
		//cerr << "Starting sentence " << lineCtr << endl;
		bool mustFillZero = false;
		int firstGoodOne = -1;
		int lastGoodOne = -1;
		// Now, fill s.unalignedWordOwners by sampling owners for each unaligned word.
		// Fill in s.alignedWordChildren at the same time.
		for (int i = 0; i < s.unalignedWords.size(); ++i) {
			// 
			// Old code that allowed any word to be the owner for each unaligned word:
			// Randomly choose an aligned word from s.alignedWords to be the owner
			// for the unaligned word i.
			int numPossibleOwners = s.alignedWords.size() - currSoloWords.size();
			vector<MultinomialEvent> samplingChoices;
			double eventLogProb = -log(numPossibleOwners);

			// Go through all aligned words and consider them as possible owners,
			// unless they are solo words.
			for (int j = 0; j < s.alignedWords.size(); ++j) {
				if (s.alignedWords[j] == s.unalignedWords[i] - 1 || s.alignedWords[j] == s.unalignedWords[i] + 1) {
					// Look for the current aligned word in the solo words set.
					currSoloWordIter = currSoloWords.find(s.alignedWords[j]);
					// If the current word is not in the solo words set, add it to be
					// considered as a multinomial event.
					if (currSoloWordIter == currSoloWords.end()) {
						MultinomialEvent me;
						me.score = eventLogProb;
						me.color = j;
						//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
						samplingChoices.push_back(me);
					}
				}
			}
			if (samplingChoices.size() > 0) {
				if (firstGoodOne == -1) firstGoodOne = i;
				lastGoodOne = i;
				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledOwner = samplingChoices[sampledEvent].color;
				//cerr << "sampledEvent = " << sampledEvent << ", sampledColor = " << sampledColor << endl;
				s.unalignedWordOwners.push_back(sampledOwner);
				s.alignedWordChildren[sampledOwner].insert(s.unalignedWords[i]);
			} else {
				//cerr << "Warning: no sampling choices for unaligned word " << i << " (" << s.unalignedWords[i] 
				//	 << ") (" << s.words[s.unalignedWords[i]] << ")." << endl;
				if (i > lastGoodOne && lastGoodOne != -1) {
					//cerr << "Using fields from last unaligned word " << s.words[s.unalignedWords[lastGoodOne]] 
					//	 << " which had index " << lastGoodOne << " and owner = " << s.unalignedWordOwners[lastGoodOne] << endl;
					s.unalignedWordOwners.push_back(s.unalignedWordOwners[lastGoodOne]);
					s.alignedWordChildren[s.unalignedWordOwners[lastGoodOne]].insert(s.unalignedWords[i]);
				} else {
					// If i == 0, what should we do here??
					//cerr << "error: still not handled" << endl;
					s.unalignedWordOwners.push_back(-1);
					//mustFillZero = true;
				}
			}
		}
		if (firstGoodOne == -1 && s.unalignedWords.size() != 0) {
			cerr << "Error: firstGoodOne == -1! s.alignedWords.size() = " << s.alignedWords.size() 
				 << ", s.unalignedWords.size() = " << s.unalignedWords.size() << endl;
		}

		for (int i = 0; i < firstGoodOne; ++i) {
			//cerr << "For unaligned word " << i << ", using fields from first good unaligned word " << s.words[s.unalignedWords[firstGoodOne]] 
			//	 << " which had index " << firstGoodOne << " and owner = " << s.unalignedWordOwners[firstGoodOne] << endl;
			s.unalignedWordOwners[i] = s.unalignedWordOwners[firstGoodOne];
			s.alignedWordChildren[s.unalignedWordOwners[firstGoodOne]].insert(s.unalignedWords[i]);
		}
	}
	//cerr << "num_unaligned\t" << lineCtr << "\t" << s.unalignedWords.size() << endl;
	return s;
}

// Goes through sentences in memory and counts fragments (i.e., updates counts hashmaps)
// given current colors.
void GapLM::countWordsBilingNull(bool includeWidths) {
	// Clear all counts.
	if (!includeWidths) {
		m_counts.clear();
		m_numColors = 0;
	}
	// Go through all sentences.
	for (unsigned int i = 0; i < m_alignedData.size(); ++i) {
		AlignedSentence& s = m_alignedData[i];
		// Go through all aligned words in the sentence.
		for (int j = 0; j < s.alignedWords.size(); ++j) {
			if (!includeWidths) ++m_numColors;

			const set<int>& wordIndices = s.alignedWordChildren[j];
			int numWordsCurrColor = wordIndices.size();
			// Build the string containing the pattern corresponding to the current color.
			GENERIC_STRING currWordPhrase = EMPTY_STRING;
			int lastWordIndex = -1;
			set<int>::const_iterator setIter;
			// ii is the number of words processed so far for this color.
			int ii = 0;
			bool gapPresent = false;
			// Go through all words labeled with this color.
			for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
				// If there is a gap between the current word and the last one, add a "|" as a separator.
				if (ii > 0) {
					int distance = *setIter - lastWordIndex;
					if (distance > 1) {
						gapPresent = true;
						if (includeWidths || m_useGapWidths) {
							int gapWidthToPrint = distance - 1;
							if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
							currWordPhrase += BOUNDARY_STRING;
							currWordPhrase += itos(gapWidthToPrint);
							currWordPhrase += SPACE_STRING;
						} else {
							currWordPhrase += BOUNDARY_SPACE_STRING;
						}
					}
				}
				// Append the current word.
				// If the current word is an unaligned word, prepend ~
				if (*setIter != s.alignedWords[j]) currWordPhrase += UNALIGNED_PREFIX_STRING;
				currWordPhrase += s.words[*setIter]; //wordIndices[ii]];
				// If we're not at the end, append a space.
				if (ii < numWordsCurrColor - 1) {
					currWordPhrase += SPACE_STRING;
				}
				// Update the index of the last word added.
				lastWordIndex = *setIter; //wordIndices[ii];
				++ii;
			}
			if (gapPresent) cerr << "Warning: while counting, found pattern with gap: " << currWordPhrase << endl;
			//if (numWordsCurrColor > 1) cerr << "Incrementing the count of ||" << currWordPhrase << "||, ";
			if (includeWidths) {
				++m_cumCountsWidths[currWordPhrase];
			} else {
				++m_counts[currWordPhrase];
			}
			//if (numWordsCurrColor > 1) cerr << "New count: " << m_counts[currWordPhrase] << endl;
		}
	}

	if (!includeWidths) {
		cerr << "Done counting words in sentences" << endl;
	}
}

double GapLM::appendWordsToString(AlignedSentence& as, int alignedWord, const set<int>& wordIndices,
		bool computeBaseProb, double uniformWordDist, GENERIC_STRING* result) {

	int lastWordIndex = -1;
	set<int>::const_iterator setIter;
	int numWords = wordIndices.size();
	double baseProb = m_logPoissonWPDF[numWords];
	int numGaps = 0;
	int ii = 0;
	for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
		//cerr << "\tB4 " << *setIter << endl;
		// If there is a gap between the current word and the last one, add a "|" as a separator.
		if (ii > 0) {
			int distance = *setIter - lastWordIndex;
			if (distance > 1) {
				if (m_useGapWidths) {
					int gapWidthToPrint = distance - 1;
					if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
					result->append(BOUNDARY_STRING);
					result->append(itos(gapWidthToPrint));
					result->append(SPACE_STRING);
				} else {
					result->append(BOUNDARY_SPACE_STRING);
				}
				++numGaps;
			}
		}
		// Append the current word.
		// If the current word is an unaligned word, prepend ~
		if (*setIter != as.alignedWords[alignedWord]) result->append(UNALIGNED_PREFIX_STRING);
		// Append the current word for this color.
		result->append(as.words[*setIter]);

		baseProb += uniformWordDist;

		// If we're not at the end, append a space.
		if (ii < numWords - 1) {
			result->append(SPACE_STRING);
		}
		// Update the index of the last word added.
		lastWordIndex = *setIter;
		++ii;
	}

	// Log-multiply in the probability of the number of gaps.
	cerr << "Error: fix base dist!" << endl;
	//baseProb += m_logPoissonGPDF[numGaps];
	if (numGaps > 0) {
		// We use a uniform distribution over all arrangements of the gaps.
		// Where there are n tokens and k gaps, there are "n choose k" arrangements
		// of gaps. We cache log(1.0 / nchoosek) for values of n and k.
		// Log-multiply the log-probability of this particular arrangement of gaps.
		// Since this is uniform, we only care about the number of gaps (numGaps)
		// and do not look at the actual positions of the gaps.
		baseProb += m_precomputedLogOneOverNChooseK[numWords][numGaps];
	}
	return baseProb;
}

void GapLM::trainBilingNull(int numIters) {
	clock_t clock1, clock2;
	double totalTime;
	int numSentences = m_alignedData.size();

	int numTypes = m_unigramProbabilities.size();
	double uniformBaseProb;
	uniformBaseProb = log(1.0 / ((double) numTypes));

	bool specialWord = false;
	set<int>::const_iterator setIter;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	double eventLogProb;
	double denom;

	dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;

	bool discrimFalse = false;
	bool bilingualTrue = true;
	bool alignedTrue = true;
	bool computeBaseProbTrue = true;
	bool computeBaseProbFalse = false;

	// go through all the iterations
	for (int m = 1; m <= numIters; ++m) {
		clock1 = clock();
		// We will keep track of the number of times that we change a word's owner.
		// This will help in determining when sampling has converged.
		int numOwnersChanged = 0;

		// Go through all the sentences.
		for (int i = 0; i < numSentences; ++i) {
			AlignedSentence& as = m_alignedData[i];

			// Go through all unaligned words in the current sentence and sample new owners for them.
			for (unsigned int j = 0; j < as.unalignedWords.size(); ++j) {
				// First, we have to remove the current unaligned word from its owner.

				// Get the current word's owner.
				int currOwner = as.unalignedWordOwners[j];
				// Get the set of all words owned by the owner.
				set<int>& ownedWords = as.alignedWordChildren[currOwner];
				int numOwnedWords = ownedWords.size();

				// Build the string corresponding to the owned words.
				GENERIC_STRING phraseWithCurr = EMPTY_STRING;
				GENERIC_STRING initialPhraseWithCurr = EMPTY_STRING;
				appendWordsToString(as, currOwner, ownedWords, computeBaseProbFalse, 0.0, &initialPhraseWithCurr);

				// Revert the count to indicate that we are "removing" the current word from the current sentence.
				//cerr << "B4: Decrementing the count of ||" << phraseWithCurr << "||; before count = " << m_counts[phraseWithCurr] << endl;
				if (m_counts[initialPhraseWithCurr] == 0) {
					cerr << "Error: count of \"" << initialPhraseWithCurr << "\" is 0! Cannot decrement (1)" << endl;
				} else {
					if (m_counts[initialPhraseWithCurr] == 1) {
						m_counts.erase(initialPhraseWithCurr);
					} else {
						--m_counts[initialPhraseWithCurr];
					}
				}

				// Remove the current word from the word index set of its owner.
				// ownedWords holds indices in as.words, which is also what unalignedWords holds.
				ownedWords.erase(as.unalignedWords[j]);

				// Increment the count of the new smaller pattern that remains for this owner.
				// TODO_efficiency: we could do this at the same time that we do the above append (ie in the same loop in that fcn).
				GENERIC_STRING initialPhraseWithoutCurr = EMPTY_STRING;
				appendWordsToString(as, currOwner, ownedWords, computeBaseProbFalse, 0.0, &initialPhraseWithoutCurr);
				++m_counts[initialPhraseWithoutCurr];

				//////////////////////////////////////////////////////////////////////////////////
				// We have "removed" the current unaligned word from the state and adjusted     //
				// the counts accordingly. We now sample among all possible owners for the      //
				// current unaligned word. 														//
				//////////////////////////////////////////////////////////////////////////////////

				// The possible moves consist of changing the owner of the current unaligned word
				// to any of the aligned words in the sentence.
				vector<MultinomialEvent> samplingChoices;
				// Create multinomial events for all aligned words in the sentence.

				// Go through all aligned words in the sentence.
				for (int k = 0; k < as.alignedWords.size(); ++k) {
					// Get the set of all words owned by the current aligned word.
					set<int>& currOwnedWords = as.alignedWordChildren[k];
					int currNumOwnedWords = currOwnedWords.size();

					// Build the strings containing the pattern corresponding to the current owner,
					// both with and without the current unaligned word.
					GENERIC_STRING phraseWithoutCurr = EMPTY_STRING;
					GENERIC_STRING phraseWithCurr = EMPTY_STRING;

					// Count the number of gaps both with and without the current word.
					int tokensWithoutCurrNumGaps = 0;
					int tokensWithCurrNumGaps = 0;

					int lastWordIndexWithCurr = -1;
					int ii = 0;
					int lastWordIndexWithoutCurr = -1;
					int jj = 0;

					int currNumOwnedWordsWithoutCurr = currOwnedWords.size();
					// Add the current unaligned word to the current owner's list of owned words.
					currOwnedWords.insert(as.unalignedWords[j]);
					int currNumOwnedWordsWithCurr = currOwnedWords.size();

					// Initialize the base probabilities to the probability of the number of words,
					// for both the owner without the current unaligned word and with it.
					double baseWithoutCurr = m_logPoissonWPDF[currNumOwnedWordsWithoutCurr];
					double baseWithCurr = m_logPoissonWPDF[currNumOwnedWordsWithCurr];

					// Get the new phrase by going through wordIndices again now that we
					// inserted j. We do this in a separate loop because it's tricky to get
					// the spacing and separator insertion right when we do it in a single loop.

					for (setIter = currOwnedWords.begin(); setIter != currOwnedWords.end(); ++setIter) {
						// Don't add word j to phraseWithoutCurr.
						if (*setIter != as.unalignedWords[j]) {
							// Add to phraseWithoutCurr.
							// If there is a gap between the current word and the last one, add a
							// space and then "|" as a separator.
							if (jj > 0) {
								int distance = *setIter - lastWordIndexWithoutCurr;
								if (distance > 1) {
									if (m_useGapWidths) {
										int gapWidthToPrint = distance - 1;
										if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
										phraseWithoutCurr += SPACE_STRING;
										phraseWithoutCurr += BOUNDARY_STRING;
										phraseWithoutCurr += itos(gapWidthToPrint);
									} else {
										phraseWithoutCurr += SPACE_BOUNDARY_STRING;
									}
									// Increment the number of gaps in the phrase without the current word.
									++tokensWithoutCurrNumGaps;
								}
							}
							//if (jj > 0 && (*setIter - lastWordIndexWithoutCurr) > 1) {
							//	phraseWithoutCurr += SPACE_BOUNDARY_STRING;
							//	// Increment the number of gaps in the phrase without the current word.
							//	++tokensWithoutCurrNumGaps;
							//}
							// If we're not at the beginning, insert a space before the current word.
							if (jj > 0) {
								phraseWithoutCurr += SPACE_STRING;
							}
							// If the current word is an unaligned word, prepend ~
							if (*setIter != as.alignedWords[k]) phraseWithoutCurr += UNALIGNED_PREFIX_STRING;
							// Append the current word.
							phraseWithoutCurr += as.words[*setIter];

							// Check if the current word is a solo word.
							specialWord = false;
							if (currNumOwnedWordsWithoutCurr > 1) {
								soloWordsIter = m_soloWords.find(as.words[*setIter]);
								if (soloWordsIter != m_soloWords.end()) specialWord = true;
							}

							if (specialWord) {
								// Set the base probability to zero because a solo word is with another word
								// in a single pattern.
								baseWithoutCurr = neg_inf;
							} else {
								// Log-multiply in the unigram token probability for the single token.
								//baseWithoutCurr += m_unigramProbabilities[s.words[*setIter]];
								baseWithoutCurr += uniformBaseProb;
							}

							// Update the index of the last word added to the phrase without word j.
							lastWordIndexWithoutCurr = *setIter;

							// Increment the counter of the number of words we've seen so far
							// in the phrase without the current word.
							++jj;
						}
						// If there is a gap between the current word and the last one, add a "|" as a separator.
						if (ii > 0) {
							int distance = *setIter - lastWordIndexWithCurr;
							if (distance > 1) {
								if (m_useGapWidths) {
									int gapWidthToPrint = distance - 1;
									if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
									phraseWithCurr += SPACE_STRING;
									phraseWithCurr += BOUNDARY_STRING;
									phraseWithCurr += itos(gapWidthToPrint);
								} else {
									phraseWithCurr += SPACE_BOUNDARY_STRING;
								}
								++tokensWithCurrNumGaps;
							}
						}
						//if (ii > 0 && (*setIter - lastWordIndexWithCurr) > 1) {
						//	phraseWithCurr += SPACE_BOUNDARY_STRING;
						//	++tokensWithCurrNumGaps;
						//}

						// If we're not at the beginning, insert a space before the current word.
						if (ii > 0) {
							phraseWithCurr += SPACE_STRING;
						}
						// If the current word is an unaligned word, prepend ~
						if (*setIter != as.alignedWords[k]) phraseWithCurr += UNALIGNED_PREFIX_STRING;
						// Append the current word.
						phraseWithCurr += as.words[*setIter];

						// Check if the current word is a solo word.
						specialWord = false;
						if (currNumOwnedWordsWithCurr > 1) {
							soloWordsIter = m_soloWords.find(as.words[*setIter]);
							if (soloWordsIter != m_soloWords.end()) specialWord = true;
						}

						if (specialWord) {
							// Set the base probability to zero because a solo word is with another word
							// in a single pattern.
							baseWithCurr = neg_inf;
						} else {
							// Log-multiply in the unigram token probability for the single token.
							//baseWithCurr += m_unigramProbabilities[s.words[*setIter]];
							baseWithCurr += uniformBaseProb;
						}

						// Update the index of the last word added.
						lastWordIndexWithCurr = *setIter;

						// Increment the counter of the number of words we've seen so far
						// in the phrase with the current word.
						++ii;
					}

					// Log-multiply in the terms for the number of gaps and the arrangement of gaps.
					baseWithoutCurr += computeGapPortionOfBaseProbability(currNumOwnedWordsWithoutCurr, tokensWithoutCurrNumGaps);
					baseWithCurr += computeGapPortionOfBaseProbability(currNumOwnedWordsWithCurr, tokensWithCurrNumGaps);

					// Now that we have the string, remove the current word from the current color's list to
					// reset things to how they were at the start of this loop iteration.
					currOwnedWords.erase(as.unalignedWords[j]);

					//cerr << "Score for " << iter->first << ": ||" << phraseWithCurr << "||, dividing ||" << phraseWithoutCurr << "||: ";

					if (tokensWithCurrNumGaps == 0) {
						if (baseWithCurr == neg_inf) {
							eventLogProb = neg_inf;
						} else {
							checkCountsIter = m_counts.find(phraseWithCurr);
							// If we found an entry for the current word, use the count.
							if (checkCountsIter != m_counts.end()) {
								eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithCurr);
							} else {
								// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
								eventLogProb = m_logalpha + baseWithCurr;
							}
						}

						if (baseWithoutCurr == neg_inf) {
							denom = neg_inf;
						} else {
							checkCountsIter = m_counts.find(phraseWithoutCurr);
							if (checkCountsIter != m_counts.end()) {
								denom = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithoutCurr);
							} else {
								// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
								denom = m_logalpha + baseWithoutCurr;
							}
						}

						if (denom == neg_inf) {
							eventLogProb = neg_inf;
						} else {
							eventLogProb -= denom;
						}
						//cerr << eventLogProb << " (" << exp(eventLogProb) << ")" << endl;
						MultinomialEvent me;
						me.phraseWithCurr = phraseWithCurr;
						me.phraseWithoutCurr = phraseWithoutCurr;
						me.score = eventLogProb;
						me.color = k;
						samplingChoices.push_back(me);
					}
				}

				if (samplingChoices.size() != 0) {
					int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
					int sampledOwnerIndex = samplingChoices[sampledEvent].color;
					//cerr << "Sampled color: " << sampledColor << endl;

					// Set the owner of the current unaligned word to sampledOwnerIndex.
					as.unalignedWordOwners[j] = sampledOwnerIndex;
					as.alignedWordChildren[sampledOwnerIndex].insert(as.unalignedWords[j]);

					//sp.wordsToColors[currSentIndex][j] = sampledColor;

					// If we changed the owner of the current word, increment the counter.
					if (sampledOwnerIndex != currOwner) {
						++numOwnersChanged;
					}

					// Increment the count of the new pattern we added.
					++m_counts[samplingChoices[sampledEvent].phraseWithCurr];

					// Decrement the count of the pattern we removed.
					//cerr << "PS: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr << "||; before count = " << m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
					if (m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
						m_counts.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
					} else {
						--m_counts[samplingChoices[sampledEvent].phraseWithoutCurr];
					}
				} else {
					// There were no possible events to sample from, so revert the counts to how they
					// were at the start of this loop iteration.
					++m_counts[initialPhraseWithCurr];

					if (m_counts[initialPhraseWithoutCurr] == 0) {
						cerr << "Error: count of \"" << initialPhraseWithoutCurr << "\" is 0! Cannot decrement (2)" << endl;
					} else {
						if (m_counts[initialPhraseWithoutCurr] == 1) {
							m_counts.erase(initialPhraseWithoutCurr);
						} else {
							--m_counts[initialPhraseWithoutCurr];
						}
					}
				}
			}
		}
		clock2 = clock();
		totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
		//cerr << "total time:\t" << totalTime << endl;

		cerr << "Iteration " << m << " done (" << totalTime << " s). Changed " << numOwnersChanged << " owners.";
		// if we're past the burn-in period, add the current counts to our cumulative counts hashmaps
		if (m > m_burnin) {
			cerr << " Accumulating counts ";
			clock1 = clock();
			accumulateCounts(discrimFalse, bilingualTrue, alignedTrue);
			clock2 = clock();
			double accumTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
			cerr << "(" << accumTime << " s).";
		}
		cerr << endl;
		cerr << "Size of m_counts: " << m_counts.size() << endl;
		// print the corpus if this is a multiple of K
		/*if (m_saveEveryK >= 0 && m % m_saveEveryK == 0) {
			string newOutputFilename = m_outputFilename + "." + itos(m);
			printSegmentedCorpus(newOutputFilename, false);
			string newDecodeOutputFilename = m_outputFilename + ".dec." + itos(m);
			decodeFile(m_decodeFilename, newDecodeOutputFilename, false);
		}*/
	}
	// now that training is completed, divide the accumulated counts by
	// the number of post-burn-in iterations
	clock1 = clock();
	divideCounts(discrimFalse, numIters - m_burnin);
	clock2 = clock();
	double divideTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done computing average counts (" << divideTime << " s).";

	// After inference, print colors with count > 1.
	/*dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::iterator countIter;
	for (countIter = m_counts.begin(); countIter != m_counts.end(); ++countIter) {
		if (countIter->second > 1) {
			cerr << countIter->second << "\t" << countIter->first << endl;
		}
	}*/
	vector<Color> colorsToPrint;
	vector<Color> colorsWidthsToPrint;

	//map<GENERIC_STRING, sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> > condProbs;
	//map<GENERIC_STRING, sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring> >::iterator condProbsIter;

	clock1 = clock();

	double logSumCounts = log(m_cumSumCounts);

	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator cumCountIter;
	for (cumCountIter = m_cumCounts.begin(); cumCountIter != m_cumCounts.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			string::size_type spacePosition = cumCountIter->first.find_first_of(' ');
			if (spacePosition != string::npos) {// && cumCountIter->second > 1.0) {
				Color c;
				c.phrase = cumCountIter->first;
				c.count = cumCountIter->second;
				c.logProbability = log(cumCountIter->second) - logSumCounts;
				colorsToPrint.push_back(c);
			}
		}
	}

	/*for (cumCountIter = m_cumCountsWidths.begin(); cumCountIter != m_cumCountsWidths.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsWidthsToPrint.push_back(c);
		}
	}*/

	clock2 = clock();
	double prelimTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done compiling preliminary statistics (" << prelimTime << " s).";

	//// First, print raw phrase counts.
	cerr << endl << "Printing phrases sorted by counts (cut-off of 20.0): " << endl;
	cerr << "Total count sum: " << m_cumSumCounts << "\t" << logSumCounts << endl;

	// Print phrases to stderr in decreasing order of raw count.
	sort(colorsToPrint.begin(), colorsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsToPrint.size(); ++i) {
		if (colorsToPrint[i].count < 20.0) break;
		cerr << colorsToPrint[i].count << "\t\t" << colorsToPrint[i].logProbability << "\t\t" << colorsToPrint[i].phrase << endl;
	}

	// Print phrases to the count output file in decreasing order of their counts.
	GENERIC_OFSTREAM countOutfile(m_countOutputFilename.c_str());
	bool printToOutputFile = true;
	if (!countOutfile.good()) {
		cerr << "Error opening count output file " << m_countOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsToPrint.size(); ++i) {
			countOutfile << colorsToPrint[i].count << "\t" << colorsToPrint[i].logProbability << "\t" << colorsToPrint[i].phrase << endl;
		}
		countOutfile.close();
	}

	// Now print phrases with gap widths.
	/*cerr << endl << "Printing phrases with widths sorted by counts (cut-off of 20.0): " << endl;
	// Print phrases with widths to stderr in decreasing order of raw count.
	sort(colorsWidthsToPrint.begin(), colorsWidthsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
		if (colorsWidthsToPrint[i].count < 20.0) break;
		cerr << colorsWidthsToPrint[i].count << "\t\t" << colorsWidthsToPrint[i].logProbability
			<< "\t\t" << colorsWidthsToPrint[i].phrase << endl;
	}

	// Print phrases with gap widths to the count with gap width output file in decreasing order of their counts.
	GENERIC_OFSTREAM countWidthOutfile(m_countWidthOutputFilename.c_str());
	printToOutputFile = true;
	if (!countWidthOutfile.good()) {
		cerr << "Error opening count width output file " << m_countWidthOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countWidthOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countWidthOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
			countWidthOutfile << colorsWidthsToPrint[i].count << "\t" << colorsWidthsToPrint[i].logProbability
				<< "\t" << colorsWidthsToPrint[i].phrase << endl;
		}
		countWidthOutfile.close();
	}*/
}
