/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// This file contains code that implements the monolingual gappy pattern model. 
// I tried to provide sufficient comments to understand what each part of the 
// code does. Also, note that this code could be made much more efficient by 
// storing patterns as vectors of ints instead of strings. I use strings just because 
// it makes it easier to experiment with variations (e.g., including gap widths with 
// the gap or specifying different classes of gaps). The bilingual model (in 
// gaplm_biling.cpp) is also one such extension. --Kevin Gimpel, 7/18/2011

#include "gaplm.h"
#include "utils.h"

// Converts a line of input into a Sentence instance, which is returned.
// Assumes tokens are delimited by spaces.
Sentence GapLM::readSentenceMonoling(GENERIC_STRING& str1, int lineCtr) {
	Sentence s;
	s.seq = str1;

	// Tokenize the sentence.
	s.words.clear();
	tokenize(str1, ' ', s.words);
	int numTokens = s.words.size();
	s.numWords = numTokens;
	s.wordIDs.clear();
	s.wordIDs.reserve(numTokens);
	// Fill the vector of word IDs.
	for (int i = 0; i < numTokens; ++i) {
		s.wordIDs.push_back(m_vocabulary.storeIfNew(s.words[i]));
	}

	s.numColors = 0;
	s.colorsToWords.set_empty_key(9999999);
	s.colorsToWords.set_deleted_key(8888888);

	dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;

	set<int> soloWordColors;
	set<int>::iterator soloColorIter;

	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	for (int i = 0; i < numTokens; ++i) {
		bool soloWord = false;
		soloWordsIter = m_soloWords.find(s.words[i]);
		if (soloWordsIter != m_soloWords.end()) soloWord = true;

		if (soloWord) {
			// The new color will be s.numColors.
			int newColor = s.numColors;
			// Increment the number of colors for this sentence.
			++(s.numColors);
			// Add the color assignment for this word.
			s.wordsToColors.push_back(newColor);
			// Add this word to the list of word indices for the new color.
			set<int> wordIndices;
			wordIndices.insert(i);
			s.colorsToWords.insert(make_pair(newColor, wordIndices));
			// Add this new color to the hashset of solo word colors.
			soloWordColors.insert(newColor);
		} else {
			// The probability of creating a new color for this word is m_pInitNewColor.
			// The remaining probability mass is divided among the colors used already in this sentence.

			bool createNewColor = true;
			// We MUST create a new color if we are on the first word or if all previous colors
			// were assigned to solo words.
			// So, if we are past the first word and there are non-solo-word colors, see whether
			// or not we should create a new color.
			if (i > 0 && s.numColors > soloWordColors.size()) {
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				if (randDouble > m_pInitNewColor) {
					createNewColor = false;
				}
			}

			if (createNewColor) {
				// The new color will be s.numColors.
				int newColor = s.numColors;
				// Increment the number of colors for this sentence.
				++(s.numColors);
				// Add the color assignment for this word.
				s.wordsToColors.push_back(newColor);
				// Add this word to the list of word indices for the new color.
				set<int> wordIndices;
				wordIndices.insert(i);
				s.colorsToWords.insert(make_pair(newColor, wordIndices));
			} else {
				// If we're not creating a new color, choose uniformly from among the colors
				// that already exist for this sentence, except for any solo word colors.
				int numPossibleColors = s.numColors - soloWordColors.size();
				vector<MultinomialEvent> samplingChoices;
				double eventLogProb = -log(numPossibleColors);

				for (int ii = 0; ii < s.numColors; ++ii) {
					// Look for the current color in the solo colors set.
					soloColorIter = soloWordColors.find(ii);
					// If the current color is not in the solo colors set, add it to be
					// considered as a multinomial event.
					if (soloColorIter == soloWordColors.end()) {
						MultinomialEvent me;
						me.score = eventLogProb;
						me.color = ii;
						//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
						samplingChoices.push_back(me);
					}
				}
				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				//cerr << "sampledEvent = " << sampledEvent << ", sampledColor = " << sampledColor << endl;

				// Add the color assignment for this word.
				s.wordsToColors.push_back(sampledColor);

				// Search the color map for the color we just sampled.
				iter = s.colorsToWords.find(sampledColor);
				// If it already exists (should always be the case! -- make this more efficient),
				if (iter != s.colorsToWords.end()) {
					// Add the index of the current word to the list of words for the current color.
					iter->second.insert(i);
				} else {
					cerr << "Error: sampled color (" << sampledColor << ") not found in hash map!" << endl;
					// Create a new vector and add it to the hash map.
					set<int> wordIndices;
					wordIndices.insert(i);
					s.colorsToWords.insert(make_pair(sampledColor, wordIndices));
				}
			}
			//cerr << s.words[i] << "~" << prevColor << " ";
		}
		// Increment the count of this word for the unigram probabilities.
		m_unigramProbabilities[s.words[i]] += 1.0;
		//cerr << tokens[i] << "\t" << m_unigramProbabilities[tokens[i]] << endl;
	}
	//cerr << endl;

	// We want to add colors to the freeColors stack for each word in the sentence.
	// First, figure out how many more colors there can possibly be.
	if (s.numColors != s.colorsToWords.size()) {
		cerr << "Warning: s.colorsToWords.size() = " << s.colorsToWords.size() << ", s.numColors = " << s.numColors << endl;
	}
	//int numMoreColors = s.numWords - s.numColors;
	// TODO: we only use <= in the following line instead of < because we want to add one extra
	// color in case anything bad happens. But the algorithm should only need s.numWords colors,
	// so the <= should be able to be changed to a <.
	for (int i = s.numColors; i <= s.numWords; ++i) {
		s.freeColors.push(i);
	}

	if (s.wordsToColors.size() != numTokens) {
		cerr << "Warning: s.wordsToColors.size() (" << s.wordsToColors.size()
			 << ") != numTokens (" << numTokens << ")" << endl;
	}
	// Print out the data structures in the Sentence to ensure there weren't any bugs.
	/*cerr << "Colors for each word: " << endl;
	for (int i = 0; i < s.numWords; ++i) {
		cerr << s.wordsToColors[i] << endl;
	}
	cerr << "Words for each color: " << endl;
	for (int i = 0; i < s.numColors; ++i) {
		cerr << i << ":" << endl;
		vector<int> wordIndices = s.colorsToWords[i];
		for (int j = 0; j < wordIndices.size(); ++j) {
			cerr << "\t" << wordIndices[j] << endl;
		}
	}*/

	return s;
}

// Goes through sentences in memory and counts fragments (i.e., updates counts hashmaps)
// given current colors.
void GapLM::countWordsMonoling(bool includeWidths) {
	// Clear all counts.
	if (!includeWidths) {
		m_counts.clear();
		m_numColors = 0;
	}
	// Go through all sentences.
	for (unsigned int i = 0; i < m_data.size(); ++i) {
		Sentence& s = m_data[i];
		// Go through all colors in the sentence.
		dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::const_iterator iter;
		for (iter = s.colorsToWords.begin(); iter != s.colorsToWords.end(); ++iter) {
			if (!includeWidths) ++m_numColors;
			const set<int>& wordIndices = iter->second;
			int numWordsCurrColor = wordIndices.size();
			// Build the string containing the fragment corresponding to the current color.
			GENERIC_STRING currWordPhrase = EMPTY_STRING;
			int lastWordIndex = -1;
			set<int>::const_iterator setIter;
			// ii is the number of words processed so far for this color.
			int ii = 0;
			// Go through all words labeled with this color.
			for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
				// If there is a gap between the current word and the last one, add a "|" as a separator.
				if (ii > 0) {
					int distance = *setIter - lastWordIndex;
					if (distance > 1) {
						if (includeWidths || m_useGapWidths) {
							int gapWidthToPrint = distance - 1;
							if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
							currWordPhrase += BOUNDARY_STRING;
							currWordPhrase += itos(gapWidthToPrint);
							currWordPhrase += SPACE_STRING;
						} else {
							currWordPhrase += BOUNDARY_SPACE_STRING;
						}
					}
				}
				// Append the current word.
				currWordPhrase += s.words[*setIter]; //wordIndices[ii]];
				// If we're not at the end, append a space.
				if (ii < numWordsCurrColor - 1) {
					currWordPhrase += SPACE_STRING;
				}
				// Update the index of the last word added.
				lastWordIndex = *setIter; //wordIndices[ii];

				++ii;
			}
			//cerr << "Incrementing the count of ||" << currWordPhrase << "||, ";
			if (includeWidths) {
				++m_cumCountsWidths[currWordPhrase];
			} else {
				++m_counts[currWordPhrase];
			}
			//cerr << "New count: " << m_counts[currWordPhrase] << endl;
		}
	}

	if (!includeWidths) {
		cerr << "Done counting words in sentences" << endl;
	}
}

double GapLM::appendWordsToString(Sentence& s, const set<int>& wordIndices,
		bool computeBaseProb, double uniformWordDist, GENERIC_STRING* result) {

	int lastWordIndex = -1;
	set<int>::const_iterator setIter;
	int numWords = wordIndices.size();
	double baseProb = neg_inf;
	if (computeBaseProb) baseProb = m_logPoissonWPDF[numWords];
	int numGaps = 0;
	int ii = 0;
	//cerr << "B4 num words = " << numWordsCurrColor << endl;
	for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
		//cerr << "\tB4 " << *setIter << endl;
		// If there is a gap between the current word and the last one, add a "|" as a separator.
		if (ii > 0) {
			int distance = *setIter - lastWordIndex;
			if (distance > 1) {
				if (m_useGapWidths) {
					int gapWidthToPrint = distance - 1;
					if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
					result->append(BOUNDARY_STRING);
					result->append(itos(gapWidthToPrint));
					result->append(SPACE_STRING);
				} else {
					result->append(BOUNDARY_SPACE_STRING);
				}
				++numGaps;
			}
		}
		// Append the current word.
		result->append(s.words[*setIter]);
		if (computeBaseProb) baseProb += uniformWordDist;
		// If we're not at the end, append a space.
		// If we're not at the end, append a space.
		if (ii < numWords - 1) {
			result->append(SPACE_STRING);
		}
		// Update the index of the last word added.
		lastWordIndex = *setIter;

		++ii;
	}

	// Log-multiply in the probability of the number of gaps.
	if (computeBaseProb) baseProb += computeGapPortionOfBaseProbability(numWords, numGaps);

	return baseProb;
}

void GapLM::trainMonoling(int numIters) {
	clock_t clock1, clock2;
	double totalTime;
	int numSentences = m_data.size();

	int numTypes = m_unigramProbabilities.size();
	if (numTypes == 0) {
		cerr << "Error: 0 words found in unigram distribution!" << endl;
		numTypes = 50000;
	}
	double uniformBaseProb = log(1.0 / ((double) numTypes));

	bool soloWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	bool computeBaseProbFalse = false;
	bool discrimFalse = false;
	bool bilingualFalse = false;

	GENERIC_STRING phraseWithCurr;
	set<int>::const_iterator setIter;
	int ii;

	// Go through numIters sampling iterations.
	for (int m = 1; m <= numIters; ++m) {
		clock1 = clock();

		// Go through every word position in every sentence and record the number of times we change a color.
		int numColorsChanged = 0;

		// Go through all the sentences.
		for (int i = 0; i < numSentences; ++i) {
			Sentence& s = m_data[i];
			// Go through all words in the current sentence.
			for (unsigned int j = 0; j < s.numWords; ++j) {

				// Get the color of the current word.
				int currColor = s.wordsToColors[j];
				// Get the set of words that are colored with this color.
				set<int>& currWordIndices = s.colorsToWords[currColor];
				// Get the number of words colored with this color.
				int numWordsCurrColor = currWordIndices.size();

				// Revert counts to indicate that we are "removing" the current word from the corpus,
				// which is the standard way of thinking about Gibbs sampling.

				// First, build the string
				phraseWithCurr = EMPTY_STRING;
				appendWordsToString(s, currWordIndices, computeBaseProbFalse, uniformBaseProb, &phraseWithCurr);
				if (m_counts[phraseWithCurr] == 1) {
					m_counts.erase(phraseWithCurr);
				} else {
					--m_counts[phraseWithCurr];
				}

				int newColor;
				if (numWordsCurrColor == 1) {
					// Remove the color of the current word if it's alone.

					// Since the word is alone, remove the entire color entry from the hash map.
					s.colorsToWords.erase(currColor);
					// Remove the color by adding it to the free color stack.
					s.freeColors.push(currColor);
					// Decrement the total number of colors in the corpus.
					--m_numColors;
				} else {
					// Remove the current word from the word index set for its color.
					currWordIndices.erase(j);

					// Increment the count of the new smaller fragment that remains for this color.
					phraseWithCurr = EMPTY_STRING;
					appendWordsToString(s, currWordIndices, computeBaseProbFalse, uniformBaseProb, &phraseWithCurr);
					++m_counts[phraseWithCurr];
					// TODO: should we also change/update wordsToColors?
				}

				// Get the next free color from the freeColors stack.
				if (s.freeColors.empty()) {
					cerr << "Error: empty free colors!" << endl;
				}
				newColor = s.freeColors.top();
				// Discard the color we just obtained.
				s.freeColors.pop();

				// The possible moves consist of changing the color of the current word to any of the
				// colors in the sentence, including newColor.

				vector<MultinomialEvent> samplingChoices;

				// First, create the multinomial event corresponding to the new color.

				// Compute the log-probability from the base distribution for tokens.
				double base = m_logPoissonWPDF[1];	// There is just one token.
				// Log-multiply in the unigram token probability for the single token.
				//base += m_unigramProbabilities[s.words[j]];
				// We just use a uniform distribution for this.
				base += uniformBaseProb;
				// We don't have to log-multiply in the probability of the number of gaps
				// since there's only one word, there are no gaps, and the probability of
				// 0 gaps given 1 word is 1.0.
				// We also don't need to log-multiply in the log-probability of the
				// arrangement of gaps since there aren't any gaps.
				// Therefore, the following line is commented out. If the following function
				// call does not return 0.0, then uncomment the line.
				//base += computeGapPortionOfBaseProbability(1, 0);

				double eventLogProb;
				dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter;
				checkCountsIter = m_counts.find(s.words[j]);
				// If we found an entry for the current word, use the count.
				if (checkCountsIter != m_counts.end()) {
					eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + base);
				} else {
					// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
					eventLogProb = m_logalpha + base;
				}
				// We have the numerator, so we now need to compute the denominator.
				double denom = logAdd(log((double) m_numColors), m_logalpha);
				// Subtract the denominator from the numerator.
				eventLogProb -= denom;

				// Build a multinomial event for the choice of a new color.
				MultinomialEvent me;
				me.phraseWithCurr = s.words[j];
				me.phraseWithoutCurr = EMPTY_STRING;
				me.score = eventLogProb;
				me.color = newColor;
				samplingChoices.push_back(me);

				// Note: nowhere do we include a term in the event score for changing the number of
				// patterns in the sentence. This is because we use a uniform distribution over
				// the number of patterns given the number of words. Therefore, if we increase
				// or decrease the number of patterns, the probability of that new number is
				// the same as that of the old (as long as we don't move to an illegal number of
				// patterns outside the range [1, s.numWords]).

				// Create multinomial events for the existing colors in the sentence.
				// Go through all colors in the sentence.
				dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;
				for (iter = s.colorsToWords.begin(); iter != s.colorsToWords.end(); ++iter) {
					set<int>& wordIndices = iter->second;

					// Build the strings containing the fragment corresponding to the current color,
					// both with and without the current word.
					GENERIC_STRING phraseWithoutCurr = EMPTY_STRING;
					GENERIC_STRING phraseWithCurr = EMPTY_STRING;

					// Count the number of gaps both with and without the current word.
					int tokensWithoutCurrNumGaps = 0;
					int tokensWithCurrNumGaps = 0;

					int lastWordIndexWithCurr = -1;
					ii = 0;
					int lastWordIndexWithoutCurr = -1;
					int jj = 0;

					int numWordsCurrColorWithoutCurr = wordIndices.size();
					// Add the current word to the current color.
					wordIndices.insert(j);
					int numWordsCurrColorWithCurr = wordIndices.size();

					// Initialize the base probabilities to the probability of the number of words,
					// for both the color without the current word and with the current word.
					double baseWithoutCurr = m_logPoissonWPDF[numWordsCurrColorWithoutCurr];
					double baseWithCurr = m_logPoissonWPDF[numWordsCurrColorWithCurr];

					// Get the new phrase by going through wordIndices again now that we
					// inserted j. We do this in a separate loop because it's tricky to get
					// the spacing and separator insertion right when we do it in a single
					// loop.
					for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
						// Don't add word j to phraseWithoutCurr.
						if (*setIter != j) {
							// Add to phraseWithoutCurr.
							// If there is a gap between the current word and the last one, add a
							// space and then "|" as a separator.
							if (jj > 0) {
								int distance = *setIter - lastWordIndexWithoutCurr;
								if (distance > 1) {
									if (m_useGapWidths) {
										int gapWidthToPrint = distance - 1;
										if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
										phraseWithoutCurr += SPACE_STRING;
										phraseWithoutCurr += BOUNDARY_STRING;
										phraseWithoutCurr += itos(gapWidthToPrint);
									} else {
										phraseWithoutCurr += SPACE_BOUNDARY_STRING;
									}
									// Increment the number of gaps in the phrase without the current word.
									++tokensWithoutCurrNumGaps;
								}
							}
							//if (jj > 0 && (*setIter - lastWordIndexWithoutCurr) > 1) {
							//	phraseWithoutCurr += SPACE_BOUNDARY_STRING;
							//	// Increment the number of gaps in the phrase without the current word.
							//	++tokensWithoutCurrNumGaps;
							//}
							// If we're not at the beginning, insert a space before the current word.
							if (jj > 0) {
								phraseWithoutCurr += SPACE_STRING;
							}
							// Append the current word.
							phraseWithoutCurr += s.words[*setIter];

							// Check if the current word is a solo word.
							soloWord = false;
							if (numWordsCurrColorWithoutCurr > 1) {
								soloWordsIter = m_soloWords.find(s.words[*setIter]);
								if (soloWordsIter != m_soloWords.end()) soloWord = true;
							}

							if (soloWord) {
								// Set the base probability to zero because a solo word is with another word
								// in a pattern.
								baseWithoutCurr = neg_inf;
							} else {
								// Log-multiply in the unigram token probability for the single token.
								//baseWithoutCurr += m_unigramProbabilities[s.words[*setIter]];
								baseWithoutCurr += uniformBaseProb;
							}

							// Update the index of the last word added to the phrase without word j.
							lastWordIndexWithoutCurr = *setIter;

							// Increment the counter of the number of words we've seen so far
							// in the phrase without the current word.
							++jj;
						}
						// If there is a gap between the current word and the last one, add a "|" as a separator.
						if (ii > 0) {
							int distance = *setIter - lastWordIndexWithCurr;
							if (distance > 1) {
								if (m_useGapWidths) {
									int gapWidthToPrint = distance - 1;
									if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
									phraseWithCurr += SPACE_STRING;
									phraseWithCurr += BOUNDARY_STRING;
									phraseWithCurr += itos(gapWidthToPrint);
								} else {
									phraseWithCurr += SPACE_BOUNDARY_STRING;
								}
								++tokensWithCurrNumGaps;
							}
						}
						//if (ii > 0 && (*setIter - lastWordIndexWithCurr) > 1) {
						//	phraseWithCurr += SPACE_BOUNDARY_STRING;
						//	++tokensWithCurrNumGaps;
						//}

						// If we're not at the beginning, insert a space before the current word.
						if (ii > 0) {
							phraseWithCurr += SPACE_STRING;
						}
						// Append the current word.
						phraseWithCurr += s.words[*setIter];

						// Check if the current word is a solo word.
						soloWord = false;
						//if ((s.words[*setIter] == "." || s.words[*setIter] == "the") && numWordsCurrColorWithCurr > 1) {
						if (numWordsCurrColorWithCurr > 1) {
							soloWordsIter = m_soloWords.find(s.words[*setIter]);
							if (soloWordsIter != m_soloWords.end()) soloWord = true;
						}

						if (soloWord) {
							// Set the base probability to zero because a solo word is with another word
							// in a single pattern.
							baseWithCurr = neg_inf;
						} else {
							// Log-multiply in the unigram token probability for the single token.
							//baseWithCurr += m_unigramProbabilities[s.words[*setIter]];
							baseWithCurr += uniformBaseProb;
						}

						// Update the index of the last word added.
						lastWordIndexWithCurr = *setIter;

						// Increment the counter of the number of words we've seen so far
						// in the phrase with the current word.
						++ii;
					}

					// Log-multiply in the terms for the number of gaps and the arrangement of gaps.
					baseWithoutCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithoutCurr, tokensWithoutCurrNumGaps);
					baseWithCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithCurr, tokensWithCurrNumGaps);

					// Now that we have the string, remove the current word from the current color's list to
					// reset things to how they were at the start of this loop iteration.
					wordIndices.erase(j);

					if (baseWithCurr == neg_inf) {
						eventLogProb = neg_inf;
					} else {
						checkCountsIter = m_counts.find(phraseWithCurr);
						// If we found an entry for the current word, use the count.
						if (checkCountsIter != m_counts.end()) {
							eventLogProb = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							eventLogProb = m_logalpha + baseWithCurr;
						}
					}

					if (baseWithoutCurr == neg_inf) {
						denom = neg_inf;
					} else {
						checkCountsIter = m_counts.find(phraseWithoutCurr);
						if (checkCountsIter != m_counts.end()) {
							denom = logAdd(log((double)(checkCountsIter->second)), m_logalpha + baseWithoutCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							denom = m_logalpha + baseWithoutCurr;
						}
					}

					if (denom == neg_inf) {
						eventLogProb = neg_inf;
					} else {
						eventLogProb -= denom;
					}
					//cerr << eventLogProb << " (" << exp(eventLogProb) << ")" << endl;
					MultinomialEvent me;
					me.phraseWithCurr = phraseWithCurr;
					me.phraseWithoutCurr = phraseWithoutCurr;
					me.score = eventLogProb;
					me.color = iter->first;
					samplingChoices.push_back(me);
				}

				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				//cerr << "Sampled color: " << sampledColor << endl;

				// Set the color of the current word to sampledColor.
				s.wordsToColors[j] = sampledColor;

				// If we changed the color of the current word, increment the counter.
				if (sampledColor != currColor) {
					++numColorsChanged;
				}

				if (sampledColor == newColor) {
					// We sampled a new color, so create a new set for it containing only j.
					set<int> newColorWordIndices;
					newColorWordIndices.insert(j);
					s.colorsToWords.insert(make_pair(newColor, newColorWordIndices));

					// Increment the count of the new fragment we just added.
					++m_counts[samplingChoices[sampledEvent].phraseWithCurr];

					// Increment the total number of colors since we just created a new one.
					++m_numColors;
				} else {
					// We sampled an existing color, so just add j to its set.
					s.colorsToWords[sampledColor].insert(j);

					// Increment the count of the new fragment we just added.
					++m_counts[samplingChoices[sampledEvent].phraseWithCurr];

					// Decrement the count of the fragment we removed.
					if (m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
						m_counts.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
					} else {
						--m_counts[samplingChoices[sampledEvent].phraseWithoutCurr];
					}

					// Since we didn't sample a new color, push the new color that we
					// were considering back onto the stack.
					s.freeColors.push(newColor);
				}
			}
		}
		clock2 = clock();
		totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
		//cerr << "total time:\t" << totalTime << endl;

		cerr << "Iteration " << m << " done (" << totalTime << " s). Changed " << numColorsChanged << " colors.";
		// if we're past the burn-in period, add the current counts to our cumulative counts hashmaps
		if (m > m_burnin) {
			cerr << " Accumulating counts ";
			clock1 = clock();
			accumulateCounts(discrimFalse, bilingualFalse, false);
			clock2 = clock();
			double accumTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
			cerr << "(" << accumTime << " s).";
		}
		cerr << endl;
		cerr << "Size of m_counts: " << m_counts.size() << endl;
	}
	// Now that training is completed, divide the accumulated counts by
	// the number of post-burn-in iterations.
	clock1 = clock();
	divideCounts(discrimFalse, numIters - m_burnin);
	clock2 = clock();
	double divideTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done computing average counts (" << divideTime << " s)." << endl;

	// After inference, print colors with count > 1.
	/*dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::iterator countIter;
	for (countIter = m_counts.begin(); countIter != m_counts.end(); ++countIter) {
		if (countIter->second > 1) {
			cerr << countIter->second << "\t" << countIter->first << endl;
		}
	}*/
	vector<Color> colorsToPrint;
	vector<Color> colorsWidthsToPrint;

	clock1 = clock();

	double logSumCounts = log(m_cumSumCounts);

	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator cumCountIter;
	for (cumCountIter = m_cumCounts.begin(); cumCountIter != m_cumCounts.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsToPrint.push_back(c);
		}
	}

	for (cumCountIter = m_cumCountsWidths.begin(); cumCountIter != m_cumCountsWidths.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsWidthsToPrint.push_back(c);
		}
	}

	clock2 = clock();
	double prelimTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done compiling preliminary statistics (" << prelimTime << " s).";

	//// First, print raw phrase counts.

	cerr << endl << "Printing phrases sorted by counts (cut-off of 20.0): " << endl;
	cerr << "Total count sum: " << m_cumSumCounts << "\t" << logSumCounts << endl;

	if (colorsToPrint.size() == 0) {
		cerr << "Warning! colorsToPrint is empty!" << endl;
	}
	// Print phrases to stderr in decreasing order of raw count.
	sort(colorsToPrint.begin(), colorsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsToPrint.size(); ++i) {
		if (colorsToPrint[i].count < 20.0) break;
		cerr << colorsToPrint[i].count << "\t\t" << colorsToPrint[i].logProbability << "\t\t" << colorsToPrint[i].phrase << endl;
	}

	// Print phrases to the count output file in decreasing order of their counts.
	GENERIC_OFSTREAM countOutfile(m_countOutputFilename.c_str());
	bool printToOutputFile = true;
	if (!countOutfile.good()) {
		cerr << "Error opening count output file " << m_countOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsToPrint.size(); ++i) {
			countOutfile << colorsToPrint[i].count << "\t" << colorsToPrint[i].logProbability << "\t" << colorsToPrint[i].phrase << endl;
		}
		countOutfile.close();
	}

	// Now print phrases with gap widths.
	cerr << endl << "Printing phrases with widths sorted by counts (cut-off of 20.0): " << endl;
	// Print phrases with widths to stderr in decreasing order of raw count.
	sort(colorsWidthsToPrint.begin(), colorsWidthsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
		if (colorsWidthsToPrint[i].count < 20.0) break;
		cerr << colorsWidthsToPrint[i].count << "\t\t" << colorsWidthsToPrint[i].logProbability
			<< "\t\t" << colorsWidthsToPrint[i].phrase << endl;
	}

	// Print phrases with gap widths to the count with gap width output file in decreasing order of their counts.
	GENERIC_OFSTREAM countWidthOutfile(m_countWidthOutputFilename.c_str());
	printToOutputFile = true;
	if (!countWidthOutfile.good()) {
		cerr << "Error opening count width output file " << m_countWidthOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countWidthOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countWidthOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
			countWidthOutfile << colorsWidthsToPrint[i].count << "\t" << colorsWidthsToPrint[i].logProbability
				<< "\t" << colorsWidthsToPrint[i].phrase << endl;
		}
		countWidthOutfile.close();
	}
}
