/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
// This file contains code that implements a monolingual gappy pattern model that uses scores 
// from some external source for each sentence (e.g., BLEU scores for translations for which the 
// provided sentences are references) and incorporates these scores into the pattern generation 
// process. Two distributions are maintained, each pattern in a sentence is generated from one of 
// the distributions, and the choice of distribution depends on the score. The idea is to discover 
// some patterns that appear in sentences that are easy to reach with our MT system and separately 
// discover patterns that appear in sentences that are difficult to reach. I haven't done a lot of 
// testing with this model, and it probably needs some more work to better accomplish what we're 
// trying to do, but the basic functionality seems to be working. Let me know if you are interested 
// in doing more with this model. --Kevin Gimpel, 7/18/2011
#include "gaplm.h"
#include "utils.h"

// Converts a line of input into a Sentence instance, which is returned.
// Assumes tokens are delimited by spaces.
Sentence GapLM::readSentenceMonolingDiscrim(GENERIC_STRING& str1, int lineCtr, double sentenceScore) {
	Sentence s;
	s.seq = str1;

	// Tokenize the sentence.
	s.words.clear();
	tokenize(str1, ' ', s.words);
	int numTokens = s.words.size();
	s.numWords = numTokens;
	s.wordIDs.clear();
	s.wordIDs.reserve(numTokens);
	// Fill the vector of word IDs.
	for (int i = 0; i < numTokens; ++i) {
		s.wordIDs.push_back(m_vocabulary.storeIfNew(s.words[i]));
	}

	s.numColors = 0;
	s.colorsToWords.set_empty_key(9999999);
	s.colorsToWords.set_deleted_key(8888888);
	s.score = sentenceScore;

	dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;

	set<int> specialWordColors;
	set<int>::iterator specialColorIter;

	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	for (int i = 0; i < numTokens; ++i) {
		bool specialWord = false;
		soloWordsIter = m_soloWords.find(s.words[i]);
		if (soloWordsIter != m_soloWords.end()) specialWord = true;
		//if (s.words[i] == "." || s.words[i] == "the") specialWord = true;

		if (specialWord) {
			// The new color will be s.numColors.
			int newColor = s.numColors;
			// Increment the number of colors for this sentence.
			++(s.numColors);
			// Add the color assignment for this word.
			s.wordsToColors.push_back(newColor);
			// Add this word to the list of word indices for the new color.
			set<int> wordIndices;
			wordIndices.insert(i);
			s.colorsToWords.insert(make_pair(newColor, wordIndices));
			// Assign the new color to one of the distributions.
			int sampledDistribution = 1;
			double randDouble = ((double) rand()) / ((double) RAND_MAX);
			// The score should be the probability of choosing distribution 2.
			if (randDouble < s.score) {
				sampledDistribution = 2;
			}
			//cerr << "For score " << s.score << ", (solo) newColor " << newColor << ", sampledDist = " << sampledDistribution << endl;
			s.colorsToDists.insert(make_pair(newColor, sampledDistribution));

			// Add this new color to the hashset of special word colors.
			specialWordColors.insert(newColor);
		} else {
			// The probability of creating a new color for this word is m_pInitNewColor.
			// The remaining probability mass is divided among the colors used already in this sentence.

			bool createNewColor = true;
			// We MUST create a new color if we are on the first word or if all previous colors
			// were assigned to special words.
			// So, if we are past the first word and there are non-special-word colors, see whether
			// or not we should create a new color.
			if (i > 0 && s.numColors > specialWordColors.size()) {
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				if (randDouble > m_pInitNewColor) {
					createNewColor = false;
				}
			}

			if (createNewColor) {
				// The new color will be s.numColors.
				int newColor = s.numColors;
				// Increment the number of colors for this sentence.
				++(s.numColors);
				// Add the color assignment for this word.
				s.wordsToColors.push_back(newColor);
				// Add this word to the list of word indices for the new color.
				set<int> wordIndices;
				wordIndices.insert(i);
				s.colorsToWords.insert(make_pair(newColor, wordIndices));
				// Assign the new color to one of the distributions.
				int sampledDistribution = 1;
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				// The score should be the probability of choosing distribution 2.
				if (randDouble < s.score) {
					sampledDistribution = 2;
				}
				//cerr << "For score " << s.score << ", newColor " << newColor << ", sampledDist = " << sampledDistribution << endl;
				s.colorsToDists.insert(make_pair(newColor, sampledDistribution));
			} else {
				// If we're not creating a new color, choose uniformly from among the colors
				// that already exist for this sentence, except for any special word colors.
				int numPossibleColors = s.numColors - specialWordColors.size();
				vector<MultinomialEvent> samplingChoices;
				double eventLogProb = -log(numPossibleColors);

				for (int ii = 0; ii < s.numColors; ++ii) {
					// Look for the current color in the special colors set.
					specialColorIter = specialWordColors.find(ii);
					// If the current color is not in the special colors set, add it to be
					// considered as a multinomial event.
					if (specialColorIter == specialWordColors.end()) {
						MultinomialEvent me;
						me.score = eventLogProb;
						me.color = ii;
						//cerr << ii << "\t" << me.score << "\t" << exp(me.score) << endl;
						samplingChoices.push_back(me);
					}
				}
				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				//cerr << "sampledEvent = " << sampledEvent << ", sampledColor = " << sampledColor << endl;

				// Add the color assignment for this word.
				s.wordsToColors.push_back(sampledColor);

				// Search the color map for the color we just sampled.
				iter = s.colorsToWords.find(sampledColor);
				// If it already exists (should always be the case! -- make this more efficient),
				if (iter != s.colorsToWords.end()) {
					// Add the index of the current word to the list of words for the current color.
					iter->second.insert(i);
				} else {
					cerr << "Error: sampled color (" << sampledColor << ") not found in hash map!" << endl;
					// Create a new vector and add it to the hash map.
					set<int> wordIndices;
					wordIndices.insert(i);
					s.colorsToWords.insert(make_pair(sampledColor, wordIndices));
				}
			}
			//cerr << s.words[i] << "~" << prevColor << " ";
		}
		// Increment the count of this word for the unigram probabilities.
		m_unigramProbabilities[s.words[i]] += 1.0;
		//cerr << tokens[i] << "\t" << m_unigramProbabilities[tokens[i]] << endl;
	}
	//cerr << endl;

	// We want to add colors to the freeColors stack for each word in the sentence.
	// First, figure out how many more colors there can possibly be.
	if (s.numColors != s.colorsToWords.size()) {
		cerr << "Warning: s.colorsToWords.size() = " << s.colorsToWords.size() << ", s.numColors = " << s.numColors << endl;
	}
	if (s.numColors != s.colorsToDists.size()) {
		cerr << "Warning: s.colorsToDists.size() = " << s.colorsToDists.size() << ", s.numColors = " << s.numColors << endl;
	}
	if (s.colorsToWords.size() != s.colorsToDists.size()) {
		cerr << "Warning: s.colorsToDists.size() = " << s.colorsToDists.size() << ", s.colorsToWords.size() = " << s.colorsToWords.size() << endl;
	}
	//int numMoreColors = s.numWords - s.numColors;
	// TODO: we only use <= in the following line instead of < because we want to add one extra
	// color in case anything bad happens. But the algorithm should only need s.numWords colors,
	// so the <= should be able to be changed to a <.
	for (int i = s.numColors; i <= s.numWords; ++i) {
		s.freeColors.push(i);
	}

	if (s.wordsToColors.size() != numTokens) {
		cerr << "Warning: s.wordsToColors.size() (" << s.wordsToColors.size()
			 << ") != numTokens (" << numTokens << ")" << endl;
	}
	// Print out the data structures in the Sentence to ensure there weren't any bugs.
	/*cerr << "Colors for each word: " << endl;
	for (int i = 0; i < s.numWords; ++i) {
		cerr << s.wordsToColors[i] << endl;
	}
	cerr << "Words for each color: " << endl;
	for (int i = 0; i < s.numColors; ++i) {
		cerr << i << ":" << endl;
		vector<int> wordIndices = s.colorsToWords[i];
		for (int j = 0; j < wordIndices.size(); ++j) {
			cerr << "\t" << wordIndices[j] << endl;
		}
	}*/

	return s;
}

// Goes through sentences in memory and counts fragments (i.e., updates counts hashmaps)
// given current colors.
void GapLM::countWordsMonolingDiscrim(bool includeWidths) {
	// Clear all counts.
	if (!includeWidths) {
		m_counts.clear();
		m_counts2.clear();
		m_numColors = 0;
		m_numColors2 = 0;
	}
	// Go through all sentences.
	for (unsigned int i = 0; i < m_data.size(); ++i) {
		Sentence& s = m_data[i];
		// Go through all colors in the sentence.
		dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::const_iterator iter;
		for (iter = s.colorsToWords.begin(); iter != s.colorsToWords.end(); ++iter) {
			const set<int>& wordIndices = iter->second;
			int numWordsCurrColor = wordIndices.size();
			// Build the string containing the fragment corresponding to the current color.
			GENERIC_STRING currWordPhrase = EMPTY_STRING;
			int lastWordIndex = -1;
			set<int>::const_iterator setIter;
			// ii is the number of words processed so far for this color.
			int ii = 0;
			// Go through all words labeled with this color.
			for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
				// If there is a gap between the current word and the last one, add a "|" as a separator.
				if (ii > 0) {
					int distance = *setIter - lastWordIndex;
					if (distance > 1) {
						if (includeWidths || m_useGapWidths) {
							int gapWidthToPrint = distance - 1;
							if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
							currWordPhrase += BOUNDARY_STRING;
							currWordPhrase += itos(gapWidthToPrint);
							currWordPhrase += SPACE_STRING;
						} else {
							currWordPhrase += BOUNDARY_SPACE_STRING;
						}
					}
				}
				// Append the current word.
				currWordPhrase += s.words[*setIter]; //wordIndices[ii]];
				// If we're not at the end, append a space.
				if (ii < numWordsCurrColor - 1) {
					currWordPhrase += SPACE_STRING;
				}
				// Update the index of the last word added.
				lastWordIndex = *setIter; //wordIndices[ii];

				++ii;
			}
			//cerr << "Incrementing the count of ||" << currWordPhrase << "||, ";

			// Get the distribution that this color belongs to.
			int distIndex = s.colorsToDists[iter->first];

			if (includeWidths) {
				if (distIndex == 1) {
					++m_cumCountsWidths[currWordPhrase];
				} else if (distIndex == 2) {
					++m_cumCountsWidths2[currWordPhrase];
				} else {
					cerr << "Error: distIndex == " << distIndex << endl;
				}
			} else {
				if (distIndex == 1) {
					++m_counts[currWordPhrase];
					++m_numColors;
					//if (m_counts[currWordPhrase] > 4) {
					//	cerr << "1\t" << currWordPhrase << "\t" << m_counts[currWordPhrase] << "\t(" << m_counts2[currWordPhrase] << ")" << endl;
					//}
				} else if (distIndex == 2) {
					++m_counts2[currWordPhrase];
					++m_numColors2;
					//if (m_counts2[currWordPhrase] > 4) {
					//	cerr << "2\t" << currWordPhrase << "\t" << m_counts2[currWordPhrase] << "\t(" << m_counts[currWordPhrase] << ")" << endl;
					//}
				} else {
					cerr << "Error: distIndex == " << distIndex << endl;
				}
			}
			//cerr << "New count: " << m_counts[currWordPhrase] << endl;
		}
	}

	if (!includeWidths) {
		cerr << "Done counting words in sentences" << endl;
	}
}

void GapLM::trainMonolingDiscrim(int numIters, const string& count_out_filename2) {
	clock_t clock1, clock2;
	double totalTime;
	int numSentences = m_data.size();

	int numTypes = m_unigramProbabilities.size();
	if (numTypes == 0) {
		cerr << "Error: 0 words found in unigram distribution!" << endl;
		numTypes = 50000;
	}
	double uniformBaseProb = log(1.0 / ((double) numTypes));

	bool specialWord = false;
	dense_hash_set<GENERIC_STRING, MurmurHash2, eqstring>::const_iterator soloWordsIter;

	bool discrimTrue = true;
	bool computeBaseProbFalse = false;
	bool bilingualFalse = false;

	GENERIC_STRING phraseWithCurr;
	set<int>::const_iterator setIter;
	int ii;

	// Go through numIters sampling iterations.
	for (int m = 1; m <= numIters; ++m) {
		clock1 = clock();

		// Go through every word position in every sentence and record the number of times we change a color.
		int numColorsChanged = 0;
		int numDistIndexesChanged = 0;

		// Go through all the sentences.
		for (int i = 0; i < numSentences; ++i) {
			Sentence& s = m_data[i];
			// Go through all words in the current sentence.
			for (unsigned int j = 0; j < s.numWords; ++j) {

				// Get the color of the current word.
				int currColor = s.wordsToColors[j];
				// Get the set of words that are colored with this color.
				set<int>& currWordIndices = s.colorsToWords[currColor];
				// Get the number of words colored with this color.
				int numWordsCurrColor = currWordIndices.size();
				// Get the distribution index of this color.
				int currDistIndex = s.colorsToDists[currColor];
				if (currDistIndex != 1 && currDistIndex != 2) cerr << "Error: currDistIndex = " << currDistIndex << endl;

				// Revert counts to indicate that we are "removing" the current word from the corpus,
				// which is the standard way of thinking about Gibbs sampling.

				// First, build the string
				phraseWithCurr = EMPTY_STRING;
				appendWordsToString(s, currWordIndices, computeBaseProbFalse, uniformBaseProb, &phraseWithCurr);
				if (currDistIndex == 1) {
					if (m_counts[phraseWithCurr] == 1) {
						m_counts.erase(phraseWithCurr);
					} else {
						--m_counts[phraseWithCurr];
					}
				} else {
					if (m_counts2[phraseWithCurr] == 1) {
						m_counts2.erase(phraseWithCurr);
					} else {
						--m_counts2[phraseWithCurr];
					}
				}

				int newColor;
				if (numWordsCurrColor == 1) {
					// Remove the color of the current word if it's alone.

					// Since the word is alone, remove the entire color entry from the hash map.
					s.colorsToWords.erase(currColor);
					// Remove the color by adding it to the free color stack.
					s.freeColors.push(currColor);
					// Remove the color->distribution index element from the map.
					s.colorsToDists.erase(currColor);
					// Decrement the total number of colors in the corpus.
					if (currDistIndex == 1) {
						--m_numColors;
					} else {
						--m_numColors2;
					}
				} else {
					// Remove the current word from the word index set for its color.
					currWordIndices.erase(j);

					// Increment the count of the new smaller fragment that remains for this color.
					phraseWithCurr = EMPTY_STRING;
					appendWordsToString(s, currWordIndices, computeBaseProbFalse, uniformBaseProb, &phraseWithCurr);
					if (currDistIndex == 1) {
						++m_counts[phraseWithCurr];
					} else {
						++m_counts2[phraseWithCurr];
					}
					// TODO: should we also change/update wordsToColors?
				}

				// Get the next free color from the freeColors stack.
				if (s.freeColors.empty()) {
					cerr << "Error: empty free colors!" << endl;
				}
				newColor = s.freeColors.top();
				// Discard the color we just obtained.
				s.freeColors.pop();

				// The possible moves consist of changing the color of the current word to any of the
				// colors in the sentence, including newColor.

				vector<MultinomialEvent> samplingChoices;

				// First, create the multinomial event corresponding to the new color.

				// Compute the log-probability from the base distribution for tokens.
				double base = m_logPoissonWPDF[1];	// There is just one token.
				// Log-multiply in the unigram token probability for the single token.
				//base += m_unigramProbabilities[s.words[j]];
				base += uniformBaseProb;
				// We don't have to log-multiply in the probability of the number of gaps
				// since there's only one word, there are no gaps, and the probability of
				// 0 gaps given 1 word is 1.0.
				// We also don't need to log-multiply in the log-probability of the
				// arrangement of gaps since there aren't any gaps.
				// Therefore, the following line is commented out. If the following function
				// call does not return 0.0, then uncomment the line.
				//base += computeGapPortionOfBaseProbability(1, 0);

				double eventLogProb1, eventLogProb2;
				dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::const_iterator checkCountsIter1, checkCountsIter2;
				checkCountsIter1 = m_counts.find(s.words[j]);
				// If we found an entry for the current word, use the count.
				if (checkCountsIter1 != m_counts.end()) {
					if (checkCountsIter1->second < 1) {
						cerr << "Warning: count of \"" << s.words[j] << "\" = " << checkCountsIter1->second << endl;
					}
					eventLogProb1 = logAdd(log((double)(checkCountsIter1->second)), m_logalpha + base);
				} else {
					// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
					eventLogProb1 = m_logalpha + base;
				}
				checkCountsIter2 = m_counts2.find(s.words[j]);
				if (checkCountsIter2 != m_counts2.end()) {
					if (checkCountsIter2->second < 1) {
						cerr << "Warning: count of \"" << s.words[j] << "\" = " << checkCountsIter2->second << endl;
					}
					eventLogProb2 = logAdd(log((double)(checkCountsIter2->second)), m_logalpha2 + base);
				} else {
					// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
					eventLogProb2 = m_logalpha2 + base;
				}
				// We have the numerator, so we now need to compute the denominator.
				double denom1 = logAdd(log((double) m_numColors), m_logalpha);
				double denom2 = logAdd(log((double) m_numColors2), m_logalpha2);
				// Subtract the denominator from the numerator.
				eventLogProb1 -= denom1;
				eventLogProb2 -= denom2;
				// Log-multiply in the log-probability of choosing this distribution.
				double logDistribution1Probability = log(1.0 - s.score);
				double logDistribution2Probability = log(s.score);

				// Build a multinomial event for the choice of a new color with distribution index 1.
				MultinomialEvent meNew;
				meNew.phraseWithCurr = s.words[j];
				meNew.phraseWithoutCurr = EMPTY_STRING;
				meNew.score = eventLogProb1 + logDistribution1Probability;
				meNew.color = newColor;
				meNew.distIndex = 1;
				samplingChoices.push_back(meNew);

				// Build a multinomial event for the choice of a new color with distribution index 2.
				MultinomialEvent meNew2;
				meNew2.phraseWithCurr = s.words[j];
				meNew2.phraseWithoutCurr = EMPTY_STRING;
				meNew2.score = eventLogProb2 + logDistribution2Probability;
				meNew2.color = newColor;
				meNew2.distIndex = 2;
				samplingChoices.push_back(meNew2);

				// Note: nowhere do we include a term in the event score for changing the number of
				// patterns in the sentence. This is because we use a uniform distribution over
				// the number of patterns given the number of words. Therefore, if we increase
				// or decrease the number of patterns, the probability of that new number is
				// the same as that of the old (as long as we don't move to an illegal number of
				// patterns outside the range [1, s.numWords]).

				// Create multinomial events for the existing colors in the sentence.
				// Go through all colors in the sentence.
				dense_hash_map<unsigned int, set<int>, MurmurHash2UnsignedInt, equnsignedint>::iterator iter;
				for (iter = s.colorsToWords.begin(); iter != s.colorsToWords.end(); ++iter) {
					set<int>& wordIndices = iter->second;

					// Build the strings containing the fragment corresponding to the current color,
					// both with and without the current word.
					GENERIC_STRING phraseWithoutCurr = EMPTY_STRING;
					GENERIC_STRING phraseWithCurr = EMPTY_STRING;

					// Count the number of gaps both with and without the current word.
					int tokensWithoutCurrNumGaps = 0;
					int tokensWithCurrNumGaps = 0;

					int lastWordIndexWithCurr = -1;
					ii = 0;
					int lastWordIndexWithoutCurr = -1;
					int jj = 0;

					int numWordsCurrColorWithoutCurr = wordIndices.size();
					// Add the current word to the current color.
					wordIndices.insert(j);
					int numWordsCurrColorWithCurr = wordIndices.size();

					int oldDistIndex = s.colorsToDists[iter->first];

					// Initialize the base probabilities to the probability of the number of words,
					// for both the color without the current word and with the current word.
					double baseWithoutCurr = m_logPoissonWPDF[numWordsCurrColorWithoutCurr];
					double baseWithCurr = m_logPoissonWPDF[numWordsCurrColorWithCurr];

					// Get the new phrase by going through wordIndices again now that we
					// inserted j. We do this in a separate loop because it's tricky to get
					// the spacing and separator insertion right when we do it in a single
					// loop.
					for (setIter = wordIndices.begin(); setIter != wordIndices.end(); ++setIter) {
						// Don't add word j to phraseWithoutCurr.
						if (*setIter != j) {
							// Add to phraseWithoutCurr.
							// If there is a gap between the current word and the last one, add a
							// space and then "|" as a separator.
							if (jj > 0) {
								int distance = *setIter - lastWordIndexWithoutCurr;
								if (distance > 1) {
									if (m_useGapWidths) {
										int gapWidthToPrint = distance - 1;
										if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
										phraseWithoutCurr += SPACE_STRING;
										phraseWithoutCurr += BOUNDARY_STRING;
										phraseWithoutCurr += itos(gapWidthToPrint);
									} else {
										phraseWithoutCurr += SPACE_BOUNDARY_STRING;
									}
									// Increment the number of gaps in the phrase without the current word.
									++tokensWithoutCurrNumGaps;
								}
							}
							//if (jj > 0 && (*setIter - lastWordIndexWithoutCurr) > 1) {
							//	phraseWithoutCurr += SPACE_BOUNDARY_STRING;
							//	// Increment the number of gaps in the phrase without the current word.
							//	++tokensWithoutCurrNumGaps;
							//}
							// If we're not at the beginning, insert a space before the current word.
							if (jj > 0) {
								phraseWithoutCurr += SPACE_STRING;
							}
							// Append the current word.
							phraseWithoutCurr += s.words[*setIter];

							// Check if the current word is a solo word.
							specialWord = false;
							if (numWordsCurrColorWithoutCurr > 1) {
								soloWordsIter = m_soloWords.find(s.words[*setIter]);
								if (soloWordsIter != m_soloWords.end()) specialWord = true;
							}

							if (specialWord) {
								// Set the base probability to zero because a solo word is with another word
								// in a pattern.
								baseWithoutCurr = neg_inf;
							} else {
								// Log-multiply in the unigram token probability for the single token.
								//baseWithoutCurr += m_unigramProbabilities[s.words[*setIter]];
								baseWithoutCurr += uniformBaseProb;
							}

							// Update the index of the last word added to the phrase without word j.
							lastWordIndexWithoutCurr = *setIter;

							// Increment the counter of the number of words we've seen so far
							// in the phrase without the current word.
							++jj;
						}
						// If there is a gap between the current word and the last one, add a "|" as a separator.
						if (ii > 0) {
							int distance = *setIter - lastWordIndexWithCurr;
							if (distance > 1) {
								if (m_useGapWidths) {
									int gapWidthToPrint = distance - 1;
									if (gapWidthToPrint > m_maxGapWidth) gapWidthToPrint = m_maxGapWidth;
									phraseWithCurr += SPACE_STRING;
									phraseWithCurr += BOUNDARY_STRING;
									phraseWithCurr += itos(gapWidthToPrint);
								} else {
									phraseWithCurr += SPACE_BOUNDARY_STRING;
								}
								++tokensWithCurrNumGaps;
							}
						}
						//if (ii > 0 && (*setIter - lastWordIndexWithCurr) > 1) {
						//	phraseWithCurr += SPACE_BOUNDARY_STRING;
						//	++tokensWithCurrNumGaps;
						//}

						// If we're not at the beginning, insert a space before the current word.
						if (ii > 0) {
							phraseWithCurr += SPACE_STRING;
						}
						// Append the current word.
						phraseWithCurr += s.words[*setIter];

						// Check if the current word is a solo word.
						specialWord = false;
						//if ((s.words[*setIter] == "." || s.words[*setIter] == "the") && numWordsCurrColorWithCurr > 1) {
						if (numWordsCurrColorWithCurr > 1) {
							soloWordsIter = m_soloWords.find(s.words[*setIter]);
							if (soloWordsIter != m_soloWords.end()) specialWord = true;
						}

						if (specialWord) {
							// Set the base probability to zero because a solo word is with another word
							// in a single pattern.
							baseWithCurr = neg_inf;
						} else {
							// Log-multiply in the unigram token probability for the single token.
							//baseWithCurr += m_unigramProbabilities[s.words[*setIter]];
							baseWithCurr += uniformBaseProb;
						}

						// Update the index of the last word added.
						lastWordIndexWithCurr = *setIter;

						// Increment the counter of the number of words we've seen so far
						// in the phrase with the current word.
						++ii;
					}

					// Log-multiply in the terms for the number of gaps and the arrangement of gaps.
					if (baseWithoutCurr != neg_inf) {
						baseWithoutCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithoutCurr, tokensWithoutCurrNumGaps);
					}
					if (baseWithCurr != neg_inf) {
						baseWithCurr += computeGapPortionOfBaseProbability(numWordsCurrColorWithCurr, tokensWithCurrNumGaps);
					}

					// Now that we have the string, remove the current word from the current color's list to
					// reset things to how they were at the start of this loop iteration.
					wordIndices.erase(j);

					if (baseWithCurr == neg_inf) {
						eventLogProb1 = neg_inf;
						eventLogProb2 = neg_inf;
					} else {
						checkCountsIter1 = m_counts.find(phraseWithCurr);
						// If we found an entry for the current word, use the count.
						if (checkCountsIter1 != m_counts.end()) {
							if (checkCountsIter1->second < 1) {
								cerr << "Warning: count of \"" << phraseWithCurr << "\" = " << checkCountsIter1->second << endl;
							}
							eventLogProb1 = logAdd(log((double)(checkCountsIter1->second)), m_logalpha + baseWithCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							eventLogProb1 = m_logalpha + baseWithCurr;
						}

						checkCountsIter2 = m_counts2.find(phraseWithCurr);
						// If we found an entry for the current word, use the count.
						if (checkCountsIter2 != m_counts2.end()) {
							if (checkCountsIter2->second < 1) {
								cerr << "Warning: count of \"" << phraseWithCurr << "\" = " << checkCountsIter2->second << endl;
							}
							eventLogProb2 = logAdd(log((double)(checkCountsIter2->second)), m_logalpha2 + baseWithCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							eventLogProb2 = m_logalpha2 + baseWithCurr;
						}
					}

					if (baseWithoutCurr == neg_inf) {
						denom1 = neg_inf;
						denom2 = neg_inf;
					} else {
						checkCountsIter1 = m_counts.find(phraseWithoutCurr);
						if (checkCountsIter1 != m_counts.end()) {
							if (checkCountsIter1->second < 1) {
								cerr << "Warning: count of \"" << phraseWithoutCurr << "\" = " << checkCountsIter1->second << endl;
							}
							denom1 = logAdd(log((double)(checkCountsIter1->second)), m_logalpha + baseWithoutCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							denom1 = m_logalpha + baseWithoutCurr;
						}

						checkCountsIter2 = m_counts2.find(phraseWithoutCurr);
						if (checkCountsIter2 != m_counts2.end()) {
							if (checkCountsIter2->second < 1) {
								cerr << "Warning: count of \"" << phraseWithoutCurr << "\" = " << checkCountsIter2->second << endl;
							}
							denom2 = logAdd(log((double)(checkCountsIter2->second)), m_logalpha2 + baseWithoutCurr);
						} else {
							// If we didn't find an entry for the current word, the count is zero so we can use the simpler formula below.
							denom2 = m_logalpha2 + baseWithoutCurr;
						}
					}

					if (denom1 == neg_inf) {
						eventLogProb1 = neg_inf;
					} else {
						eventLogProb1 -= denom1;
						eventLogProb1 += logDistribution1Probability;
					}
					if (denom2 == neg_inf) {
						eventLogProb2 = neg_inf;
					} else {
						eventLogProb2 -= denom2;
						eventLogProb2 += logDistribution2Probability;
					}
					//cerr << eventLogProb << " (" << exp(eventLogProb) << ")" << endl;
					MultinomialEvent me1;
					me1.phraseWithCurr = phraseWithCurr;
					me1.phraseWithoutCurr = phraseWithoutCurr;
					me1.score = eventLogProb1;
					me1.color = iter->first;
					me1.oldDistIndex = oldDistIndex;
					me1.distIndex = 1;
					samplingChoices.push_back(me1);

					MultinomialEvent me2;
					me2.phraseWithCurr = phraseWithCurr;
					me2.phraseWithoutCurr = phraseWithoutCurr;
					me2.score = eventLogProb2;
					me2.color = iter->first;
					me2.oldDistIndex = oldDistIndex;
					me2.distIndex = 2;
					samplingChoices.push_back(me2);
				}

				int sampledEvent = normalizeAndSampleMultinomial(samplingChoices);
				int sampledColor = samplingChoices[sampledEvent].color;
				int sampledDistIndex = samplingChoices[sampledEvent].distIndex;
				int sampledOldDistIndex = samplingChoices[sampledEvent].oldDistIndex;
				//cerr << "Sampled color: " << sampledColor << endl;

				// Set the color of the current word to sampledColor.
				s.wordsToColors[j] = sampledColor;

				// If we changed the color of the current word, increment the counter.
				if (sampledColor != currColor) {
					++numColorsChanged;
				}
				if (sampledDistIndex != currDistIndex) {
					++numDistIndexesChanged;
				}

				if (sampledColor == newColor) {
					// We sampled a new color, so create a new set for it containing only j.
					set<int> newColorWordIndices;
					newColorWordIndices.insert(j);
					s.colorsToWords.insert(make_pair(newColor, newColorWordIndices));
					s.colorsToDists.insert(make_pair(newColor, sampledDistIndex));
					if (sampledDistIndex == 1) {
						// Increment the count of the new fragment we just added.
						++m_counts[samplingChoices[sampledEvent].phraseWithCurr];
						// Increment the total number of colors since we just created a new one.
						++m_numColors;
					} else {
						// Increment the count of the new fragment we just added.
						++m_counts2[samplingChoices[sampledEvent].phraseWithCurr];
						// Increment the total number of colors since we just created a new one.
						++m_numColors2;
					}
				} else {
					// We sampled an existing color, so just add j to its set.
					s.colorsToWords[sampledColor].insert(j);

					// Update the distribution index of the color.
					s.colorsToDists[sampledColor] = sampledDistIndex;

					// Increment the count of the new fragment we just added.
					if (sampledDistIndex == 1) {
						//cerr << "PS: Incrementing the count of ||" << samplingChoices[sampledEvent].phraseWithCurr
						// << "||; before count = " << m_counts[samplingChoices[sampledEvent].phraseWithCurr] << endl;
						++m_counts[samplingChoices[sampledEvent].phraseWithCurr];
					} else {
						++m_counts2[samplingChoices[sampledEvent].phraseWithCurr];
					}


					// Decrement the count of the fragment we removed.
					if (sampledOldDistIndex == 1) {
						//cerr << "PS: Decrementing the count of ||" << samplingChoices[sampledEvent].phraseWithoutCurr
						// << "||; before count = " << m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] << endl;
						if (m_counts[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
							m_counts.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
						} else {
							--m_counts[samplingChoices[sampledEvent].phraseWithoutCurr];
						}
					} else {
						if (m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr] == 1) {
							m_counts2.erase(samplingChoices[sampledEvent].phraseWithoutCurr);
						} else {
							--m_counts2[samplingChoices[sampledEvent].phraseWithoutCurr];
						}
					}

					// Since we didn't sample a new color, push the new color that we
					// were considering back onto the stack.
					s.freeColors.push(newColor);
				}
			}
		}
		clock2 = clock();
		totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
		//cerr << "total time:\t" << totalTime << endl;

		cerr << "Iteration " << m << " done (" << totalTime << " s). Changed " << numColorsChanged
			 << " colors and " << numDistIndexesChanged << " indexes.";
		// if we're past the burn-in period, add the current counts to our cumulative counts hashmaps
		if (m > m_burnin) {
			cerr << " Accumulating counts ";
			clock1 = clock();
			accumulateCounts(discrimTrue, bilingualFalse, false);
			clock2 = clock();
			double accumTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
			cerr << "(" << accumTime << " s).";
		}
		cerr << endl;
		cerr << "1: " << m_counts.size() << "\t2: " << m_counts2.size() << endl;
	}
	// Now that training is completed, divide the accumulated counts by
	// the number of post-burn-in iterations.
	clock1 = clock();
	divideCounts(discrimTrue, numIters - m_burnin);
	clock2 = clock();
	double divideTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done computing average counts (" << divideTime << " s).";

	// After inference, print colors with count > 1.
	/*dense_hash_map<GENERIC_STRING, int, MurmurHash2, eqstring>::iterator countIter;
	for (countIter = m_counts.begin(); countIter != m_counts.end(); ++countIter) {
		if (countIter->second > 1) {
			cerr << countIter->second << "\t" << countIter->first << endl;
		}
	}*/
	vector<Color> colorsToPrint, colorsToPrint2;
	vector<Color> colorsWidthsToPrint, colorsWidthsToPrint2;

	set<string> dist1[5], dist2[5];
	int cutoffs[5];
	cutoffs[0] = 50;
	cutoffs[1] = 100;
	cutoffs[2] = 500;
	cutoffs[3] = 1000;
	cutoffs[4] = 5000;

	clock1 = clock();

	double logSumCounts = log(m_cumSumCounts);
	double logSumCounts2 = log(m_cumSumCounts2);

	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator cumCountIter;
	for (cumCountIter = m_cumCounts.begin(); cumCountIter != m_cumCounts.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsToPrint.push_back(c);
		}
	}

	for (cumCountIter = m_cumCounts2.begin(); cumCountIter != m_cumCounts2.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts2;
			colorsToPrint2.push_back(c);
		}
	}

	for (cumCountIter = m_cumCountsWidths.begin(); cumCountIter != m_cumCountsWidths.end(); ++cumCountIter) {
		if (cumCountIter->second > 0.0) {
			Color c;
			c.phrase = cumCountIter->first;
			c.count = cumCountIter->second;
			c.logProbability = log(cumCountIter->second) - logSumCounts;
			colorsWidthsToPrint.push_back(c);
		}
	}

	clock2 = clock();
	double prelimTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Done compiling preliminary statistics (" << prelimTime << " s).";

	// First, print raw phrase counts.
	cerr << endl << "Printing phrases sorted by counts (cut-off of 25.0): " << endl;
	cerr << "Total count sum 1: " << m_cumSumCounts << "\t" << logSumCounts << endl;

	// Sort the phrases in decreasing order of raw count.
	sort(colorsToPrint.begin(), colorsToPrint.end(), m_compareColorsObject);
	sort(colorsToPrint2.begin(), colorsToPrint2.end(), m_compareColorsObject);
	
	// Print phrases to stderr in decreasing order of raw count.
	for (int i = 0; i < colorsToPrint.size(); ++i) {
		if (colorsToPrint[i].count >= 75.0) {
			cerr << colorsToPrint[i].count << "\t\t" << colorsToPrint[i].logProbability << "\t\t" << colorsToPrint[i].phrase << endl;
		}
		for (int j = 0; j < 5; ++j) {
			if (i < cutoffs[j]) {
				dist1[j].insert(colorsToPrint[i].phrase);
			}
		}		
	}

	cerr << endl << "Total count sum 2: " << m_cumSumCounts2 << "\t" << logSumCounts2 << endl;
	for (int i = 0; i < colorsToPrint2.size(); ++i) {
		if (colorsToPrint2[i].count < 75.0) break;
		cerr << colorsToPrint2[i].count << "\t\t" << colorsToPrint2[i].logProbability << "\t\t" << colorsToPrint2[i].phrase << endl;
		for (int j = 0; j < 5; ++j) {
			if (i < cutoffs[j]) {
				dist2[j].insert(colorsToPrint2[i].phrase);
			}
		}
	}

	cerr << endl;
	// We have now filled dist1 and dist2.
	// Go through the first distribution and print phrases that don't appear in the 
	// second distribution, at various cut-off points.
	set<string>::iterator stringIter;
	cerr << "Printing phrases that appear in distribution 1 that don't appear in distribution 2." << endl;
	for (int i = 0; i < 5; ++i) {
		cerr << "For cutoff " << cutoffs[i] << ":" << endl;
		int maxIndex = cutoffs[i];
		if (maxIndex > colorsToPrint.size()) maxIndex = colorsToPrint.size();
		for (int j = 0; j < maxIndex; ++j) {
			stringIter = dist2[i].find(colorsToPrint[j].phrase);
			if (stringIter == dist2[i].end()) {
				// If it was not found, print it.
				cerr << colorsToPrint[j].count << "\t\t" << colorsToPrint[j].logProbability << "\t\t" << colorsToPrint[j].phrase << endl;
			}
		}
		cerr << endl;
	}
	cerr << endl;
	cerr << "Printing phrases that appear in distribution 2 that don't appear in distribution 1." << endl;
	for (int i = 0; i < 5; ++i) {
		cerr << "For cutoff " << cutoffs[i] << ":" << endl;
		int maxIndex = cutoffs[i];
		if (maxIndex > colorsToPrint2.size()) maxIndex = colorsToPrint2.size();
		for (int j = 0; j < maxIndex; ++j) {
			stringIter = dist1[i].find(colorsToPrint2[j].phrase);
			if (stringIter == dist1[i].end()) {
				// If it was not found, print it.
				cerr << colorsToPrint2[j].count << "\t\t" << colorsToPrint2[j].logProbability << "\t\t" << colorsToPrint2[j].phrase << endl;
			}
		}
		cerr << endl;
	}

	// Print phrases to the count output file in decreasing order of their counts.
	GENERIC_OFSTREAM countOutfile(m_countOutputFilename.c_str());
	bool printToOutputFile = true;
	if (!countOutfile.good()) {
		cerr << "Error opening count output file " << m_countOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsToPrint.size(); ++i) {
			countOutfile << colorsToPrint[i].count << "\t" << colorsToPrint[i].logProbability << "\t" << colorsToPrint[i].phrase << endl;
		}
		//countOutfile.close();
	}

	if (printToOutputFile) {
		GENERIC_OFSTREAM countOutfile2(count_out_filename2.c_str());
		if (!countOutfile2.good()) {
			cerr << "Error opening count output file 2 " << count_out_filename2 << ". Only printing abbreviated output to stderr." << endl;
		} else {
			if (USING_WSTRINGS) {
    			countOutfile2.imbue(locale("en_US.utf8"));
			}
			countOutfile2 << m_cumSumCounts2 << "\t" << logSumCounts2 << endl;
			for (int i = 0; i < colorsToPrint2.size(); ++i) {
				countOutfile2 << colorsToPrint2[i].count << "\t" << colorsToPrint2[i].logProbability << "\t" << colorsToPrint2[i].phrase << endl;
			}
			countOutfile2.close();
		}
	}

	// Now print phrases with gap widths.
	/*cerr << endl << "Printing phrases with widths sorted by counts (cut-off of 20.0): " << endl;
	// Print phrases with widths to stderr in decreasing order of raw count.
	sort(colorsWidthsToPrint.begin(), colorsWidthsToPrint.end(), m_compareColorsObject);
	for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
		if (colorsWidthsToPrint[i].count < 20.0) break;
		cerr << colorsWidthsToPrint[i].count << "\t\t" << colorsWidthsToPrint[i].logProbability
			<< "\t\t" << colorsWidthsToPrint[i].phrase << endl;
	}

	// Print phrases with gap widths to the count with gap width output file in decreasing order of their counts.
	GENERIC_OFSTREAM countWidthOutfile(m_countWidthOutputFilename.c_str());
	printToOutputFile = true;
	if (!countWidthOutfile.good()) {
		cerr << "Error opening count width output file " << m_countWidthOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	countWidthOutfile.imbue(locale("en_US.utf8"));
	}
	if (printToOutputFile) {
		//sort(colorsToPrint.begin(), colorsToPrint.end(), CompareColorsAlphabeticalObject);
		countWidthOutfile << m_cumSumCounts << "\t" << logSumCounts << endl;
		for (int i = 0; i < colorsWidthsToPrint.size(); ++i) {
			countWidthOutfile << colorsWidthsToPrint[i].count << "\t" << colorsWidthsToPrint[i].logProbability
				<< "\t" << colorsWidthsToPrint[i].phrase << endl;
		}
		countWidthOutfile.close();
	}*/

	// Now, print conditional probabilities.
	/*cerr << endl << "Printing conditional probabilities: " << endl;
	sparse_hash_map<GENERIC_STRING, double, MurmurHash2, eqstring>::iterator entriesIter;

	GENERIC_OFSTREAM outfile(m_condProbOutputFilename.c_str());
	printToOutputFile = true;
	if (!outfile.good()) {
		cerr << "Error opening conditional probability output file " << m_condProbOutputFilename << ". Only printing abbreviated output to stderr." << endl;
		printToOutputFile = false;
	}
	if (USING_WSTRINGS && printToOutputFile) {
    	outfile.imbue(locale("en_US.utf8"));
	}

	for (condProbsIter = condProbs.begin(); condProbsIter != condProbs.end(); ++condProbsIter) {
		vector<CPTEvent> cptEvents;
		double sum = neg_inf;
		// Go through all entries in the conditional probability table for the current word.
		for (entriesIter = condProbsIter->second.begin(); entriesIter != condProbsIter->second.end(); ++entriesIter) {
			sum = logAdd(sum, log(entriesIter->second));
		}

		// Only print to stderr if the sum of all occurrences of the first word is at least 20.
		bool printToStderr = true;
		if (exp(sum) >= 20.0) {
			cerr << condProbsIter->first << "\t" << exp(sum) << endl;
		} else {
			printToStderr = false;
		}

		outfile << condProbsIter->first << "\t" << exp(sum) << endl;
		for (entriesIter = condProbsIter->second.begin(); entriesIter != condProbsIter->second.end(); ++entriesIter) {
			CPTEvent cptEvent;
			cptEvent.phrase = entriesIter->first;
			cptEvent.condProb = exp(log(entriesIter->second) - sum);
			cptEvent.rawCount = entriesIter->second;
			cptEvents.push_back(cptEvent);
			//cerr << "\t" << entriesIter->first << "\t" << exp(log(entriesIter->second) - sum) << "\t" << entriesIter->second << endl;
		}
		sort(cptEvents.begin(), cptEvents.end(), compareCPTEvents);
		for (int i = 0; i < cptEvents.size(); ++i) {
			// Stop printing entries to stderr once we get past probability 0.0001.
			if (printToStderr && cptEvents[i].condProb >= 0.0002 && i <= 100) {
				cerr << "\t" << cptEvents[i].phrase << "\t\t" << cptEvents[i].condProb << "\t\t" << cptEvents[i].rawCount << endl;
			}
			if (printToOutputFile) {
				outfile << "\t" << cptEvents[i].phrase << "\t" << cptEvents[i].condProb << "\t" << cptEvents[i].rawCount << endl;
			}
		}
		if (printToStderr) cerr << endl;
	}
	if (printToOutputFile) outfile.close();
	*/
}
