/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
//
// File:   gaplmtrain.cpp
// Author: Kevin Gimpel
//
// Created on July 19, 2010
//

#include "gaplmtrain.h"

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("gaplmtrain -- train nonparametric Bayesian gappy pattern models");
    params.addConstraint("input_file", "file containing input sentences", false); // not optional
    params.addConstraint("count_output_file", "filename in which to output pattern counts from posterior samples (source pattern counts if model = biling)", false); // not optional
    params.addConstraint("count_output_file2", "filename in which to output pattern counts for target patterns (used if model = biling)", true); // optional
    //params.addConstraint("score_file", "file containing baseline Moses scores for input sentences, used if model = monoling-discrim", true); // optional
    params.addConstraint("solo_words_file", "file containing \"solo words\" (words to force to be in their own patterns), one per line", true); // optional
    params.addConstraint("srclang_input_file", "file containing source language input sentences, only used if model = biling", true); // optional
    params.addConstraint("srctgt_align_file", "file containing word alignments from source language to target language, only used if model = biling", true); // optional
    params.addConstraint("srclang_solo_words_file", "file containing source language solo words, only used if model = biling", true); // optional
	params.addConstraint("model", "type of model to train; one of {monoling, biling} (default: monoling)", true); // optional
	//params.addConstraint("model", "type of model to train; one of {monoling, monoling-discrim, biling, biling-null} (default: monoling)", true); // optional
	params.addConstraint("num_iters", "number of iterations for which to run Gibbs sampling (default: 100)", true); // optional
	params.addConstraint("alpha", "concentration parameter for DP prior (used for both DP priors when model = biling) (default: 100.0)", true);
	//params.addConstraint("alpha2", "value of alpha2 -- concentration parameter for second DP prior (only used if model = monoling-discrim) (default: 1.0)", true);
	params.addConstraint("lambda_w", "parameter for Poisson distribution over number of words in pattern (default: 3)", true);
	params.addConstraint("lambda_s", "parameter for Poisson distribution over number of source words linked to target word in bilingual pattern (only used if model = biling) (default: 1)", true);
	params.addConstraint("init_new_color", "probability of generating a new color during initialization (default: 0.7)", true);
	params.addConstraint("init_propagate", "probability of propagating a target color to a source word across a word link during initialization (default: 0.4)", true);
	params.addConstraint("use_gap_widths", "if true, uses gap widths in gappy patterns (default: false)", true);
	params.addConstraint("max_gap_width", "all wider gaps are binned with this gap width (default: 5)", true);
    params.addConstraint("burnin", "number of iterations of burn-in to disregard when storing cumulative counts (default: 50)", true); // optional

   	if (!params.runConstraints("gaplmtrain")) {
       	return 0;
    }

	string input_file = params.asString("input_file");
	string count_output_file = params.asString("count_output_file");
	string count_output_file2 = params.asString("count_output_file2", "||NOFILESPECIFIED||");

	string score_file = params.asString("score_file", "||NOFILESPECIFIED||");
	string solo_words_file = params.asString("solo_words_file", "||NOFILESPECIFIED||");
	string srclang_solo_words_file = params.asString("srclang_solo_words_file", "||NOFILESPECIFIED||");
	string srclang_input_file = params.asString("srclang_input_file", "||NOFILESPECIFIED||");
	string srctgt_align_file = params.asString("srctgt_align_file", "||NOFILESPECIFIED||");

	string modelType = params.asString("model", "monoling");
	int numIters = atoi(params.asString("num_iters", "100").c_str());
	double alpha = atof(params.asString("alpha", "100.0").c_str());
	double alpha2 = atof(params.asString("alpha2", "1.0").c_str());
	int lambda_w = atoi(params.asString("lambda_w", "3").c_str());
	int lambda_s = atoi(params.asString("lambda_s", "1").c_str());
	double initNewColor = atof(params.asString("init_new_color", "0.7").c_str());
	double initPropagate = atof(params.asString("init_propagate", "0.4").c_str());
	bool useGapWidths = (params.asString("use_gap_widths", "false") == "true");
	int maxGapWidth = atoi(params.asString("max_gap_width", "5").c_str());
	int burnin = atoi(params.asString("burnin", "50").c_str());

	clock_t clock1, clock2;
	clock1 = clock();

	bool bigram = false;
	bool discrim = false;
	bool bilingual = false;
	bool onlyNull = false;
	if (modelType == "monoling") {
		;
	} else if (modelType == "monoling-discrim") {
		discrim = true;
	} else if (modelType == "biling") {
		bilingual = true;
	} else if (modelType == "biling-null") {
		bilingual = true;
		onlyNull = true;
	}

	GapLM gaplm(input_file, count_output_file, score_file, solo_words_file, srclang_input_file, srctgt_align_file,
		srclang_solo_words_file, discrim, bigram, bilingual, onlyNull, alpha, alpha2,
		lambda_w, lambda_s, initNewColor, initPropagate, burnin, useGapWidths, maxGapWidth);

	// Perform inference.
	if (modelType == "monoling") {
		gaplm.trainMonoling(numIters);
	} else if (modelType == "monoling-discrim") {
		gaplm.trainMonolingDiscrim(numIters, count_output_file2);
	} else if (modelType == "biling") {
		gaplm.trainBiling(numIters, count_output_file2);
	} else if (modelType == "biling-null") {
		gaplm.trainBilingNull(numIters);
	} else {
		cerr << "Error: Model type not understood: " << modelType
			 << ". Please choose one of {\"monoling\", \"monoling-discrim\", \"biling\", \"biling-null\"}." << endl;
		cerr << "Exiting..." << endl;
		return 1;
	}

	// Print the total time for inference.
	clock2 = clock();
	double totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "Total training time:\t" << totalTime << endl;

    return 0;
}
