/*
Copyright 2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
//
// File:   rank_pats.cpp
// Author: Kevin Gimpel
//
// Created on May 16, 2011
//

#include "rank_pats.h"

bool comparePatternsProductOfExperts(const Pattern& a, const Pattern& b) {
	return (a.logCondProb + a.logMargProb) > (b.logCondProb + b.logMargProb);
}

bool comparePatternsMargProb(const Pattern& a, const Pattern& b) {
	return a.logMargProb > b.logMargProb;
}

bool comparePatternsCondProb(const Pattern& a, const Pattern& b) {
	return a.logCondProb > b.logCondProb;
}

void getSubstring(const vector<string>& toks, int start, int end, string& result) {
	result = "";
	for (int i = start; i <= end; ++i) {
		if (i > start) result += " ";
		result += toks[i];
	}
}

void addAllNoGapPatterns(map<string, int>& allpats, const vector<string>& toks, int maxNoGapPatLength,
	const dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring>& checkpats, const vector<vector<string> >& substrings) {

	dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring>::const_iterator checkiter;

	for (int i = 0; i < toks.size(); ++i) {
		int limit = i + maxNoGapPatLength;
		if (limit > toks.size()) limit = toks.size();
		for (int j = i; j < limit; ++j) {
			const string& pat = substrings[i][j-i];
			checkiter = checkpats.find(pat);
			if (checkiter != checkpats.end()) {
				allpats[pat] += 1;
			}
		}
	}
}

void precomputeSubstrings(const vector<string>& toks, int maxLength, vector<vector<string> >& substrings) {
	substrings.clear();

	for (int i = 0; i < toks.size(); ++i) {
		vector<string> currSubstrings;
		int limit = i + maxLength;
		if (limit > toks.size()) limit = toks.size();
		for (int j = i; j < limit; ++j) {
			string currSubstring;
			getSubstring(toks, i, j, currSubstring);
			currSubstrings.push_back(currSubstring);

		}
		substrings.push_back(currSubstrings);
	}
}

void addAllGapPatterns(map<string, int>& allpats, const vector<string>& toks, 
	int maxFirstSubstringLength, int maxSecondSubstringLength, int maxThirdSubstringLength,
	const dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring>& checkpats,
	const dense_hash_set<string, MurmurHash2, eqstring>& patFirstPartHash,
	const dense_hash_set<string, MurmurHash2, eqstring>& patFirstSecondPartHash,
	const vector<vector<string> >& substrings) {

	dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring>::const_iterator checkiter;
	dense_hash_set<string, MurmurHash2, eqstring>::const_iterator firstPartCheckIter;
	// Go through all first substrings of the sentence.
	// They can't start at the last or second-to-last word in the sentence, since we need a gap and second substring.
	if (toks.size() > 2) {
		for (int i = 0; i < (toks.size() - 2); ++i) {
			int limit = i + maxFirstSubstringLength;
			// First substrings can't end at the last or second-to-last word in the sentence.
			// They must end before this.
			if (limit > (toks.size() - 2)) limit = (toks.size() - 2);
			for (int j = i; j < limit; ++j) {

				const string& firstPart = substrings[i][j-i];
				firstPartCheckIter = patFirstPartHash.find(firstPart);
				if (firstPartCheckIter != patFirstPartHash.end()) {
					// Go through all substrings after the first substring.
					// Second substrings must start at least two words after the first substring ended.
					// They can end on the last word of the sentence.
					for (int k = j + 2; k < toks.size(); ++k) {
						int limit2 = k + maxSecondSubstringLength;
						if (limit2 > toks.size()) limit2 = toks.size();
						for (int l = k; l < limit2; ++l) {
							//string secondPart;
							const string& secondPart = substrings[k][l-k];
							//getSubstring(toks, k, l, secondPart);
							string fullPat = firstPart + " | " + secondPart;
							checkiter = checkpats.find(fullPat);
							if (checkiter != checkpats.end()) {
								allpats[fullPat] += 1;
							}

							firstPartCheckIter = patFirstSecondPartHash.find(fullPat);
							if (firstPartCheckIter != patFirstSecondPartHash.end()) {
								// Also get 2-gap patterns if the first two parts of the pattern were seen.
								// Go through all substrings after the second substring.
								// Third substrings must start at least two words after the second substring ended.
								// They can end on the last word of the sentence.
								for (int m = l + 2; m < toks.size(); ++m) {
									int limit3 = m + maxThirdSubstringLength;
									if (limit3 > toks.size()) limit3 = toks.size();
									for (int n = m; n < limit3; ++n) {
										//string thirdPart;
										//getSubstring(toks, m, n, thirdPart);
										const string& thirdPart = substrings[m][n-m];

										string fullPat2 = fullPat + " | " + thirdPart;
										checkiter = checkpats.find(fullPat2);
										if (checkiter != checkpats.end()) {
											allpats[fullPat2] += 1;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
}

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("rank_pats -- compute statistics of patterns from a pattern file and corpus file and output ranked list to output file");
    params.addConstraint("input_file", "file containing original input corpus", false); // not optional
    params.addConstraint("pat_file", "file containing pattern file", false); // not optional
    params.addConstraint("output_file", "file containing output patterns, sorted in decreasing order by their rank statistic", false); // not optional
    params.addConstraint("rank_stat", "statistic to use when ranking patterns (can be \"marginal\", \"conditional\", or \"poe\") (default: poe)", true); // optional

   	if (!params.runConstraints("rank_pats")) return 0;

	string inFilename = params.asString("input_file");
	string patFilename = params.asString("pat_file");
	string outFilename = params.asString("output_file");
	string rankStatistic = params.asString("rank_stat", "poe");
	clock_t clock1, clock2;
	clock1 = clock();

	bool rankMarginal = false;
	bool rankConditional = false;
	bool rankPOE = false;
	if (rankStatistic == "marginal") {
		rankMarginal = true;
	} else if (rankStatistic == "conditional") {
		rankConditional = true;
	} else if (rankStatistic == "poe") {
		rankPOE = true;
	} else {
		cerr << "Error: rank statistic \"" << rankStatistic << "\" not understood. "
				"Specify \"marginal\", \"conditional\", or \"poe\". Using \"poe\"..." << endl;
		rankPOE = true;
	}

	// Open the output file.
	ofstream outfile(outFilename.c_str());
	if (!outfile.good()) {
		cerr << "Error opening output file: " << outFilename << endl;
		exit(1);
	}

	// Open the pattern file.
    ifstream patfile(patFilename.c_str());
	if (!patfile.good()) {
		cerr << "Error opening pattern file: " << patFilename << endl;
		exit(1);
	}

	// Open the input file.
    ifstream infile(inFilename.c_str());
	if (!infile.good()) {
		cerr << "Error opening input file: " << inFilename << endl;
		exit(1);
	}

	string line;
	vector<string> linetoks;
	int linectr = 0;

	dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring> patHash;
	dense_hash_set<string, MurmurHash2, eqstring> patFirstPartHash;
	dense_hash_set<string, MurmurHash2, eqstring> patFirstSecondPartHash;
	string emptyString = "";
	patHash.set_empty_key(emptyString);
	patFirstPartHash.set_empty_key(emptyString);
	patFirstSecondPartHash.set_empty_key(emptyString);

	int maxNoGapPatLength = 0;
	int maxGapPatLength = 0;
	int maxNumGaps = 0;

	int maxFirstSubstringLength = 0;
	int maxSecondSubstringLength = 0;
	int maxThirdSubstringLength = 0;

    while (!patfile.eof()) {
		getline(patfile, line);
	    if (line != "") {
			++linectr;
			linetoks.clear();

			linetoks.reserve(3);
			tokenize(line, '\t', linetoks);
			if (linetoks.size() < 3) {
				//cerr << "Warning: number of tokens in line < 3. Line = " << line << ". Skipping line..." << endl;
				outfile << line << endl;
			} else {
				// The first token is the raw count, averaged over all post-burn-in samples.
				double count = atof(linetoks[0].c_str());
				// The second token is the marginal log-probability of the pattern.
				double logprob = atof(linetoks[1].c_str());
				// The third token is the pattern itself.
				string pat = linetoks[2];

				patHash.insert(make_pair(pat, make_pair(count, logprob)));
				if (countChar(pat, '|') == 0) {
					// If the pattern has no gaps:
					int currNoGapPatLength = countChar(pat, ' ') + 1;
					if (currNoGapPatLength > maxNoGapPatLength) maxNoGapPatLength = currNoGapPatLength;
				} else {
					// The pattern does have gaps.
					int numGaps = countChar(pat, '|');
					int currGapPatLength = (countChar(pat, ' ') + 1) - numGaps;
					if (currGapPatLength > maxGapPatLength) maxGapPatLength = currGapPatLength;
					if (numGaps > maxNumGaps) maxNumGaps = numGaps;
					vector<string> parts;
					splitString(linetoks[2], " | ", parts);
					if (parts.size() > 0) {
						int firstSubstringLength = countChar(parts[0], ' ') + 1;
						if (firstSubstringLength > maxFirstSubstringLength) maxFirstSubstringLength = firstSubstringLength;
					}
					if (parts.size() > 1) {
						// Store the first part of this 2-part pattern for faster matching later on.
						patFirstPartHash.insert(parts[0]);

						int secondSubstringLength = countChar(parts[1], ' ') + 1;
						if (secondSubstringLength > maxSecondSubstringLength) maxSecondSubstringLength = secondSubstringLength;
					}
					if (parts.size() > 2) {
						// Store the first two parts of this 3-part pattern for faster matching later on.
						patFirstSecondPartHash.insert(parts[0] + " | " + parts[1]);

						int thirdSubstringLength = countChar(parts[2], ' ') + 1;
						if (thirdSubstringLength > maxThirdSubstringLength) maxThirdSubstringLength = thirdSubstringLength;
					}
				}
			}
		}
	}
	patfile.close();
	cerr << "Read patterns." << endl;
	cerr << "Max no gap pattern length = " << maxNoGapPatLength << endl;
	cerr << "Max first substring length = " << maxFirstSubstringLength << endl;
	cerr << "Max second substring length = " << maxSecondSubstringLength << endl;
	cerr << "Max third substring length = " << maxThirdSubstringLength << endl;

	linectr = 0;
	int maxSubstringLength = maxNoGapPatLength;
	if (maxFirstSubstringLength > maxSubstringLength) maxSubstringLength = maxFirstSubstringLength;
	if (maxSecondSubstringLength > maxSubstringLength) maxSubstringLength = maxSecondSubstringLength;
	if (maxThirdSubstringLength > maxSubstringLength) maxSubstringLength = maxThirdSubstringLength;
	cerr << "Max substring length = " << maxSubstringLength << endl;
	map<string, int> allpats;
    while (!infile.eof()) {
		getline(infile, line);
	    if (line != "") {
			++linectr;
			if (linectr % 100 == 0) cerr << ".";
			if (linectr % 5000 == 0) cerr << linectr << endl;

			linetoks.clear();

			linetoks.reserve(200);
			tokenize(line, ' ', linetoks);
			vector<vector<string> > substrings;
			precomputeSubstrings(linetoks, maxSubstringLength, substrings);

			//cerr << line << "\t\t" << linetoks.size() << endl;
			addAllNoGapPatterns(allpats, linetoks, maxNoGapPatLength, patHash, substrings);
			addAllGapPatterns(allpats, linetoks, maxFirstSubstringLength, maxSecondSubstringLength,
				maxThirdSubstringLength, patHash, patFirstPartHash, patFirstSecondPartHash, substrings);
		}
	}
	cerr << linectr << endl;
	infile.close();

	vector<Pattern> finalpats;
	dense_hash_map<string, pair<double, double>, MurmurHash2, eqstring>::const_iterator iter;

	map<string, int>::iterator iter2;
	for (iter = patHash.begin(); iter != patHash.end(); ++iter) {
		iter2 = allpats.find(iter->first);
		if (iter2 == allpats.end()) {
			cerr << "Warning: pattern " << iter->first << " found in patfile but not in infile!" << endl;
		} else {
			Pattern pattern;
			pattern.pat = iter->first;
			pattern.count = iter->second.first;
			pattern.condProb = iter->second.first / ((double) iter2->second);
			pattern.logCondProb = log(pattern.condProb);//iter->second.first / ((double) iter2->second));
			pattern.logMargProb = iter->second.second;
			finalpats.push_back(pattern);
		}
	}
	//cerr << numOmitted << " patterns omitted due to being empty strings." << endl;

	// Sort the patterns according to the desired criterion.
	if (rankMarginal) {
		sort(finalpats.begin(), finalpats.end(), comparePatternsMargProb);
	} else if (rankConditional) {
		sort(finalpats.begin(), finalpats.end(), comparePatternsCondProb);
	} else if (rankPOE) {
		sort(finalpats.begin(), finalpats.end(), comparePatternsProductOfExperts);
	}

	for (int i = 0; i < finalpats.size(); ++i) {
		outfile << finalpats[i].count << "\t";
		if (rankMarginal) {
			outfile << finalpats[i].logMargProb;
		} else if (rankConditional) {
			outfile << finalpats[i].logCondProb;
		} else if (rankPOE) {
			outfile << (finalpats[i].logCondProb + finalpats[i].logMargProb);
		}
		outfile << "\t" << finalpats[i].pat << endl;
	}
	/*vector<string> lastPattern;
	vector<Pattern> patterns;
	int lastNumGapTokens = 0;
	int currNumGapTokens = 0;
	int lastGapToken = -1;
	int currGapToken = -1;

	for (int i = 0; i < allPatterns.size(); ++i) {
		const vector<string>& compPattern = allPatterns[i].piecesTemplate;
		if (i == 0) {
			patterns.push_back(allPatterns[i]);
		} else {
			if (lastPattern == compPattern) {
				patterns.push_back(allPatterns[i]);
			} else {
				// The current pattern does not equal the last pattern, so we want to sum and
				// print the vector of patterns and then clear the vector.

				if (allPatterns[i-1].numGaps == 1) {
					sumAndPrintPatterns(patterns, outfile, lastPattern, allPatterns[i-1].lastGap);
				}
				patterns.clear();
				patterns.push_back(allPatterns[i]);
			}
		}
		lastPattern = compPattern;
	}
	if (allPatterns[allPatterns.size()-1].numGaps == 1) {
		sumAndPrintPatterns(patterns, outfile, lastPattern, allPatterns[allPatterns.size()-1].lastGap);
	}
*/
	outfile.close();

	clock2 = clock();
	double totalTime = (((double)(clock2 - clock1)) / ((double)(CLOCKS_PER_SEC)));
	cerr << "total time:\t" << totalTime << endl;
    return 0;
}
