/*
Copyright 2010-2011 Kevin Gimpel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
#include "utils.h"
#include <iostream>
#include <assert.h>
#include <math.h>
#include <cstring>

using namespace std;

const double neg_inf = log(0.0);
const double pos_inf = -neg_inf;
const float neg_inf_float = log(0.0);
const float pos_inf_float = -neg_inf_float;

/////////////////////////////////////////////////////////////////////////////////////////////////

// Returns the number of times that c occurs in s.
int countChar(const string& s, char c) {
	int count = 0;
	int slen = s.length();
	for (int i = 0; i < slen; ++i) {
		if (s[i] == c) ++count;
	}
	return count;
}

/////////////////////////////////////////////////////////////////////////////////////////////////

vector<string> tokenize(const string& str,const string& delimiters)
{
	vector<string> tokens;

	// skip delimiters at beginning.
    	string::size_type lastPos = str.find_first_not_of(delimiters, 0);

	// find first "non-delimiter".
    	string::size_type pos = str.find_first_of(delimiters, lastPos);

    	while (string::npos != pos || string::npos != lastPos)
    	{
        	// found a token, add it to the vector.
        	tokens.push_back(str.substr(lastPos, pos - lastPos));

        	// skip delimiters.  Note the "not_of"
        	lastPos = str.find_first_not_of(delimiters, pos);

        	// find next "non-delimiter"
        	pos = str.find_first_of(delimiters, lastPos);
    	}

	return tokens;
}

/////////////////////////////////////////////////////////////////////////////////////////////////

// This is a faster version of the tokenize function above when the delimiter
// is only a single character.
void tokenize(const string& str, char delimiter, vector<string>& tokens) {
	// Skip delimiters at beginning. Find first "non-delimiter".
   	string::size_type lastPos = str.find_first_not_of(delimiter, 0);
	// Find next instance of the delimiter.
   	string::size_type pos = str.find_first_of(delimiter, lastPos);

   	while (string::npos != pos || string::npos != lastPos) {
       	// Add the token from lastPos to pos.
       	tokens.push_back(str.substr(lastPos, pos - lastPos));

       	// Find the start of the next token.
       	lastPos = str.find_first_not_of(delimiter, pos);

       	// Find the end of the token.
       	pos = str.find_first_of(delimiter, lastPos);
   	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////

int splitString(const string& input, const string& delimiter, vector<string>& results, bool includeEmpties) {
    int iPos = 0;
    int newPos = -1;
    int sizeS2 = (int)delimiter.size();
    int isize = (int)input.size();

    if( ( isize == 0 ) || ( sizeS2 == 0 ) ) {
        return 0;
    }
    vector<int> positions;
    newPos = input.find (delimiter, 0);
    if( newPos < 0 ) {
        return 0;
    }

    int numFound = 0;

    while( newPos >= iPos ) {
        numFound++;
        positions.push_back(newPos);
        iPos = newPos;
        newPos = input.find (delimiter, iPos+sizeS2);
    }

    if( numFound == 0 ) {
        return 0;
    }

    for (int i=0; i <= (int)positions.size(); ++i) {
        string s("");
        if( i == 0 ) {
            s = input.substr( i, positions[i] );
        }
        int offset = positions[i-1] + sizeS2;
        if( offset < isize ) {
            if ( i == (int)positions.size() ) {
                s = input.substr(offset);
            }
            else if( i > 0 ) {
                s = input.substr( positions[i-1] + sizeS2,
                      positions[i] - positions[i-1] - sizeS2 );
            }
        }
        if( includeEmpties || ( s.size() > 0 ) ) {
            results.push_back(s.c_str());
        }
    }
    return numFound;
}

/////////////////////////////////////////////////////////////////////////////////////////////////

double logAdd(const double lx, const double ly) {
 /* never seen this happen ... but you might want to put it back in
      if(lx == pos_inf)
      return lx;
      if(ly == pos_inf)
      return ly;
 */
 if(lx == neg_inf) return ly;
 if(ly == neg_inf) return lx;
 double d = lx - ly;
 if (d >= 0) {
   if (d > 50) return lx;
   else return (double)(lx + log(1.0 + exp(-d)));
 }
 else {
   if (d < -50) return ly;
   else return (double)(ly + log(1.0 + exp(d)));
 }
}

/////////////////////////////////////////////////////////////////////////////////////////////////

// Converts int to string.
string itos(int i) {
		stringstream s;
		s << i;
		return s.str();
}

/////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * Returns log(x!), computed in the log domain to avoid overflow.
 */
double logFactorial(int x) {
	double ret = 0.0;
	while (x > 1) {
		ret = ret + log(x);
		x--;
	}
	return ret;
}

/////////////////////////////////////////////////////////////////////////////////////////////////

// Returns "n choose k", i.e., n! / ((n-k)! k!).
// Note: if overflow is experienced, change this implementation
// to use the log domain or write a new version of this function
// that uses the log domain.
int nchoosek(int n, int k) {
	if (k > n) return 0;
	// We use doubles because they can store a wider range of values than ints.
	double resultNum = 1.0;
	double resultDenom = 1.0;

	// Compute denominator: k(k-1)(k-2)...(2)(1)
	for (int i = 1; i <= k; ++i) {
		resultDenom *= i;
	}
	// Compute numerator: n(n-1)(n-2)...(n-k+1)
	for (int i = n - k + 1; i <= n; ++i) {
		resultNum *= i;
	}
	return (resultNum / resultDenom);
}

